/* movemail foo bar -- move file foo to file bar,
   locking file foo the way /bin/mail respects.
   Copyright (C) 1986, 1992, 1993, 1994, 1996, 1997 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/*
 * Added May, 1997 by TSUKAHARA, Hiroki <asuka@ba2.so-net.or.jp>
 *
 * Progress dialog.
 * It works only under x86-win32 systems.
 * (It requires progdlg.rc)
 *
 */

/*
 * Modified May, 1997 by YAMAGUCHI, Shuhei <yamagus@kw.netlaputa.or.jp>
 *
 * Show window without activation.
 */

#include <windows.h>
#include <commctrl.h>
void progress_open(int max);
void progress_stepit();
void progress_close();

static HINSTANCE s_hInst = NULL;
static HWND s_hWnd = NULL;
static int s_max;
static int s_current;


BOOL CALLBACK 
DlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam){
  static char buf[32];
  switch(uMsg){
  case WM_INITDIALOG:
    {
      RECT rect;
      LONG x,y;
      GetWindowRect(GetDesktopWindow(), &rect);
      x = (rect.right - rect.left)/2;
      y = (rect.bottom - rect.top)/2;
      MoveWindow(hWnd, x - (200/2), y - (100/2), 200, 100, TRUE);
      sprintf(buf, "%d/%d", s_current, s_max);
      SendMessage(GetDlgItem(hWnd, 2), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)buf);
    }
    return FALSE;

  case WM_COMMAND:
    {
      /* Progress Bar Control */
      SendMessage(GetDlgItem(hWnd, 1), PBM_STEPIT, 0, 0);
      sprintf(buf, "%d/%d", s_current, s_max);
      SendMessage(GetDlgItem(hWnd, 2), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)buf);
    }
    return TRUE;
  }
  return FALSE;
}

void progress_open(int max){
  HWND hwndProgress;
  if( s_hInst != NULL ){
    return;
  }
  InitCommonControls();

  s_max     = max;
  s_current = 0;

  s_hInst = GetModuleHandle(NULL);
  s_hWnd  = CreateDialog(s_hInst, MAKEINTRESOURCE(1), NULL, DlgProc);
  hwndProgress = GetDlgItem(s_hWnd, 1);
  SendMessage(hwndProgress, PBM_SETRANGE, 0, MAKELPARAM(0,max));
  SendMessage(hwndProgress, PBM_SETSTEP, 1, 0);
  SendMessage(hwndProgress, PBM_SETPOS, 0, 0);

  /* modified by yamagus */
  SetWindowPos(s_hWnd, HWND_TOPMOST, 0, 0, 0, 0,
	  SWP_NOSIZE|SWP_NOMOVE|SWP_NOACTIVATE|SWP_SHOWWINDOW);
  /* ShowWindow(s_hWnd, SW_SHOW); */
  UpdateWindow(s_hWnd);
}

void progress_stepit(){
  if( s_hWnd != 0 ){
    s_current++;
    if( s_current <= s_max ){
      SendMessage(s_hWnd, WM_COMMAND, 0, 0);
    }
  }
}

void progress_close(){
  DestroyWindow(s_hWnd);
  s_hWnd  = NULL;
  s_hInst = NULL;
}
