;;; win32-script.el
;;; Author: yamagus@kw.netlaputa.or.jp (YAMAGUCHI, Shuhei)
;;; Modified by H.Miyashita.
;;; Version 1.0 (Jul 08, 1997)
;; (require 'win32-script)
;; (define-process-argument-editing
;;   "/[^./]+\\'"
;;   'script-process-argument-editing-function 'last)

(defconst win32-script-version "Win32 Script version 1.0")
;; begin --- options
(defvar script-process-argument-editing-alist
  '(("/sh$" . "sh.exe")
    ("/bash$" . "bash.exe")
    ("/perl$" . "perl.exe")
    ("/t?csh$" . "tcsh.exe")))
(defvar script-process-permit-recurrence nil)
;; end --- options
(defvar script-process-pathext '(".com" ".exe" ".bat" ".cmd"))
(defvar script-process-bufsiz 256)
(defvar script-process-buffer-tmp " *script process*")

(defun win32-openp (command-name)
  "Locate the full path name of external-command COMMAND-NAME."
  (interactive "sExternal-command: ")
  (catch 'tag
    (let (path)
      (if (file-name-absolute-p command-name)
	  (if (and (file-executable-p command-name)
		   (null (file-directory-p command-name)))
	      (throw 'tag command-name)
	    (mapcar
	     (lambda (suf)
	       (setq path (expand-file-name (concat command-name suf)))
	       (and (file-executable-p path)
		    (null (file-directory-p path))
		    (throw 'tag path)))
	     (if (null script-process-pathext)
		 '("")
	       script-process-pathext)))
	(mapcar
	 (lambda (dir)
	   (mapcar
	    (lambda (suf)
	      (setq path (expand-file-name (concat command-name suf) dir))
	      (and (file-executable-p path)
		   (null (file-directory-p path))
		   (throw 'tag path)))
	    (if (null script-process-pathext)
		'("")
	      (append (list "") script-process-pathext))))
	 exec-path))) nil))

(defun script-process-get-command (script-path)
  "Find out external-command that executes the script SCRIPT-PATH.
Return value is a list of arguments, and car of the list is argv[0]."
  (interactive "fScript: ")
  (let ((buf (generate-new-buffer script-process-buffer-tmp))
	limit args)
    (unwind-protect
	(save-excursion
	  (set-buffer buf)
	  (condition-case nil
	      (progn
		(insert-file-contents script-path nil 0 script-process-bufsiz)
		(goto-line 2)
		(setq limit (point))
		(goto-char 1)
		(if (re-search-forward
 		     "\\`#![ \t]*\\([^ \t\n]+\\)[ \t]*" limit t)
		    (while
			(progn
			  (setq args
				(nconc args
				       (list
					(buffer-substring (match-beginning 1)
							  (match-end 1)))))
			  (re-search-forward "\\([^ \t\n]+\\)[ \t]*"
					     limit t))))
		args)
	    (file-error nil)))
      (kill-buffer buf))))
	    

(defun script-process-argument-editing-function (argument)
  "Find out external-command that executes the script SCRIPT-PATH.
Return value is a list of arguments, and car of the list is argv[0]."
  (interactive)
  (let ((argv0 (car argument)) sargs)
    (if (setq sargs (script-process-get-command argv0))
	(let (func ret)
	  (setq argv0 (car sargs))
	  (catch 'tag
	    (mapcar
	     (lambda (pat)
	       (and (string-match (car pat) argv0)
		    (setq argv0 (win32-openp (cdr pat)))
		    (throw 'tag t)))
	     script-process-argument-editing-alist))
	  (and (eq (setq func (find-process-argument-editing-function argv0))
		   (function script-process-argument-editing-function))
	       (not script-process-permit-recurrence)
	       (setq func default-process-argument-editing-function))
	  (if (consp (setq ret (funcall
				func
				(append (list argv0) (cdr sargs) argument))))
	      ret
	    (cons argv0 ret)))
      (funcall default-process-argument-editing-function argument))))

(provide 'win32-script)
