;;;; Mule for Win32 Specific Lisp Code.
;;;; Author: H.Miyashita.

(defvar ime-toggle-key nil
  "This is IME toggle key. nil means that it is not bind.")
(defvar win32-last-selection nil
  "It is stored the last data from Emacs.")
(defvar win32-isearch-ime-mode nil
  "Indicate isearch ime mode.")
(defvar win32-ime-on-hook nil
  "Functions to eval when IME is turned on at least.\n\
Even if IME state is not changed, these functiona are maybe called.")
(defvar win32-ime-off-hook nil
  "Functions to eval when IME is turned off at least.\n\
Even if IME state is not changed, these functiona are maybe called.")

(setq search-highlight t)

;; isearch ime keymap 
(setq isearch-ime-keymap (copy-keymap minibuffer-local-map))
(nconc isearch-ime-keymap (list (make-vector 26 'isearch-exit-win32ime)))
(define-key isearch-ime-keymap [compend] 'exit-minibuffer)
(define-key isearch-ime-keymap [kanji] 'isearch-exit-win32ime)
(define-key isearch-ime-keymap "\C-s" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\C-r" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\177" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\C-g" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\C-q" 'isearch-command-win32ime)

(define-key isearch-ime-keymap "\r" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\C-j" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\t" 'isearch-command-win32ime)
(define-key isearch-ime-keymap " " 'isearch-command-win32ime)
    
(define-key isearch-ime-keymap "\C-w" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\C-y" 'isearch-command-win32ime)

(let ((meta-map (make-sparse-keymap)))
  (define-key isearch-ime-keymap (char-to-string meta-prefix-char) meta-map)
  (define-key isearch-ime-keymap [escape] meta-map))
(define-key isearch-ime-keymap (vector meta-prefix-char t) 'isearch-exit-win32ime)
(define-key isearch-ime-keymap "\M-n" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\M-p" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\M-y" 'isearch-command-win32ime)
(define-key isearch-ime-keymap "\M-\t" 'isearch-command-win32ime)

(defvar toroku-region-yomigana nil
  "* if this variable is string, toroku-region regard this value as yomigana.")

(defun toroku-region (begin end)
  (interactive "r")
  (let ((string (buffer-substring begin end))
	(win32-isearch-ime-mode t)            ;;Very bogus!!!
	(reading toroku-region-yomigana))
    (save-excursion
      (set-buffer (window-buffer (minibuffer-window)))
      (fep-force-on nil)
      (toggle-win32ime-mode)
      (win32-set-ime-mode 'hiragana)
      (if (not (stringp reading))
	  (setq reading (read-from-minibuffer (format "\"%s\"$B$NFI$_(B:" string))))
      (win32-ime-register-word-dialog reading string))
    (if (not fep-mode) (fep-force-off nil))))

;----------------------------------------
(defvar process-argument-editing-alist nil)
(defvar default-process-argument-editing-function
  (lambda (x) (general-process-argument-editing-function
	       x 'msvc t)))

(defun remove-process-argument-editing (process)
  (let ((curelem process-argument-editing-alist))
    (if (string= (car (car curelem)) process)
	(setq process-argument-editing-alist
	      (cdr process-argument-editing-alist))
      (while (progn
	       (if (not (string= (car (car (cdr curelem))) process))
		   (setq curelem (cdr curelem))
		 (setcdr curelem (cdr (cdr curelem)))
		 nil))))))

(defun define-process-argument-editing
  (process function &optional method)
  (indirect-function function)
  (let ((elem (cons process function))
	(oelem (assoc process process-argument-editing-alist)))
    (cond ((eq method 'last)
	   (remove-process-argument-editing process)
	   (nconc process-argument-editing-alist (list elem)))
	  ((eq method 'first)
	   (remove-process-argument-editing process)
	   (setq process-argument-editing-alist
		 (cons elem process-argument-editing-alist)))
	  ((eq method 'append)
	   (if oelem
	       nil
	     (setq process-argument-editing-alist
		   (cons elem process-argument-editing-alist))))
	  ((eq method 'replace)
	   (if oelem
	       (setcdr oelem function)))
	  (t
	   (if oelem
	       (setcdr oelem function)
	     (setq process-argument-editing-alist
		   (cons elem process-argument-editing-alist)))))))

(defun find-process-argument-editing-function (process)
  (let ((alist process-argument-editing-alist)
	(elem nil))
    (while (and (null elem) (setq elem (car alist)))
      (if (string-match (car elem) process)
	  (setq elem (cdr elem))
	(setq alist (cdr alist))
	(setq elem nil)))
    (if elem
	elem
      default-process-argument-editing-function)))

(defun msvc-process-argument-quoting (arg)
  (mapcar (lambda (x)
	    (let ((start 0) (result "\"") pos end)
	      (while (string-match "\\\\*\"" x start)
		(setq pos (match-beginning 0)
		      end (match-end 0)
		      result (concat result
				     (substring x start pos)
				     (make-string (* (- end pos 1) 2) ?\\ )
				     "\\\"")
		      start end))
	      (concat result (substring x start) "\"")))
	  arg))

(defun cygnus-process-argument-quoting (arguments)
  (mapcar (lambda (arg)
	     (let ((result "\"") (start 0) pos)
	       (while (string-match "\"" arg start)
		 (setq pos (match-end 0)
		       result (concat result
				      (substring arg start pos) "\"")
		       start pos))
	       (concat result (substring arg start) "\"")))
	  arguments))

(defun general-process-argument-editing-function
  (argument quoting argv0isp &optional ep h2sp qp s2isp)
  (setq argument (cond ((eq quoting 'msvc)
			(msvc-process-argument-quoting argument))
		       ((eq quoting 'cygnus)
			(cygnus-process-argument-quoting argument))
		       (t
			argument)))
  (if (null argv0isp)
      (unix-to-dos-argument (mapconcat (function concat) argument " ")
			    ep h2sp qp s2isp)
    (concat
     (unix-to-dos-filename (car argument)) " "
     (unix-to-dos-argument (mapconcat (function concat) (cdr argument) " ")
			   ep h2sp qp s2isp))))

(define-process-argument-editing
  "/\\(tcp\\|movemail\\|fakemail\\|emacsserver\\|m2ps\\|hexl\\|wakeup\\|ctags\\|etags\\|ftp\\)\\.exe\\'"
  (lambda (x) (general-process-argument-editing-function x 'msvc t)))

(define-process-argument-editing
  "/cmd\\.exe\\'"
  (lambda (x) (general-process-argument-editing-function x nil t t nil t t)))

(define-process-argument-editing
  "/command\\.com\\'"
  (lambda (x) (general-process-argument-editing-function x nil t t nil t t)))

(define-process-argument-editing
  "/tcsh\\.exe\\'"
  (lambda (x) (general-process-argument-editing-function x 'msvc t)))

(define-process-argument-editing
  "/bash\\.exe\\'"
  (lambda (x) (general-process-argument-editing-function x 'cygnus t)))

(define-process-argument-editing
  "/trr.*\\.exe\\'"
  (lambda (x) (general-process-argument-editing-function x 'msvc t)))


;-------------------------------------------
(defun mouse-win32-fontset-select (&rest fontset-sel)
  (interactive
   (win32-popup-menu last-nonmenu-event
		     (let ((fslist (win32-fontset-list)))
		       (list "Win32 fontset Menu"
			     (cons "Select fontset"
				   (mapcar (function 
					    (lambda (fse) (list fse fse)))
					   fslist))))))
  (if fontset-sel
      (modify-frame-parameters (selected-frame)
			      (list (cons 'fontset (car fontset-sel))))))
(global-set-key [C-down-mouse-2] 'mouse-win32-fontset-select)

;;;; Created by Keisuke.Ai
(defun win32-mouse-menu-choose-yank ()
  (interactive)
  (let* ((pos (mouse-pixel-position))
	 (x (- (car (cdr pos)) 30))
	 (y 0)
	 (win (frame-first-window (car pos)))
	 (fake-event
	  (list last-input-event (list win 0 (cons x y) 0))))
    (mouse-menu-choose-yank fake-event)))

(put 'win32-mouse-menu-choose-yank 'menu-enable 'kill-ring)

(substitute-key-definition
 'mouse-menu-choose-yank 'win32-mouse-menu-choose-yank
 (current-global-map))
;;;;

(defun win32-mouse-rctrl-rbutton-emulation (button up modifier)
  (progn
    (if (= button 1)
	(if (win32-get-key-state 163)
	    (progn
	      (setq button 2)
	      (if (not (win32-get-key-state 162))
		(setq modifier (logand modifier (lognot (lsh 1 22))))))))
    (list button up modifier)))

(defun win32-mouse-operation-init ()
  (if (= (win32-get-system-metrics 43) 3)
      (progn
	(setq win32-lbutton-to-emacs-button 0)
	(setq win32-mbutton-to-emacs-button 1)
	(setq win32-rbutton-to-emacs-button 2)
	(global-set-key [C-down-mouse-3] 'mouse-win32-fontset-select)
	)))

(defun win32-ime-toggle ()
  (interactive)
  (if (fep-get-mode) (fep-force-off t) (fep-force-on t)))

(defun set-ime-toggle-key (key)
  (interactive "kIME toggle key:")
  (setq ime-toggle-key key)
  (define-key isearch-mode-map key 'isearch-fep-win32ime)
  (define-key isearch-ime-keymap key 'isearch-exit-win32ime)
  (global-set-key key 'win32-ime-toggle))

(global-set-key [compend] (function ignore))

(defun wrap-function-to-control-IME
  (function interactive-p interactive-arg &optional suffix)
  "Wrap FUNCTION, and IME control is enabled when FUNCTION is called. \n\
An original function is saved to FUNCTION-SUFFIX when suffix is string. \n\
If SUFFIX is nil, \"-original\" is added. "
  (let ((original-function
	 (intern (concat (symbol-name function)
			 (if suffix suffix "-original")))))
    (cond ((not (fboundp original-function))
	   (fset original-function
		 (symbol-function function))
	   (fset function
		 (list 
		  'lambda '(&rest arguments)
		  (if interactive-p
		      (list 'interactive interactive-arg))
		  (` (cond ((fep-get-mode)
			    (fep-force-off)
			    (unwind-protect
				(apply '(, original-function) arguments)
			      (if fep-mode
				  (fep-force-on))))
			   (t
			    (apply '(, original-function)
				   arguments))))))))))


;;; Emulate x-color-values
(defun x-color-values (name)
  (let ((value (win32-color-values name)))
    (mapcar (lambda (x) (ash x 8))
            value)))

;;; for clipboard and selection

(defun win32-select-text (text &optional push)
  (if x-select-enable-clipboard 
      (win32-set-clipboard-data text))
  (setq win32-last-selection text))

(defun win32-get-selection-value ()
  (if x-select-enable-clipboard 
      (let (text)
	
	;; Don't die if x-get-selection signals an error.
	(condition-case c
	    (setq text (win32-get-clipboard-data))
	  (error (message "dddd %s" c)))
	(if (or (string= text "") 
		(string= text win32-last-selection))
	    (setq text nil))
	text)))

;;;
;;; Win32 IME Control	by TAKADA, Toshihiro (takadat@st.rim.or.jp)
;;;

;; Copyright (C) 1996 TAKADA, Toshihiro
;; Modified by H.Miyashita.
;; Win32 IME Control is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; Win32 IME Control is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; $B$H$$$&$h$&$J%3%T!<%i%$%HI=<($r<h$j4:$($:=q$$$F$_$^$7$?$,!"<B$O(B
;;; win32ime.el $B$O(B egg.el [Coded by S.Tomura, Electrotechnical Lab.
;;; (tomura@etl.go.jp)] $B$r%,%j%,%j:o$C$F=q$-D>$7$?$@$1$G$9!#(B:-)

;;;
;;; $BMzNr(B
;;;
;;;	1996.05.23	egg.el$B$r;29M$K$7$F:n@.(B
;;;     1996.10.13      Mule for Win32 W1.10$B$N$?$a$K2~B$!#(B

;;;

;;;
;;; $BDj?t(B/$BJQ?t(B
;;;

;;; kill-all-local-variables $B$+$iJ]8n$9$k(B local variables $B$r;XDj$G$-$k(B
;;; $B$h$&$KJQ99$9$k!#(B
(defconst *protected-local-variables* 
  (append 
;   '(win32ime:*mode-on*
   '(
     mode-line-win32ime-mode)
   *protected-local-variables*)
  "List of buffer local variables protected from \"kill-all-local-variables\" ."
  )

;;; $B%_%K%P%C%U%!$N%b!<%I$r%_%K%P%C%U%!$KI=<($9$k$+H]$+(B
(defconst display-minibuffer-mode-in-minibuffer t)

;;; FEP$B%*%s%b!<%II=<($H%*%U%b!<%II=<((B
(defconst win32ime-mode-indicator "$B$"(B")
(defconst transparent-mode-indicator "--")

;;; $B8=:_$N%b!<%II=<($,F~$l$i$l$kJQ?t(B
(defvar mode-line-win32ime-mode transparent-mode-indicator)
(defvar mode-line-win32ime-mode-in-minibuffer transparent-mode-indicator)

;;; $B%_%K%P%C%U%!$,A*Br$5$l$F$$$k$+H]$+(B
(defvar minibuffer-window-selected nil)

;;;; FEP$B%b!<%I$r<($9JQ?t(B
;(defvar win32ime:*mode-on* nil "T if win32ime mode is on.")

;;; $B%b!<%I%i%$%s$KI=<($9$Y$-%_%K%P%C%U%!$N%b!<%I$,$"$k$H$-$O(BT
(defvar display-minibuffer-mode nil)

;;; $B%_%K%P%C%U%!$N%b!<%II=<(MQ$N0l;~JQ?t(B
(defvar win32ime:minibuffer-preprompt '("[" nil "]"))

;;; $B%P%C%U%!%m!<%+%k$JJQ?t$K$9$k(B
;(make-variable-buffer-local 'win32ime:*mode-on*)
(make-variable-buffer-local 'mode-line-win32ime-mode)

;;;
;;; $B%&%#%s%I%&$NA*Br$r@Z$j49$($?;~$N=hM}(B
;;; 
(defun win32ime:set-selected-window-buffer-hook (old new)
  (save-excursion
;    (set-buffer old)
;    (setq fep-mode (fep-get-mode))
    (set-buffer new)
    (if (not win32-isearch-ime-mode)
	(if (not (eq (fep-get-mode) fep-mode))
	    (cond
	     (fep-mode
		(fep-force-on nil)
		(run-hooks 'win32-ime-on-hook))
	     (t
	      (if (not (does-ime-have-undetermined-strings))
		  (progn
		    (fep-force-off nil)
		    (run-hooks 'win32-ime-off-hook)))))))))

(defun win32ime:select-window-hook (old new)
  (save-excursion
;    (set-buffer (window-buffer old))
;    (setq fep-mode (fep-get-mode))
    (set-buffer (window-buffer new))
    (if (not win32-isearch-ime-mode)
	(if (not (eq (fep-get-mode) fep-mode))
	    (cond
	     (fep-mode
		(fep-force-on nil)
		(run-hooks 'win32-ime-on-hook))
	     (t
	      (if (not (does-ime-have-undetermined-strings))
		  (progn
		    (fep-force-off nil)
		    (run-hooks 'win32-ime-off-hook)))))))
    (if (and (eq old (minibuffer-window))
	     (not (eq new (minibuffer-window))))
	(progn
	  (set-buffer (window-buffer (minibuffer-window)))
;	(fep-force-off)
	  (setq fep-mode
		nil
		mode-line-win32ime-mode-in-minibuffer
		transparent-mode-indicator
		minibuffer-preprompt 
		nil))))
  (if (eq new (minibuffer-window))
      (setq minibuffer-window-selected t)
    (setq minibuffer-window-selected nil)))

;;;
;;; $B%b!<%II=<($NJ8;zNs$r0z?t$K<h$j!"$=$l$r@_Dj$9$k!#(B
;;;
(defun win32ime:mode-line-win32ime-mode-update (str)
  (if (eq (current-buffer) (window-buffer (minibuffer-window)))
      (if display-minibuffer-mode-in-minibuffer
	  (progn
	    (setcar (nthcdr 1 win32ime:minibuffer-preprompt) str)
	    (setq minibuffer-preprompt win32ime:minibuffer-preprompt))
	  (setq display-minibuffer-mode t
		mode-line-win32ime-mode-in-minibuffer str))
      (setq display-minibuffer-mode nil
	    mode-line-win32ime-mode str))
  (set-buffer-modified-p (buffer-modified-p)))

;;;
;;; $B%b!<%II=<($NJ8;zNs(B($BNc$($P(B"--"$B$H(B"$B$"(B")$B$N@Z49!#(B
;;;
(defun win32ime:mode-line-display ()
  (if fep-mode
      (win32ime:mode-line-win32ime-mode-update win32ime-mode-indicator)
      (win32ime:mode-line-win32ime-mode-update transparent-mode-indicator)))

;;;
;;; $B%b!<%I%i%$%s$N%U%)!<%^%C%H$NJQ99(B ($B$3$l$O=i4|2=;~$K0l2s8F$P$l$k$@$1(B)
;;;
(defun win32ime:find-symbol-in-tree (item tree)
  (if (consp tree)
      (or (win32ime:find-symbol-in-tree item (car tree))
	  (win32ime:find-symbol-in-tree item (cdr tree)))
    (equal item tree)))

(defun win32ime:init-win32ime-mode-line-format ()
  (if (not (win32ime:find-symbol-in-tree
	    'mode-line-win32ime-mode mode-line-format))
      (setq-default 
       mode-line-format
       (cons (list 'mc-flag
		   (list 'display-minibuffer-mode-in-minibuffer
		         ;; $B%_%K%P%C%U%!$N%b!<%I$r%_%K%P%C%U%!$N@hF,$K=P$9>l9g(B
			 (list "[" 'mode-line-win32ime-mode "]")
		         ;; $B%_%K%P%C%U%!$N%b!<%I$r%b!<%I%i%$%s$K=P$9>l9g(B
			 (list 
			  (list 'minibuffer-window-selected
				(list 'display-minibuffer-mode
				      "m"
				      " ")
				" ")
			  "["
			  (list 'minibuffer-window-selected
				(list 'display-minibuffer-mode
				      'mode-line-win32ime-mode-in-minibuffer
				      'mode-line-win32ime-mode)
				'mode-line-win32ime-mode)
			  "]"
			  )))
	     mode-line-format))))
  
;;;
;;; IME$B$N%H%0%k(B
;;;
(defun toggle-win32ime-mode ()
  (interactive)
  (if mc-flag
      (let ((ime-state (fep-get-mode)))
	(if ime-state
	    (run-hooks 'win32-ime-on-hook)
	  (run-hooks 'win32-ime-off-hook))
	(if (not (eq ime-state fep-mode))
	    (progn
	      (setq fep-mode ime-state)
	      (win32ime:mode-line-display))))))

;;;
;;; $B=i4|2=(B
;;;

(provide 'win32misc)