/* Utility and Unix shadow routines for GNU Emacs on Windows NT.
   Copyright (C) 1994, 1995 Free Software Foundation, Inc.

   This file is part of GNU Emacs.

   GNU Emacs is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any later
   version.

   GNU Emacs is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with GNU Emacs; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Geoff Voelker (voelker@cs.washington.edu)                         7-29-94
*/

/* Define stat before including config.h.  */
#include <string.h>
#include <sys/stat.h>
int
nt_stat (char *filename, struct stat *statbuf)
{
  int r, l = strlen (filename);
  char *str = NULL;
  extern long *xmalloc ();
  extern void xfree ();
#ifdef MCPATH     /* 95.8.1 by himi */
  extern int mc_stat(char *path, struct stat *st);
#endif

  /* stat has a bug when passed a name of a directory with a trailing
     backslash (but a trailing forward slash works fine).  */
  /* On Windows95, Anyway, a last letter is a directory separater,
     it works wrong (by himi) */
  if ((l > 1) &&
      ((filename[l - 1] == '\\') || (filename[l - 1] == '/')) &&
      (filename[l - 2] != ':'))
    {
      str = (char *) xmalloc (l);
      memcpy (str, filename, l);
      str[l - 1] = (char)NULL;
#ifdef MCPATH
      r = mc_stat (str, statbuf);
#else
      r = stat (str, statbuf);
#endif
      xfree (str);
      return r;
    }
  else
#ifdef MCPATH
    return mc_stat (filename, statbuf);
#else
    return stat (filename, statbuf);
#endif
}

/* Place a wrapper around the NT version of ctime.  It returns NULL
   on network directories, so we handle that case here.  
   Define it before including config.h.  (Ulrich Leodolter, 1/11/95).  */
char *
nt_ctime (const time_t *t)
{
  char *str = (char *) ctime (t);
  return (str ? str : "Sun Jan 01 00:00:00 1970");
}

/* 95.8.1 by himi It is because nt.c is the lowest layer module. */
#ifdef MCPATH
#define MCPATH_SOURCE
#define NONSYSTEM_DIR_LIBRARY
#endif
#include <windows.h>
#ifndef __BORLANDC__ /* 97/01/13 by Y.Matsushima */
#include "config.h"
#endif
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <ctype.h>
#ifdef __BORLANDC__ /* 97/01/13 by Y.Matsushima */
#include <fcntl.h>
#include "config.h"
#endif

#define getwd _getwd
#include "lisp.h"
#undef getwd

#include <pwd.h>

#ifndef MCPATH
#include "ndir.h"
#endif
#include "ntheap.h"

#ifdef MCPATH /* 95.8.2 by himi */
extern unsigned char *decode_path (unsigned char *path,
				   unsigned char ext_path[MC_MAXPATHLEN]);
#endif

extern int report_file_error (char *, Lisp_Object);

/* Get the current working directory.  */
/* modified by himi 95.9.11 */
int
getwd (char *dir)
{
  extern void dostounix_filename (char*);
  int f;
  f = GetCurrentDirectory (MAXPATHLEN, dir);
  dostounix_filename(dir);
  return f;
}

/* Emulate gethostname. */
/* WINAPI is to match with the type of winsock API.(by himi) */  
#undef gethostname

int WINAPI
sys_gethostname (char *buffer, int size)
{
  /* NT only allows small host names, so the buffer is 
     certainly large enough.  */
#ifdef INC_WINSOCK
  int result;
  result = gethostname(buffer, size);
  if (!result) return result;
#endif

  return !GetComputerName (buffer, &size);
}

#ifdef INC_WINSOCK

CRITICAL_SECTION critsec_systemname_access;
char *system_name = NULL;

DWORD WINAPI gethostname_thread(char *hostname)
{
  struct hostent *hp;
  char *p, *fqdn;

  hp = gethostbyname(hostname);
  if (hp)
    {
      fqdn = hp->h_name;

      p = fqdn;
      while(*p && *p != '.') p++;

      if (!*p)
	{
	  /* We still don't have a fully qualified domain name.
	     Try to find one in the list of alternate names */
	  char **alias = hp->h_aliases;
	  while (*alias)
	    {
	      p = *alias;
	      while(*p && *p != '.') p++;
	      if (*p) break;
	      alias++;
	    }
	  if (*alias)
	    fqdn = *alias;
	}
      EnterCriticalSection(&critsec_systemname_access);
      system_name = (char *)xmalloc(strlen(fqdn) + 1);
      strcpy(system_name, fqdn);
      for (p = system_name; *p; p++)
	if (*p == ' ' || *p == '\t')
	  *p = '-';
      LeaveCriticalSection(&critsec_systemname_access);
      return 1;
    }
  return 0;
}

int
set_system_name_async (char *hostname)
{
  DWORD dwThreadId;
  HANDLE hthread;
  if (system_name)
    {
      xfree(system_name);
      system_name = NULL;
    }

  InitializeCriticalSection(&critsec_systemname_access);

  hthread = CreateThread(NULL, 0, gethostname_thread,
			 (LPVOID) hostname, 0, &dwThreadId);
  SetThreadPriority(hthread, THREAD_PRIORITY_ABOVE_NORMAL);
  return 1;
}

#endif

/* Emulate getloadavg.  */
int
getloadavg (double loadavg[], int nelem)
{
  int i;

  /* A faithful emulation is going to have to be saved for a rainy day.  */
  for (i = 0; i < nelem; i++) 
    {
      loadavg[i] = 0.0;
    }
  return i;
}

/* Emulate sleep...we could have done this with a define, but that
   would necessitate including windows.h in the files that used it.
   This is much easier.  */
void
nt_sleep (int seconds)
{
  Sleep (seconds * 1000);
}

/* Emulate the Unix directory procedures opendir, closedir, 
   and readdir.  We can't use the procedures supplied in sysdep.c,
   so we provide them here.  */

struct direct dir_static;       /* simulated directory contents */
static int    dir_finding;
static HANDLE dir_find_handle;

DIR *
opendir (char *filename)
{
  DIR *dirp;

  /* Opening is done by FindFirstFile.  However, a read is inherent to
     this operation, so we have a flag to handle the open at read
     time.  This flag essentially means "there is a find-handle open and
     it needs to be closed."  */

  if (!(dirp = (DIR *) malloc (sizeof (DIR)))) 
    {
      return 0;
    }

  dirp->dd_fd = 0;
  dirp->dd_loc = 0;
  dirp->dd_size = 0;

  /* This is tacky, but we need the directory name for our
     implementation of readdir.  */
  strncpy (dirp->dd_buf, filename, DIRBLKSIZ);
  return dirp;
}

void
closedir (DIR *dirp)
{
  /* If we have a find-handle open, close it.  */
  if (dir_finding) 
    {
      FindClose (dir_find_handle);
      dir_finding = 0;
    }
  xfree ((char *) dirp);
}

struct direct *
readdir (DIR *dirp)
{
  WIN32_FIND_DATA find_data;
  
  /* If we aren't dir_finding, do a find-first, otherwise do a find-next. */
  if (!dir_finding) 
    {
      char filename[MAXNAMLEN + 3];
      int ln;

      strncpy (filename, dirp->dd_buf, MAXNAMLEN);
      ln = strlen (filename)-1;

      /* 95.8.29 by himi */
      if (!IS_DIRECTORY_SEP(filename[ln]) && filename[ln] != ':')
	strcat (filename, "\\");
      strcat (filename, "*.*");

      dir_find_handle = FindFirstFile (filename, &find_data);

      if (dir_find_handle == INVALID_HANDLE_VALUE) 
	return NULL;

      dir_finding = 1;
    } 
  else 
    {
      if (!FindNextFile (dir_find_handle, &find_data))
	return NULL;
    }
  
  /* NT's unique ID for a file is 64 bits, so we have to fake it here.  
     This should work as long as we never use 0.  */
  dir_static.d_ino = 1;
  
  dir_static.d_reclen = sizeof (struct direct) - MAXNAMLEN + 3 +
    dir_static.d_namlen - dir_static.d_namlen % 4;
  
  dir_static.d_namlen = strlen (find_data.cFileName);
  strncpy (dir_static.d_name, find_data.cFileName, MAXNAMLEN);
  
  return &dir_static;
}

/* Emulate getpwuid and getpwnam.  */

int getuid ();	/* forward declaration */

static char the_passwd_name[256];
static char the_passwd_passwd[256];
static char the_passwd_gecos[256];
static char the_passwd_dir[256];
static char the_passwd_shell[256];

static struct passwd the_passwd = 
{
  the_passwd_name,
  the_passwd_passwd,
  0,
  0,
  0,
  the_passwd_gecos,
  the_passwd_dir,
  the_passwd_shell,
};

struct passwd *
getpwuid (int uid)
{
  int size = 256;
  
  if (!GetUserName (the_passwd.pw_name, &size))
    return NULL;

  the_passwd.pw_passwd[0] = '\0';
  the_passwd.pw_uid = 0;
  the_passwd.pw_gid = 0;
  strcpy (the_passwd.pw_gecos, the_passwd.pw_name);
  the_passwd.pw_dir[0] = '\0';
  the_passwd.pw_shell[0] = '\0';

  return &the_passwd;
}

struct passwd *
getpwnam (char *name)
{
  struct passwd *pw;
  
  pw = getpwuid (getuid ());
  if (!pw)
    return pw;

  if (strcmp (name, pw->pw_name))
    return NULL;

  return pw;
}


/* We don't have scripts to automatically determine the system configuration
   for Emacs before it's compiled, and we don't want to have to make the
   user enter it, so we define EMACS_CONFIGURATION to invoke this runtime
   routine.  */

static char configuration_buffer[32];

char *
get_emacs_configuration (void)
{
    char *arch, *oem, *os;

    /* Determine the processor type.  */
    switch (get_processor_type ()) 
      {
      case PROCESSOR_ARCHITECTURE_INTEL:
	arch = "i386";
	break;
#if 0
      case PROCESSOR_INTEL_860:
	arch = "i860";
	break;
#endif
      case PROCESSOR_ARCHITECTURE_MIPS:
	arch = "mips";
	break;
      case PROCESSOR_ARCHITECTURE_ALPHA:
	arch = "alpha";
	break;
      case PROCESSOR_ARCHITECTURE_PPC:
	arch = "powerpc";
	break;
      default:
	arch = "unknown";
	break;
      }

    /* Let oem be "*" until we figure out how to decode the OEM field.  */
    oem = "*";

    if (win32_platform_id == 1)
/* #ifdef WINDOWS95 */
    os = "win";
/* #else */
    else
    os = "nt";
/* #endif */

    sprintf (configuration_buffer, "%s-%s-%s%d.%d", arch, oem, os,
	     get_nt_major_version (), get_nt_minor_version ());
    return configuration_buffer;
}

/* Conjure up inode and device numbers that will serve the purpose
   of Emacs.  Return 1 upon success, 0 upon failure.  */

int
get_inode_and_device_vals (Lisp_Object filename, Lisp_Object *p_inode, 
			   Lisp_Object *p_device)
{
  /* File uids on NT are found using a handle to a file, which
     implies that it has been opened.  Since we want to be able
     to stat an arbitrary file, we must open it, get the info,
     and then close it.
     
     Also, NT file uids are 64-bits.  This is a problem.  */

  HANDLE handle;
  BOOL result;
  DWORD attrs;
  BY_HANDLE_FILE_INFORMATION info;
#ifdef MCPATH     /* 95.8.2 by himi */
  unsigned char decode_buffer[MC_MAXPATHLEN];
  decode_path (XSTRING (filename)->data, decode_buffer);
#endif

  /* FIXME:  It shouldn't be opened without READ access, but NT on x86
     doesn't allow GetFileInfo in that case (NT on mips does).  */
     
#ifdef MCPATH
  handle = CreateFile (decode_buffer,
		       0,
		       FILE_SHARE_READ | FILE_SHARE_WRITE,
		       NULL,
		       OPEN_EXISTING,
		       FILE_FLAG_BACKUP_SEMANTICS,
		       NULL);
#else
  handle = CreateFile (XSTRING (filename)->data,
		       0,
		       FILE_SHARE_READ | FILE_SHARE_WRITE,
		       NULL,
		       OPEN_EXISTING,
		       FILE_FLAG_BACKUP_SEMANTICS,
		       NULL);
#endif

  if (handle == INVALID_HANDLE_VALUE) {
    result = 0;
  } else {
    result = GetFileInformationByHandle (handle, &info);
    CloseHandle (handle);
  }
  if (!result)
    {
#ifdef MCPATH
      attrs = GetFileAttributes (decode_buffer);
#else
      attrs = GetFileAttributes (XSTRING (filename)->data);
#endif
      if (attrs == (DWORD)INVALID_HANDLE_VALUE)
	return 0;

      /* Conjure up bogus, but unique, values.  */
      attrs = GetTickCount ();
      *p_inode = make_number (attrs);
      *p_device = make_number (attrs);
      return 1;
    }

  /* This Expression is to make 32bit file index into VALBITS order number.
     A smaller order number is made by psude-CRC16....;-) */
  *p_inode = XFASTINT((unsigned long) MAKELONG(info.nFileIndexLow,
					       info.nFileIndexHigh) %
		      (unsigned long) ((1 << (VALBITS - 1)) +
				       (1 << (VALBITS - 2)) +
				       (1 << 2) + 1));
  if (!INTEGERP(*p_inode)) abort();

  *p_device = make_number (info.dwVolumeSerialNumber);

  return 1;
}

/* The following pipe routines are used to support our fork emulation.
   Since NT's crt dup always creates inherited handles, we
   must be careful in setting up pipes.  First create 
   non-inherited pipe handles, then create an inherited handle
   to the write end by dup-ing it, and then close the non-inherited
   end that was just duped.  This gives us one non-inherited handle
   on the read end and one inherited handle to the write end.  As
   the parent, we close the inherited handle to the write end after
   spawning the child.  */

#ifndef MULE
/* From callproc.c  */
extern Lisp_Object Vbinary_process_input;
extern Lisp_Object Vbinary_process_output;
#endif

void
pipe_with_inherited_out (int fds[2])
{
  int inherit_out;
  unsigned int flags = _O_NOINHERIT;

#ifdef MULE
  flags |= _O_BINARY;
#else
  if (!NILP (Vbinary_process_output))
    flags |= _O_BINARY;
#endif

  _pipe (fds, 0, flags);
  inherit_out = dup (fds[1]);
  close (fds[1]);
  fds[1] = inherit_out;
}

void
pipe_with_inherited_in (int fds[2])
{
  int inherit_in;
  unsigned int flags = _O_NOINHERIT;

#ifdef MULE
  flags |= _O_BINARY;
#else
  if (!NILP (Vbinary_process_input))
    flags |= _O_BINARY;
#endif

  _pipe (fds, 0, flags);
  inherit_in = dup (fds[0]);
  close (fds[0]);
  fds[0] = inherit_in;
}

/* The following two routines are used to manipulate stdin, stdout, and
   stderr of our child processes.

   Assuming that in, out, and err are inherited, we make them stdin,
   stdout, and stderr of the child as follows:

   - Save the parent's current standard handles.
   - Set the parent's standard handles to the handles being passed in.
     (Note that _get_osfhandle is an io.h procedure that 
     maps crt file descriptors to NT file handles.)
   - Spawn the child, which inherits in, out, and err as stdin,
     stdout, and stderr. (see Spawnve)
   - Reset the parent's standard handles to the saved handles.
     (see reset_standard_handles)
   We assume that the caller closes in, out, and err after calling us.  */

void
prepare_standard_handles (int in, int out, int err, HANDLE handles[4])
{
  HANDLE parent, stdin_save, stdout_save, stderr_save, err_handle;

/* #ifdef WINDOWS95 */

  /* The Win95 beta doesn't set the standard handles correctly.
     Handicap subprocesses until we get a version that works correctly.  
     Undefining the subprocesses macro reveals other incompatibilities,
     so, since we're expecting subprocs to work in the near future, 
     disable them here.  */
  /* On Win95 Console mode can't attach the child process in/out. (by himi) */
  if (win32_platform_id == 1 && inhibit_window_system)
    report_file_error ("Subprocesses disabled on Windows95 console mode",
		       Qnil);

  parent = GetCurrentProcess ();
  stdin_save = GetStdHandle (STD_INPUT_HANDLE);
  stdout_save = GetStdHandle (STD_OUTPUT_HANDLE);
  stderr_save = GetStdHandle (STD_ERROR_HANDLE);

  if (!DuplicateHandle (parent, 
		       GetStdHandle (STD_INPUT_HANDLE), 
		       parent,
		       &stdin_save, 
		       0, 
		       FALSE, 
		       DUPLICATE_SAME_ACCESS))
    report_file_error ("Duplicating parent's input handle", Qnil);
  
  if (!DuplicateHandle (parent,
		       GetStdHandle (STD_OUTPUT_HANDLE),
		       parent,
		       &stdout_save,
		       0,
		       FALSE,
		       DUPLICATE_SAME_ACCESS))
    report_file_error ("Duplicating parent's output handle", Qnil);
  
  if (!DuplicateHandle (parent,
		       GetStdHandle (STD_ERROR_HANDLE),
		       parent,
		       &stderr_save,
		       0,
		       FALSE,
		       DUPLICATE_SAME_ACCESS))
    report_file_error ("Duplicating parent's error handle", Qnil);
  
  if (!SetStdHandle (STD_INPUT_HANDLE, (HANDLE) _get_osfhandle (in)))
    report_file_error ("Changing stdin handle", Qnil);
  
  if (!SetStdHandle (STD_OUTPUT_HANDLE, (HANDLE) _get_osfhandle (out)))
    report_file_error ("Changing stdout handle", Qnil);
  
  /* We lose data if we use the same handle to the pipe for stdout and
     stderr, so make a duplicate.  This took a while to find.  */
  if (out == err) 
    {
      if (!DuplicateHandle (parent,
			   (HANDLE) _get_osfhandle (err),
			   parent,
			   &err_handle,
			   0,
			   TRUE,
			   DUPLICATE_SAME_ACCESS))
	report_file_error ("Duplicating out handle to make err handle.",
			  Qnil);
    } 
  else 
    {
      err_handle = (HANDLE) _get_osfhandle (err);
    }

  if (!SetStdHandle (STD_ERROR_HANDLE, err_handle))
    report_file_error ("Changing stderr handle", Qnil);

  handles[0] = stdin_save;
  handles[1] = stdout_save;
  handles[2] = stderr_save;
  handles[3] = err_handle;
}

void
reset_standard_handles (int in, int out, int err, HANDLE handles[4])
{
  HANDLE stdin_save = handles[0];
  HANDLE stdout_save = handles[1];
  HANDLE stderr_save = handles[2];
  HANDLE err_handle = handles[3];
  int i;

  if (!SetStdHandle (STD_INPUT_HANDLE, stdin_save))
    report_file_error ("Resetting input handle", Qnil);
  
  if (!SetStdHandle (STD_OUTPUT_HANDLE, stdout_save))
    {
      i = GetLastError ();
      report_file_error ("Resetting output handle", Qnil);
    }
  
  if (!SetStdHandle (STD_ERROR_HANDLE, stderr_save))
    report_file_error ("Resetting error handle", Qnil);
  
  if (out == err) 
    {
      /* If out and err are the same handle, then we duplicated out
	 and stuck it in err_handle.  Close the duplicate to clean up.  */
      if (!CloseHandle (err_handle))
	report_file_error ("Closing error handle duplicated from out.", 
			  Qnil);
    }
}

#ifdef __BORLANDC__ /* 97/01/13 by Y.Matsushima */
#undef random()
#endif
int
random ()
{
  /* rand () on NT gives us 15 random bits...hack together 30 bits.  */
  return ((rand () << 15) | rand ());
}

void
srandom (int seed)
{
  srand (seed);
}

/* Destructively turn backslashes into slashes.  */
void
dostounix_filename (p)
     register char *p;
{
  while (*p)
    {
      if (*p == '\\')
	*p = '/';
      p++;
    }
}

/* 95.8.24 by himi */
void
unixtodos_filename (p)
     register char *p;
{
  while (*p)
    {
      if (*p == '/')
	*p = '\\';
      p++;
    }
}

void
unixtodos_argument (p, ep, h2sp, qp, s2isp)
     register char *p;
     int ep, h2sp, qp, s2isp;
{
  int qf;
  char *o;

  qf = 0;
  o = p;
  while (-1)
    {
      if (*p == '\\' && 
	  ep)
	{
	  p++;
	  *o = *p;
	  if (!*p) break;
	  o++;
	  p++;
	  continue;
	}
      if (qf)
	{
	  if (*p == qf) 
	    {
	      qf = 0;
	      p++;
	      continue;
	    }
	  else *o = *p;
	}
      else
	{
	  switch (*p)
	    {
	    case '/':
	      if (!s2isp)
		*o = *p;
	      else
		*o = '\\';
	      break;
	    case '-':
	      if (!h2sp)
		*o = *p;
	      else
		*o = '/';
	      break;

	    case '\'':
	    case '"':
	      if (!qp)
		*o = *p;
	      else
		{
		  qf = *p;
		  p++;
		  continue;
		}
	      break;
	    default :
	      *o = *p;
	      break;
	    }
	}
      if (!*p)
	{
	  *o = *p;
	  break;
	}
      o++;
      p++;
    }
}

/* Routines that are no-ops on NT but are defined to get Emacs to compile.  */


int 
sigsetmask (int signal_mask) 
{ 
  return 0;
}

int 
sigblock (int sig) 
{ 
  return 0;
}

int 
kill (int pid, int signal) 
{ 
  return 0;
}

int 
setpgrp (int pid, int gid) 
{ 
  return 0;
}

int 
alarm (int seconds) 
{ 
  return 0;
}

int 
unrequest_sigio (void) 
{ 
  return 0;
}

int 
request_sigio (void) 
{ 
  return 0;
}

int 
getuid () 
{ 
  return 0; 
}

int 
geteuid () 
{ 
  return 0; 
}

#ifndef MULE
/* Remove all CR's that are followed by a LF.
   (From msdos.c...probably should figure out a way to share it,
   although this code isn't going to ever change.)  */
int
crlf_to_lf (n, buf)
     register int n;
     register unsigned char *buf;
{
  unsigned char *np = buf;
  unsigned char *startp = buf;
  unsigned char *endp = buf + n;

  if (n == 0)
    return n;
  while (buf < endp - 1)
    {
      if (*buf == 0x0d)
	{
	  if (*(++buf) != 0x0a)
	    *np++ = 0x0d;
	}
      else
	*np++ = *buf++;
    }
  if (buf < endp)
    *np++ = *buf++;
  return np - startp;
}
#endif

#ifdef HAVE_TIMEVAL
#include <sys/timeb.h>

/* Emulate gettimeofday (Ulrich Leodolter, 1/11/95).  */
void 
gettimeofday (struct timeval *tv, struct timezone *tz)
{
  struct _timeb tb;
  _ftime (&tb);

  tv->tv_sec = tb.time;
  tv->tv_usec = tb.millitm * 1000L;
  if (tz) 
    {
#if defined(__BORLANDC__) && (__BORLANDC__ < 0x500)
#define timezone  _timezone
#endif
      tz->tz_minuteswest = tb.timezone;	/* minutes west of Greenwich  */
      tz->tz_dsttime = tb.dstflag;	/* type of dst correction  */
#if defined(__BORLANDC__) && (__BORLANDC__ < 0x500)
#undef timezone
#endif
    }
}
#endif /* HAVE_TIMEVAL */


#ifdef PIGSFLY
Keep this around...we might need it later.
#ifdef WINDOWSNT
{
  /*
   * Find the user's real name by opening the process token and looking
   * up the name associated with the user-sid in that token.
   */

  char            b[256], Name[256], RefD[256];
  DWORD           length = 256, rlength = 256, trash;
  HANDLE          Token;
  SID_NAME_USE    User;

  if (1)
    Vuser_real_login_name = build_string ("foo");
  else if (!OpenProcessToken (GetCurrentProcess (), TOKEN_QUERY, &Token))
    {
      Vuser_real_login_name = build_string ("unknown");
    }
  else if (!GetTokenInformation (Token, TokenUser, (PVOID)b, 256,
				 &trash))
    {
      CloseHandle (Token);
      Vuser_real_login_name = build_string ("unknown");
    }
  else if (!LookupAccountSid ((void *)0, (PSID)b, Name, &length, RefD,
			      &rlength, &User))
    {
      CloseHandle (Token);
      Vuser_real_login_name = build_string ("unknown");
    }
  else
    Vuser_real_login_name = build_string (Name);
}
#else   /* not WINDOWSNT */
#endif  /* not WINDOWSNT */
#endif  /* PIGSFLY */

/* 95.8.23 by himi */
DEFUN ("dos-to-unix-filename", Fdos_to_unix_filename, Sdos_to_unix_filename,
       1, 1, 0,
       "Return filename in unix form. To put it concretely, make '\\' in a \n\
filename into '/'.(indestructively)")
     (filename)
     Lisp_Object filename;
{
  int size;
  Lisp_Object rfilename;
  char* filetmp;
  
  CHECK_STRING (filename, 0);
  size = XSTRING(filename) -> size;
  filetmp = (char *) alloca(size + 1);
  if (!filetmp)
    {
      error ("Cannot allocate memory!.");
    }
  memcpy(filetmp, XSTRING(filename) -> data, size + 1);

  dostounix_filename(filetmp);

  return make_string (filetmp, size);
}

DEFUN ("unix-to-dos-filename", Funix_to_dos_filename, Sunix_to_dos_filename,
       1, 1, 0,
       "Return filename in dos form. To put it concretely, make '/' in a \n\
filename into '\\'.(indestructively). ")
     (filename)
     Lisp_Object filename;
{
  int size;
  Lisp_Object rfilename;
  char *filetmp;
  
  CHECK_STRING (filename, 0);
  size = XSTRING(filename) -> size;
  filetmp = (char *) alloca(size + 1);
  if (!filetmp)
    {
      error ("Cannot allocate memory!.");
    }
  memcpy(filetmp, XSTRING(filename) -> data, size + 1);

  unixtodos_filename(filetmp);

  return make_string (filetmp, size);
}

DEFUN ("unix-to-dos-argument", Funix_to_dos_argument,
       Sunix_to_dos_argument,
       5, 5, 0,
       "Return argument in dos form.")
     (filename, ep, h2sp, qp, s2isp)
     Lisp_Object filename, ep, h2sp, qp, s2isp;
{
  char *filetmp;

  CHECK_STRING (filename, 0);

  filetmp = (char *) alloca(XSTRING(filename) -> size + 1);
  if (!filetmp)
    {
      error ("Cannot allocate memory!.");
    }
  memcpy(filetmp, XSTRING(filename) -> data, XSTRING(filename) -> size + 1);
  unixtodos_argument(filetmp, !NILP(ep), !NILP(h2sp), !NILP(qp), !NILP(s2isp));
  
  return build_string(filetmp);

}

DEFUN ("w32-sfn-to-lfn", Fw32_sfn_to_lfn, Sw32_sfn_to_lfn, 1, 1, 0,
       "Convert short filename(8.3) to long filename.\n\
If system fails to retrieve a long filename, return nil.")
     (sfn)
     Lisp_Object sfn;
{
  HANDLE hff;
  WIN32_FIND_DATA w32FindData;
  Lisp_Object directory, fullpath;
  struct gcpro gcpro1, gcpro2;

  CHECK_STRING(sfn, 0);
  GCPRO2(directory, fullpath);

  fullpath = Fexpand_file_name(sfn, Qnil);
  directory = Ffile_name_directory(fullpath);
  hff = FindFirstFile(XSTRING(fullpath)->data, &w32FindData); 
  if(hff == INVALID_HANDLE_VALUE){
    return Qnil;
  }
  fullpath = Fexpand_file_name(build_string(w32FindData.cFileName), directory);
  FindClose(hff);

  UNGCPRO;
  return fullpath;
}

DEFUN ("w32-lfn-to-sfn", Fw32_lfn_to_sfn, Sw32_lfn_to_sfn, 1, 1, 0,
       "Convert long filename to short filename(8.3).\n\
If system fails to retrieve a short filename, return nil.")
     (lfn)
     Lisp_Object lfn;
{
#ifdef WIN32_VER4
  char buf[MAX_PATH+1];
  int  len;
  CHECK_STRING(lfn, 0);
  len = GetShortPathName(XSTRING(lfn)->data, buf, MAX_PATH);
  if (!len) 
    return Qnil;
  return Fexpand_file_name(make_string(buf, len), Qnil);
#else
  return Qnil;
#endif
}
  
void
syms_of_nt_misc ()
{
  defsubr (&Sdos_to_unix_filename);
  defsubr (&Sunix_to_dos_filename);
  defsubr (&Sunix_to_dos_argument);
  defsubr (&Sw32_sfn_to_lfn);
  defsubr (&Sw32_lfn_to_sfn);
}
