/* Win32 clipboard module for MS Windows NT and Windows 3.1
   Copyright (C) 1995 Free Software Foundation, Inc.

This file is part of Mule (MULtilingual Enhancement of GNU Emacs).

Mule is free software distributed in the form of patches to GNU Emacs.
You can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

Mule is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* 95.5.2 created by himi */

#include "config.h"

#ifdef MULE

#include <windows.h>
#include "lisp.h"
#include "coding.h"
#include "buffer.h"
#include "blockinput.h"

#define WRITE_BUF_SIZE 1024


Lisp_Object Vwin32_clipboard_coding_system;

DEFUN ("win32-clipboard-yank", Fwin32_clipboard_yank, Swin32_clipboard_yank,
       0, 1, "",
       "Insert contents of clipboard (TEXT FORMAT) after point.\n\
Returns list of coding-system used for reading and length of data inserted\n\
argument CODE apecifies the coding-system used in the current \n\
clipboard(CF_TEXT) which defaults to win32-clipboard-coding-system.")
 (code)
       Lisp_Object code;
{
  HANDLE clipboard_handle;
  HGLOBAL clipboard_data;
  int size;
  int inserted = 0, converted;
  coding_type mccode;
  Lisp_Object found_code = code;

  if (!NILP (current_buffer->read_only))
    Fbarf_if_buffer_read_only();

  if (NILP (code)) code = Vwin32_clipboard_coding_system;

  encode_code(Fcheck_code(code), &mccode);

  if (!OpenClipboard((HWND) 0)) return Fcons(Qnil, Qnil);
  if (!IsClipboardFormatAvailable(CF_TEXT)) {
    CloseClipboard();
    return Fcons(make_number (0), Qnil);
  }
  clipboard_handle = GetClipboardData(CF_TEXT);
  clipboard_data = GlobalLock(clipboard_handle);
  size = strlen(clipboard_data);

  move_gap (point);

  {
    register Lisp_Object temp;
    
    if (CODE_TYPE(&mccode) >= AUTOCONV) {
      int required_buf;	/* 93.2.10 by K.Handa */
      
      if (size == 0) {
	CODE_CNTL(&mccode) |= CC_END;
      }
      required_buf = ENCODE_BUF_SIZE (size, &mccode);
      if (GAP_SIZE < required_buf) /* 93.1.21 by K.Handa */
	make_gap (required_buf - GAP_SIZE);
      converted = encode(&mccode, clipboard_data,
			 &FETCH_CHAR (point + inserted - 1) + 1, size,
			 &found_code);
      if (converted) {
	XSET (temp, Lisp_Int, Z + converted);
	if (Z + converted != XINT (temp)) {
	  size = -2;		/* -2 means exceeding maximum buffer size */
	}
      }
    } else {
      if (GAP_SIZE < size)
	make_gap (size - GAP_SIZE);
      memcpy(&FETCH_CHAR (point+inserted-1) + 1, clipboard_data, size);
      converted = size;
    }
    GPT += converted;
    GAP_SIZE -= converted;
    ZV += converted;
    Z += converted;
    inserted += converted;
  }
      
  GlobalUnlock(clipboard_handle);
  CloseClipboard();

  if (inserted > 0)
    MODIFF++;
  record_insert (point, inserted);

  if (size == -2)
    error ("maximum buffer size exceeded");

  return Fcons (CODE_TYPE(&mccode) == AUTOCONV ? Qnil : found_code
		, Fcons (make_number (inserted), Qnil));
}

int RequiredBufSize(char *addr, int len, coding_type *mccode)
/* 95.5.16 by himi */
{
  if (!EQ (current_buffer->selective_display, Qt)
      && CODE_TYPE(mccode) <= ITNCODE)
    return len;
  else
    return DECODE_BUF_SIZE(len, mccode);
}

char *write_clipboard(char *clb_addr,
		      char *addr,
		      int len,
		      coding_type *mccode)
{
  char buf[WRITE_BUF_SIZE];
  int n = len, done, n_buf;

  if (!EQ (current_buffer->selective_display, Qt)
      && CODE_TYPE(mccode) <= ITNCODE)
    {
      memcpy(clb_addr, addr, len);
      return clb_addr + len;
    }
  else
    {
      unsigned int cntl_end =	/* 92.11.20, 92.12.20 by K.Handa */
	CODE_CNTL(mccode) & CC_END;

      CODE_CNTL(mccode) &= ~CC_END;
      while (n > 0) {
	n_buf = DECODE_BUF_SIZE(n, mccode);
	if (n_buf > WRITE_BUF_SIZE)
	  done = n * ((float) WRITE_BUF_SIZE / n_buf);
	else
	  done = n;
	if (done == n) CODE_CNTL(mccode) |= cntl_end;
	len = decode(mccode, addr, buf, done);
	memcpy(clb_addr, buf, len);
	clb_addr += len;
	n -= done, addr += done;
      }
    }
  return clb_addr;
}

DEFUN ("win32-set-region-to-clipboard", Fwin32_set_region_to_clipboard, 
  Swin32_set_region_to_clipboard, 2, 4,
  "rSet region to clipboard: ",
  "Write current region into clipboard.\n\
When called from a program, takes two arguments:\n\
START and, END .  START and END are buffer positions.\n\
Optional third argument APPEND if non-nil means\n\
  append to existing clipboard contents (if any CF_TEXT).\n\
Optional fourth argument CODE specifies the coding-system object\n\
 used in the clipboard, which defaults to win32-clipboard-coding-system.")
  (start, end, append, code)
     Lisp_Object start, end, append, code;
{
  int tem;
  char *share_data;
  HANDLE share_data_handle;
  coding_type mccode;
  int size = 0;

  if (NILP (code)) code = Vwin32_clipboard_coding_system;
  encode_code(Fcheck_code(code), &mccode);

  if (EQ (current_buffer->selective_display, Qt))
    CODE_CNTL(&mccode) |= CC_SELECTIVE;

  if (NILP (start))
    {
      XFASTINT (start) = BEG;
      XFASTINT (end) = Z;
    }
  else
    validate_region (&start, &end);

  if (XINT (start) == XINT (end)) return Qnil;

  if (XINT (start) < GPT)
    {
      register int end1 = XINT (end);
      CODE_CNTL(&mccode) |= CC_END;
      tem = XINT (start);
      size = RequiredBufSize(&FETCH_CHAR (tem),
			     min ((int)GPT, end1) - tem, &mccode);
    }
  
  if (XINT (end) > GPT)
    {
      tem = XINT (start);
      CODE_CNTL(&mccode) |= CC_END;
      tem = max (tem, GPT);
      size += RequiredBufSize(&FETCH_CHAR (tem),
				 XINT (end) - tem, &mccode);
    }

  if (!OpenClipboard(0))
    {
      message("ERROR: can't open clipboard!");
      return Qnil;
    }
  
  if (NILP (append) || !IsClipboardFormatAvailable(CF_TEXT))
    {
      EmptyClipboard();
      share_data_handle = GlobalAlloc(GMEM_SHARE, size + 1);
      share_data = (char *)GlobalLock(share_data_handle);
    }
  else
    {
      HANDLE org_clipboard_handle;
      char *org_clipboard_data;
      int org_size;
      org_clipboard_handle = GetClipboardData(CF_TEXT);
      if (!org_clipboard_handle)
	{
	  message("ERROR: can't get clipboard handle");
	  CloseClipboard();
	  return Qnil;
	}
      org_clipboard_data= (char *)GlobalLock(org_clipboard_handle);
      org_size = strlen(org_clipboard_data);
      share_data_handle = GlobalAlloc(GMEM_SHARE, size + org_size + 1);
      share_data = (char *)GlobalLock(share_data_handle);
      memcpy(share_data, org_clipboard_data, org_size);
      share_data += org_size;
      GlobalUnlock(org_clipboard_handle);
      EmptyClipboard();
    }
  
  if (XINT (start) < GPT)
    {
      register int end1 = XINT (end);
      if (end1 <= GPT)	/* 92.11.20, 92.12.20 by K.Handa */
	CODE_CNTL(&mccode) |= CC_END;
      tem = XINT (start);
      share_data =
	write_clipboard(share_data, &FETCH_CHAR (tem),
			min (GPT, end1) - tem, &mccode);
    }
  
  if (XINT (end) > GPT)
    {
      CODE_CNTL(&mccode) |= CC_END; /* 92.11.20 by K.Handa */
      tem = XINT (start);
      tem = max (tem, GPT);
      share_data =
	write_clipboard(share_data, &FETCH_CHAR (tem),
			XINT (end) - tem, &mccode);
    }

  *share_data = (char)NULL;
  GlobalUnlock(share_data_handle);
  SetClipboardData(CF_TEXT, share_data_handle);
  CloseClipboard();

  message ("Set data to clipboard");

  return Qnil;
}

/* From w32select.c */
/* 95.7.29 Modified by Shigeki Miura */

DEFUN ("win32-set-clipboard-data", Fwin32_set_clipboard_data,
       Swin32_set_clipboard_data, 1, 2, 0,
       "This sets the clipboard data to the given text.\n\
Optional argument CODE specifies the coding-system object\n\
used in the clipboard, which defaults to win32-clipboard-coding-system.")
     (string, code)
     Lisp_Object string, code;
{
  BOOL ok = TRUE;
  HANDLE htext;
  coding_type mccode;

  CHECK_STRING (string, 0);

  if (NILP (code))
    code = Vwin32_clipboard_coding_system;
  encode_code (Fcheck_code (code), &mccode);

  BLOCK_INPUT;

  if ((htext = GlobalAlloc (GMEM_MOVEABLE | GMEM_DDESHARE,
			    DECODE_BUF_SIZE (XSTRING (string)->size + 1,
					     &mccode))) == NULL)
    goto error;

  {
    unsigned char *lptext;
    int size;
    
    if ((lptext = (unsigned char *) GlobalLock (htext)) == NULL)
      goto error;
	    
    CODE_CNTL (&mccode) |= CC_END;
    size = decode (&mccode, XSTRING (string)->data, lptext,
		   XSTRING (string)->size);
    lptext[size] = '\0';

    GlobalUnlock (htext);
  }

  if (!OpenClipboard (NULL))
    goto error;

  ok = EmptyClipboard () && SetClipboardData (CF_TEXT, htext);

  CloseClipboard ();

  if (ok)
    goto done;

 error:
  ok = FALSE;
  if (htext)
    GlobalFree (htext);

 done:
  UNBLOCK_INPUT;

  return (ok ? string : Qnil);
}


DEFUN ("win32-get-clipboard-data", Fwin32_get_clipboard_data,
       Swin32_get_clipboard_data, 0, 1, 0,
       "This gets the clipboard data in text format.\n\
Optional argument CODE specifies the coding-system object\n\
used in the clipboard, which defaults to win32-clipboard-coding-system.")
     (code)
     Lisp_Object code;
{
  HANDLE htext;
  Lisp_Object ret = Qnil;
  coding_type mccode;
  Lisp_Object found_code = code;

  if (NILP (code))
    code = Vwin32_clipboard_coding_system;
  encode_code (Fcheck_code (code), &mccode);

  BLOCK_INPUT;

  if (!OpenClipboard (NULL))
    goto done;
    
  if ((htext = GetClipboardData (CF_TEXT)) == NULL)
    goto closeclip;
	    
  {
    unsigned char *lptext, *buf;
    int nbytes, size;
    
    if ((lptext = (unsigned char *) GlobalLock (htext)) == NULL)
      goto closeclip;
	    
    nbytes = strlen (lptext);
    CODE_CNTL (&mccode) |= CC_END;
    /* Modified by himi 95.12.20 */
    buf = (unsigned char*)xmalloc (ENCODE_BUF_SIZE (nbytes, &mccode));
    size = encode (&mccode, lptext, buf, nbytes, &found_code);
    ret = make_string (buf, size);
    xfree (buf);
    /* end */

    GlobalUnlock (htext);
  }

 closeclip:
  CloseClipboard ();

 done:
  UNBLOCK_INPUT;

  return ret;
}

DEFUN ("win32-clipboard-data-exist-p", Fwin32_clipboard_data_exist_p,
       Swin32_clipboard_data_exist_p, 0, 1, 0,
       "This inspects whether clipboard have any data.\n\
t means it has something. nil means it has nothing. Optional argument DUMMY\n\
is not used. This is to make compatiblity with x-selection-exists-p.")
     (dummy)
     Lisp_Object dummy;
{
  return IsClipboardFormatAvailable(CF_TEXT)? Qt:Qnil;
}

syms_of_win32clipboard ()
{
  DEFVAR_LISP ("win32-clipboard-coding-system",
	       &Vwin32_clipboard_coding_system,
	       "coding system in clipboard");

  defsubr (&Swin32_set_region_to_clipboard);
  defsubr (&Swin32_clipboard_yank);

  defsubr (&Swin32_set_clipboard_data);
  defsubr (&Swin32_get_clipboard_data);
  defsubr (&Swin32_clipboard_data_exist_p);
}

#endif /* MULE */
