/* Functions for the X window system.
   Copyright (C) 1989, 1992, 1993, 1994 Free Software Foundation.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Written by Kevin Gallo */


#include <signal.h>
#include <config.h>
#include <stdio.h>

#include "lisp.h"
#include "frame.h"
#ifdef WIN32_FONTSET
#include "win32fontset.h"
#endif
#include "win32term.h"
#include "win32reg.h"
#include "window.h"
#include "buffer.h"
#include "dispextern.h"
#include "keyboard.h"
#include "blockinput.h"

#ifdef MULE
#include "coding.h"
#endif

extern void abort ();

extern void free_frame_menubar ();

extern HINSTANCE hinst ;
extern HINSTANCE hprevinst ;
extern LPSTR lpCmdLine ;
extern int nCmdShow ;
    
/* The class of this X application.  */
#ifdef MULE
#define EMACS_CLASS "MULE"
#else
#define EMACS_CLASS "Emacs"
#endif

/* Title name and application name for X stuff. */
extern char *win32_id_name;

/* The background and shape of the mouse pointer, and shape when not
   over text or in the modeline. */
Lisp_Object Vwin32_pointer_shape, Vwin32_nontext_pointer_shape, Vwin32_mode_pointer_shape;
/* The shape when over mouse-sensitive text.  */
Lisp_Object Vwin32_sensitive_text_pointer_shape;

/* Color of chars displayed in cursor box. */
Lisp_Object Vwin32_cursor_fore_pixel;

/* The colormap for converting color names to RGB values */
Lisp_Object Vwin32_color_map;

/* Height of this X screen in pixels. */
int win32_screen_height;

/* Width of this X screen in pixels. */
int win32_screen_width;

/* Height of this X screen in inches. */
int win32_screen_height_in;

/* Width of this X screen in inches. */
int win32_screen_width_in;

/* Number of planes for this screen. */
int win32_screen_planes;

/* Non nil if no window manager is in use. */
Lisp_Object Vwin32_no_window_manager;

/* `t' if a mouse button is depressed. */

Lisp_Object Vmouse_depressed;

extern unsigned int win32_mouse_x, win32_mouse_y, win32_mouse_grabbed;

/* The last 23 bits of the timestamp of the last mouse button event. */
Time mouse_timestamp;

HANDLE hMutex = NULL ;

#if 0
ACCEL undo_accel = {FCONTROL, '/', IDM_UNDO};
HACCEL UndoAccel;
#endif

/* Evaluate this expression to rebuild the section of syms_of_xfns
   that initializes and staticpros the symbols declared below.  Note
   that Emacs 18 has a bug that keeps C-x C-e from being able to
   evaluate this expression.

(progn
  ;; Accumulate a list of the symbols we want to initialize from the
  ;; declarations at the top of the file.
  (goto-char (point-min))
  (search-forward "/\*&&& symbols declared here &&&*\/\n")
  (let (symbol-list)
    (while (looking-at "Lisp_Object \\(Q[a-z_]+\\)")
      (setq symbol-list
	    (cons (buffer-substring (match-beginning 1) (match-end 1))
		  symbol-list))
      (forward-line 1))
    (setq symbol-list (nreverse symbol-list))
    ;; Delete the section of syms_of_... where we initialize the symbols.
    (search-forward "\n  /\*&&& init symbols here &&&*\/\n")
    (let ((start (point)))
      (while (looking-at "^  Q")
	(forward-line 2))
      (kill-region start (point)))
    ;; Write a new symbol initialization section.
    (while symbol-list
      (insert (format "  %s = intern (\"" (car symbol-list)))
      (let ((start (point)))
	(insert (substring (car symbol-list) 1))
	(subst-char-in-region start (point) ?_ ?-))
      (insert (format "\");\n  staticpro (&%s);\n" (car symbol-list)))
      (setq symbol-list (cdr symbol-list)))))

  */        

/*&&& symbols declared here &&&*/
Lisp_Object Qauto_raise;
Lisp_Object Qauto_lower;
Lisp_Object Qbackground_color;
Lisp_Object Qbar;
Lisp_Object Qborder_color;
Lisp_Object Qborder_width;
Lisp_Object Qbox;
Lisp_Object Qcursor_color;
Lisp_Object Qcursor_type;
Lisp_Object Qfont;
Lisp_Object Qforeground_color;
Lisp_Object Qgeometry;
Lisp_Object Qicon_left;
Lisp_Object Qicon_top;
Lisp_Object Qicon_type;
Lisp_Object Qinternal_border_width;
Lisp_Object Qleft;
Lisp_Object Qmouse_color;
Lisp_Object Qnone;
Lisp_Object Qparent_id;
Lisp_Object Qsuppress_icon;
Lisp_Object Qtop;
Lisp_Object Qundefined_color;
Lisp_Object Qvertical_scroll_bars;
Lisp_Object Qvisibility;
Lisp_Object Qwindow_id;
Lisp_Object Qwin32_frame_parameter;
Lisp_Object Qwin32_resource_name;
Lisp_Object Quser_position;
Lisp_Object Quser_size;
#ifdef MULE			/* 94.6.17 by K.Handa */
Lisp_Object Qline_space;
#endif  /* MULE */
#ifdef WIN32_FONTSET           /* 96.2.1 by himi */
Lisp_Object Qfontset;
Lisp_Object Qfont_property;
#endif

/* The below are defined in frame.c. */
extern Lisp_Object Qheight, Qminibuffer, Qname, Qonly, Qwidth;
extern Lisp_Object Qunsplittable, Qmenu_bar_lines;

extern Lisp_Object Vwindow_system_version;

#if 0
DEFUN ("win32-rgb", Fwin32_rgb, Swin32_rgb, 3, 3, 0,
       "Convert RGB numbers to a windows color reference.")
    (red,green,blue)
    Lisp_Object red,green,blue ;
{
    Lisp_Object rgb ;

    CHECK_NUMBER (red, 0);
    CHECK_NUMBER (green, 0);
    CHECK_NUMBER (blue, 0);

    XSET (rgb, Lisp_Int, RGB(XUINT(red),XUINT(green),XUINT(blue)));

    return (rgb) ;
}


#else
/* The default colors for the win32 color map */
typedef struct ColorMap_t {
    char * name ;
    COLORREF colorref ;
} ColorMap_t ;

ColorMap_t win32_color_map[] = {
#include "win32rgb.h"
} ;

DEFUN ("win32-default-color-map", Fwin32_default_color_map, Swin32_default_color_map,
       0, 0, 0, "Return the default color map.")
  ()
{
    int i ;
    ColorMap_t * pc = win32_color_map ;
    Lisp_Object cmap ;

    BLOCK_INPUT;

    cmap = Qnil ;

    for (i = 0; i < sizeof (win32_color_map) / sizeof (win32_color_map[0]); pc++,i++)
	cmap = Fcons(Fcons (build_string(pc->name),
			    make_number (pc->colorref)),
		     cmap) ;

    UNBLOCK_INPUT ;

    return (cmap) ;
}
#endif

Lisp_Object win32_to_x_color(rgb)
    Lisp_Object rgb ;
{
    Lisp_Object color ;

    CHECK_NUMBER (rgb, 0);

    BLOCK_INPUT;

    color = Frassq(rgb,Vwin32_color_map) ;

    UNBLOCK_INPUT ;

    if (!NILP(color))
	return (Fcar(color)) ;
    else
	return Qnil ;
}

static
int get_hex1(int numstr)
{
  if (isalpha(numstr)) {
    numstr = toupper(numstr);
    if (!(numstr >= 'A' && numstr <= 'F'))
      return -1;
    numstr -= ('A' - 10);
    return numstr;
  }
  if (isdigit(numstr)) {
    numstr -= '0';
    return numstr;
  }
  return -1;
}

static
int color_radix_change(TCHAR *colstr, int size)
{
  int i;
  int ret, tmp;

  ret = 0;
  for(i = 0;i < size;i++) {
    ret <<= 4;
    tmp = get_hex1(*colstr);
    if (tmp == -1) 
      return -1;
    ret += tmp;
    colstr++;
  }
  return ret;
}

static
Lisp_Object x_old_rgb_names(TCHAR *colstr, int len)
{
  Lisp_Object cdef;
  int red, green, blue;

  switch(len) {
  case 4:			/* #RGB */
    red = color_radix_change(colstr+1, 1) << 4;
    green = color_radix_change(colstr+2, 1) << 4;
    blue = color_radix_change(colstr+3, 1) << 4;
    break;

  case 7:			/* #RRGGBB */
    red = color_radix_change(colstr+1, 2);
    green = color_radix_change(colstr+3, 2);
    blue = color_radix_change(colstr+5, 2);
    break;
  case 10:			/* #RRRGGGBBB */
    red = color_radix_change(colstr+1, 3) >> 4;
    green = color_radix_change(colstr+4, 3) >> 4;
    blue = color_radix_change(colstr+7, 3) >> 4;
    break;
  case 13:			/* #RRRRGGGGBBBB */
    red = color_radix_change(colstr+1, 4) >> 8;
    green = color_radix_change(colstr+5, 4) >> 8;
    blue = color_radix_change(colstr+9, 4) >> 8;
    break;

  default :
    return Qnil;
  }
  if ((red == -1) || (green == -1) || (blue == -1))
    return Qnil;
  return RGB(red, green, blue);

}

static
Lisp_Object x_rgb_names(TCHAR *colstr, int len)
{
  TCHAR *colstrend;
  Lisp_Object ret;
  int red, green, blue;

  colstr += 4;
  len -= 4;
  colstrend = memchr(colstr, '/', len);
  if (!colstrend)
    return Qnil;
  red = color_radix_change(colstr, (int) (colstrend - colstr));
  if (red == -1) return Qnil;

  len -= (int)(colstrend - colstr + 1);
  colstr = colstrend + 1;
  colstrend = memchr(colstr, '/', len);
  if (!colstrend)
    return Qnil;
  green = color_radix_change(colstr, (int) (colstrend - colstr));
  if (green == -1) return Qnil;

  len -= (int)(colstrend - colstr + 1);
  colstr = colstrend + 1;
  blue = color_radix_change(colstr, len);
  if (blue == -1) return Qnil;

  red >>= 8;
  green >>= 8;
  blue >>= 8;

  XSET(ret, Lisp_Int, RGB(red, green, blue));
  return ret;
}

Lisp_Object x_to_win32_color(colorname)
    Lisp_Object colorname ;
{
    TCHAR *colstr;
    int colstrlen;
    register Lisp_Object tail,ret = Qnil ;

    CHECK_STRING (colorname,0) ;

    BLOCK_INPUT;

    colstr = XSTRING(colorname)->data;
    colstrlen = XSTRING(colorname)->size;
    if (*colstr == '#') {
      return x_old_rgb_names(colstr, colstrlen);
    }
    
    if (colstrlen > 4) {
      if (memcmp(colstr, "rgb:", 4) == 0) {
	return x_rgb_names(colstr, colstrlen);
      }
#if 0
      if (memcmp(colstr, "rgbi:", 5) == 0) {
	return x_rgbi_names(arg, colstr, colstrlen);
      }
#endif
    }
    

    for (tail = Vwin32_color_map; !NILP (tail); tail = Fcdr (tail))
    {
	register Lisp_Object elt, tem;

	elt = Fcar (tail);
	if (!CONSP (elt)) continue;

	tem = Fcar (elt);

/* 96.10.12 modified by himi */
	if ((XSTRING(colorname) -> size) == (XSTRING(tem) -> size)){
	  if (memicmp(XSTRING(colorname) -> data, XSTRING(tem) -> data,
		      XSTRING(colorname) -> size) == 0) {
	    ret = Fcdr (elt) ;
	    break ;
	  }
	}

	QUIT ;
    }

    UNBLOCK_INPUT ;

    return ret ;
}

/* Return the Emacs frame-object corresponding to an X window.
   It could be the frame's main window or an icon window.  */

/* This function can be called during GC, so use XGCTYPE.  */

struct frame *
win32_window_to_frame (wdesc)
    HWND wdesc;
{
  Lisp_Object tail, frame;
  struct frame *f;

  if (wdesc == 0) return (0) ;

  for (tail = Vframe_list; XGCTYPE (tail) == Lisp_Cons;
       tail = XCONS (tail)->cdr)
    {
      frame = XCONS (tail)->car;
      if (XGCTYPE (frame) != Lisp_Frame)
        continue;
      f = XFRAME (frame);
      if (f->display.nothing == 1) 
	return 0;
      if (FRAME_WIN32_WINDOW (f) == wdesc
          || f->display.win32->icon_desc == wdesc)
        return f;
    }
  return 0;
}

/* Connect the frame-parameter names for X frames
   to the ways of passing the parameter values to the window system.

   The name of a parameter, as a Lisp symbol,
   has an `x-frame-parameter' property which is an integer in Lisp
   but can be interpreted as an `enum win32_frame_parm' in C.  */

enum win32_frame_parm
{
  WIN32_PARM_FOREGROUND_COLOR,
  WIN32_PARM_BACKGROUND_COLOR,
  WIN32_PARM_MOUSE_COLOR,
  WIN32_PARM_CURSOR_COLOR,
  WIN32_PARM_BORDER_COLOR,
  WIN32_PARM_ICON_TYPE,
  WIN32_PARM_FONT,
#ifdef WIN32_FONTSET
  WIN32_PARM_FONTSET,
  WIN32_PARM_FONT_PROPERTY,
#endif
#ifdef MULE
  WIN32_PARM_LINE_SPACE,
#endif  /* MULE */
  WIN32_PARM_BORDER_WIDTH,
  WIN32_PARM_INTERNAL_BORDER_WIDTH,
  WIN32_PARM_NAME,
  WIN32_PARM_AUTORAISE,
  WIN32_PARM_AUTOLOWER,
  WIN32_PARM_VERT_SCROLL_BAR,
  WIN32_PARM_VISIBILITY,
  WIN32_PARM_MENU_BAR_LINES
};


struct win32_frame_parm_table
{
  char *name;
  void (*setter)( /* struct frame *frame, Lisp_Object val, oldval */ );
};

void win32_set_foreground_color ();
void win32_set_background_color ();
void win32_set_mouse_color ();
void win32_set_cursor_color ();
void win32_set_border_color ();
void win32_set_cursor_type ();
void win32_set_font ();
#ifdef MULE
void win32_set_line_space ();
#endif  /* MULE */
#ifdef WIN32_FONTSET
void win32_set_fontset ();
void win32_set_font_property ();
#endif
void win32_set_border_width ();
void win32_set_internal_border_width ();
void win32_explicitly_set_name ();
void win32_set_autoraise ();
void win32_set_autolower ();
void win32_set_vertical_scroll_bars ();
void win32_set_visibility ();
void win32_set_menu_bar_lines ();

static struct win32_frame_parm_table win32_frame_parms[] =
{
  "foreground-color", win32_set_foreground_color,
  "background-color", win32_set_background_color,
  "mouse-color", win32_set_mouse_color,
  "cursor-color", win32_set_cursor_color,
  "border-color", win32_set_border_color,
  "cursor-type", win32_set_cursor_type,
  "font", win32_set_font,
#ifdef WIN32_FONTSET
  "fontset", win32_set_fontset,
  "font_property", win32_set_font_property,
#endif
#ifdef MULE
  "line-space", win32_set_line_space,
#endif  /* MULE */
  "border-width", win32_set_border_width,
  "internal-border-width", win32_set_internal_border_width,
  "name", win32_explicitly_set_name,
  "auto-raise", win32_set_autoraise,
  "auto-lower", win32_set_autolower,
  "vertical-scroll-bars", win32_set_vertical_scroll_bars,
  "visibility", win32_set_visibility,
  "menu-bar-lines", win32_set_menu_bar_lines,
};

/* Attach the `x-frame-parameter' properties to
   the Lisp symbol names of parameters relevant to X.  */

init_win32_parm_symbols ()
{
  int i;

  for (i = 0; i < sizeof (win32_frame_parms) / sizeof (win32_frame_parms[0]); i++)
    Fput (intern (win32_frame_parms[i].name), Qwin32_frame_parameter,
	  make_number (i));
}

/* Change the parameters of FRAME as specified by ALIST.
   If a parameter is not specially recognized, do nothing;
   otherwise call the `win32_set_...' function for that parameter.  */

void
win32_set_frame_parameters (f, alist)
     FRAME_PTR f;
     Lisp_Object alist;
{
  Lisp_Object tail;

  /* If both of these parameters are present, it's more efficient to
     set them both at once.  So we wait until we've looked at the
     entire list before we set them.  */
  Lisp_Object width, height;

  /* Same here.  */
  Lisp_Object left, top;

  /* Record in these vectors all the parms specified.  */
  Lisp_Object *parms;
  Lisp_Object *values;
  int i;
  
  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    i++;

  parms = (Lisp_Object *) alloca (i * sizeof (Lisp_Object));
  values = (Lisp_Object *) alloca (i * sizeof (Lisp_Object));

  /* Extract parm names and values into those vectors.  */

  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    {
      Lisp_Object elt, prop, val;

      elt = Fcar (tail);
      parms[i] = Fcar (elt);
      values[i] = Fcdr (elt);
      i++;
    }

  width = height = top = left = Qunbound;

  /* Now process them in reverse of specified order.  */
  for (i--; i >= 0; i--)
    {
      Lisp_Object prop, val;

      prop = parms[i];
      val = values[i];

      if (EQ (prop, Qwidth))
	width = val;
      else if (EQ (prop, Qheight))
	height = val;
      else if (EQ (prop, Qtop))
	top = val;
      else if (EQ (prop, Qleft))
	left = val;
      else
	{
	  register Lisp_Object param_index, old_value;

	  param_index = Fget (prop, Qwin32_frame_parameter);
	  old_value = get_frame_param (f, prop);
	  store_frame_param (f, prop, val);
	  if (XTYPE (param_index) == Lisp_Int
	      && XINT (param_index) >= 0
	      && (XINT (param_index)
		  < sizeof (win32_frame_parms)/sizeof (win32_frame_parms[0])))
	    (*win32_frame_parms[XINT (param_index)].setter)(f, val, old_value);
	}
    }

  /* Don't die if just one of these was set.  */
  if (EQ (left, Qunbound))
    XSET (left, Lisp_Int, f->display.win32->left_pos);
  if (EQ (top, Qunbound))
    XSET (top, Lisp_Int, f->display.win32->top_pos);

  /* Don't die if just one of these was set.  */
  if (EQ (width, Qunbound))
    XSET (width, Lisp_Int, FRAME_WIDTH (f));
  if (EQ (height, Qunbound))
    XSET (height, Lisp_Int, FRAME_HEIGHT (f));

  /* Don't set these parameters these unless they've been explicitly
     specified.  The window might be mapped or resized while we're in
     this function, and we don't want to override that unless the lisp
     code has asked for it.

     Don't set these parameters unless they actually differ from the
     window's current parameters; the window may not actually exist
     yet.  */
  {
    Lisp_Object frame;

    check_frame_size (f, &height, &width);

    XSET (frame, Lisp_Frame, f);

    if ((NUMBERP (width) && XINT (width) != FRAME_WIDTH (f))
	|| (NUMBERP (height) && XINT (height) != FRAME_HEIGHT (f)))
      Fset_frame_size (frame, width, height);

    if ((!NILP (left) || !NILP (top))
	&& ! (NUMBERP (left) && XINT (left) == f->display.win32->left_pos
	      && NUMBERP (top) && XINT (top) == f->display.win32->top_pos))
      {
	int leftpos = (NUMBERP (left) ? XINT (left) : 0);
	int toppos = (NUMBERP (top) ? XINT (top) : 0);

	/* Store the numeric value of the position.  */
	f->display.win32->top_pos = toppos;
	f->display.win32->left_pos = leftpos;

	/* Record the signs.  */
#if 0
	f->display.win32->size_hint_flags &= ~ (XNegative | YNegative);
	if (EQ (left, Qminus) || (NUMBERP (left) && XINT (left) < 0))
	  f->display.win32->size_hint_flags |= XNegative;
	if (EQ (top, Qminus) || (NUMBERP (top) && XINT (top) < 0))
	  f->display.win32->size_hint_flags |= YNegative;
	f->display.win32->win_gravity = NorthWestGravity;
#endif

	/* Actually set that position, and convert to absolute.  */
	win32_set_offset (f, leftpos, toppos, 0);
      }
  }
}

#ifndef WIN32_FONTSET
/* The font conversion stuff between x and win32 */

/* X font string is as follows (from faces.el)
 * (let ((- 		"[-?]")
 *      (foundry	"[^-]+")
 *      (family 	"[^-]+")
 *      (weight		"\\(bold\\|demibold\\|medium\\)")		; 1
 *      (weight\?	"\\([^-]*\\)")					; 1
 *      (slant		"\\([ior]\\)")					; 2
 *      (slant\?	"\\([^-]?\\)")					; 2
 *      (swidth		"\\([^-]*\\)")					; 3
 *      (adstyle	"[^-]*")					; 4
 *      (pixelsize	"[0-9]+")
 *      (pointsize	"[0-9][0-9]+")
 *      (resx		"[0-9][0-9]+")
 *      (resy		"[0-9][0-9]+")
 *      (spacing	"[cmp?*]")
 *      (avgwidth	"[0-9]+")
 *      (registry	"[^-]+")
 *      (encoding	"[^-]+")
 *      )
 *  (setq x-font-regexp
 *	(concat "\\`\\*?[-?*]"
 *		foundry - family - weight\? - slant\? - swidth - adstyle -
 *		pixelsize - pointsize - resx - resy - spacing - registry -
 *		encoding "[-?*]\\*?\\'"
 *		))
 *  (setq x-font-regexp-head
 *	(concat "\\`[-?*]" foundry - family - weight\? - slant\?
 *		"\\([-*?]\\|\\'\\)"))
 *  (setq x-font-regexp-slant (concat - slant -))
 *  (setq x-font-regexp-weight (concat - weight -))
 * nil)	    
 */
    
#define FONT_START       "[-?]"
#define FONT_FOUNDRY     "[^-]+"
#define FONT_FAMILY      "\\([^-]+\\)"                      /* 1 */
#define FONT_WEIGHT      "\\(bold\\|demibold\\|medium\\)"   /* 2 */
#define FONT_WEIGHT_Q    "\\([^-]*\\)"                      /* 2 */
#define FONT_SLANT       "\\([ior]\\)"                      /* 3 */
#define FONT_SLANT_Q     "\\([^-]?\\)"                      /* 3 */
#define FONT_SWIDTH      "\\([^-]*\\)"                      /* 4 */
#define FONT_ADSTYLE     "[^-]*"
#define FONT_PIXELSIZE   "[^-]*"
#define FONT_POINTSIZE   "\\([0-9][0-9]+\\|\\*\\)"          /* 5 */
#define FONT_RESX        "[0-9][0-9]+"
#define FONT_RESY        "[0-9][0-9]+"
#define FONT_SPACING     "[cmp?*]"
#define FONT_AVGWIDTH    "[0-9]+"
#define FONT_REGISTRY    "[^-]+"
#define FONT_ENCODING    "[^-]+"

#define FONT_REGEXP      ("\\`\\*?[-?*]"     \
			  FONT_FOUNDRY   "-" \
			  FONT_FAMILY    "-" \
			  FONT_WEIGHT_Q  "-" \
			  FONT_SLANT_Q   "-" \
			  FONT_SWIDTH    "-" \
			  FONT_ADSTYLE   "-" \
			  FONT_PIXELSIZE "-" \
			  FONT_POINTSIZE "-" \
			  "[-?*]\\|\\'")

#define FONT_REGEXP_HEAD ("\\`[-?*]"        \
			  FONT_FOUNDRY  "-" \
			  FONT_FAMILY   "-" \
			  FONT_WEIGHT_Q "-" \
			  FONT_SLANT_Q      \
			  "\\([-*?]\\|\\'\\)")

#define FONT_REGEXP_SLANT  "-" FONT_SLANT  "-"
#define FONT_REGEXP_WEIGHT "-" FONT_WEIGHT "-"

LONG x_to_win32_weight (lpw)
    char * lpw ;
{
    if (!lpw) return (FW_DONTCARE) ;

    if (stricmp (lpw,"bold") == 0)
	return (FW_BOLD) ;
    else if (stricmp (lpw,"demibold") == 0)
	return (FW_SEMIBOLD) ;
    else if (stricmp (lpw,"medium") == 0)
	return (FW_MEDIUM) ;
    else if (stricmp (lpw,"normal") == 0)
	return (FW_NORMAL) ;
    else
	return (FW_DONTCARE) ;
}

char * win32_to_x_weight (fnweight)
    int fnweight ;
{
    if (fnweight >= FW_BOLD) 
	return ("bold") ;
    else if (fnweight >= FW_SEMIBOLD) 
	return ("demibold") ;
    else if (fnweight >= FW_MEDIUM) 
	return ("medium") ;
    else  
	return ("normal") ;
}

BOOL win32_to_x_font(lplogfont,lpxstr,len)
    LOGFONT * lplogfont ;
    char * lpxstr ;
    int len ;
{
    if (!lpxstr) return (FALSE) ;

    if (lplogfont) {
	_snprintf (lpxstr,len - 1,
		   "-*-%s-%s-%c-%s-%s-*-%d-*-*-%c-%d-*-*-",
		   lplogfont->lfFaceName,
		   win32_to_x_weight(lplogfont->lfWeight),
		   lplogfont->lfItalic?'i':'r',
		   "*","*",
		   (lplogfont->lfHeight * 1440) / win32_screen_height_in,
		   ((lplogfont->lfPitchAndFamily & 0x3) == VARIABLE_PITCH)?'p':'c',
		   (lplogfont->lfWidth * 1440) / win32_screen_width_in) ;
    } else {
	strncpy (lpxstr,"-*-*-*-*-*-*-*-*-*-*-*-*-*-*-",len - 1) ;
    }

    lpxstr[len - 1] = 0 ; /* just to be sure */
    return (TRUE) ;
}

BOOL x_to_win32_font(lpxstr,lplogfont)
    char * lpxstr ;
    LOGFONT * lplogfont ;
{
    if (!lplogfont) return (FALSE) ;

    memset(lplogfont,0,sizeof(*lplogfont)) ;

#ifdef MULE
    lplogfont->lfCharSet = SHIFTJIS_CHARSET;
#else
    lplogfont->lfCharSet = OEM_CHARSET ;
#endif
    lplogfont->lfOutPrecision = OUT_DEFAULT_PRECIS ;
    lplogfont->lfClipPrecision = CLIP_DEFAULT_PRECIS ;
    lplogfont->lfQuality = DEFAULT_QUALITY ;
    
    if (lpxstr && *lpxstr == '-') lpxstr++ ;

    {
	int fields ;
	char name[50], weight[20], slant, pitch , height[10], width[10] ;
	
	fields = (lpxstr
		  ?sscanf(lpxstr,
			  "%*[^-]-%[^-]-%[^-]-%c-%*[^-]-%*[^-]-%*[^-]-%[^-]-%*[^-]-%*[^-]-%c-%[^-]",
			  name,weight,&slant,height,&pitch,width)
		  :0) ;

	if (fields == EOF) return (FALSE) ;

	if (fields > 0 && name[0] != '*') {
	    strncpy (lplogfont->lfFaceName,name,LF_FACESIZE) ;
	} else {
	    lplogfont->lfFaceName[0] = 0 ;
	}

	fields-- ;

	lplogfont->lfWeight = 0; /* x_to_win32_weight((fields > 0?weight:"")) ;*/

	fields-- ;

	lplogfont->lfItalic = (fields > 0 && slant == 'i') ;

	fields-- ;

	if (fields > 0 && height[0] != '*')
	    lplogfont->lfHeight = (atoi(height) * win32_screen_height_in) / 1440 ;

	fields-- ;

	lplogfont->lfPitchAndFamily = (fields > 0 && pitch == 'p')?VARIABLE_PITCH:FIXED_PITCH ;

	fields-- ;

	if (fields > 0 && width[0] != '*')
	    lplogfont->lfWidth = (atoi(width) * win32_screen_width_in) / 1440 ;
    }

    return (TRUE) ;
}

BOOL win32_font_match (lpszfont1,lpszfont2)
    char * lpszfont1 ;
    char * lpszfont2 ;
{
    char * s1 = lpszfont1, *e1 ;
    char * s2 = lpszfont2, *e2 ;

    if (s1 == NULL || s2 == NULL) return (FALSE) ;

    if (*s1 == '-') s1++ ;
    if (*s2 == '-') s2++ ;

    while (1) {
	int len1, len2 ;

	e1 = strchr (s1,'-') ;
	e2 = strchr (s2,'-') ;

	if (e1 == NULL || e2 == NULL) return (TRUE) ;

	len1 = e1 - s1 ;
	len2 = e2 - s2 ;

	if (*s1 != '*' && *s2 != '*' && 
	    (len1 != len2 || strnicmp(s1,s2,len1) != 0))
	    return (FALSE) ;

	s1 = e1 + 1 ;
	s2 = e2 + 1 ;
    }
}
#endif /* not WIN32_FONTSET */

/* Store the positions of frame F into XPTR and YPTR.
   These are the positions of the containing window manager window,
   not Emacs's own window.  */

void
win32_real_positions (f, xptr, yptr)
     FRAME_PTR f;
     int *xptr, *yptr;
{
  POINT pt ;

  BLOCK_INPUT;

  {
      RECT rect ;
      
      GetClientRect(FRAME_WIN32_WINDOW(f),&rect) ;
      AdjustWindowRect(&rect,f->display.win32->dwStyle,FRAME_EXTERNAL_MENU_BAR(f)) ;
      
      pt.x = rect.left ;
      pt.y = rect.top ;
  }

  ClientToScreen (FRAME_WIN32_WINDOW(f),&pt) ;

  UNBLOCK_INPUT;

  *xptr = pt.x ;
  *yptr = pt.y ;
}

/* Insert a description of internally-recorded parameters of frame X
   into the parameter alist *ALISTPTR that is to be given to the user.
   Only parameters that are specific to the X window system
   and whose values are not correctly recorded in the frame's
   param_alist need to be considered here.  */

win32_report_frame_params (f, alistptr)
     struct frame *f;
     Lisp_Object *alistptr;
{
  char buf[16];

  store_in_alist (alistptr, Qleft, make_number (f->display.win32->left_pos));
  store_in_alist (alistptr, Qtop, make_number (f->display.win32->top_pos));
  store_in_alist (alistptr, Qborder_width,
       	   make_number (f->display.win32->border_width));
  store_in_alist (alistptr, Qinternal_border_width,
       	   make_number (f->display.win32->internal_border_width));
  sprintf (buf, "%d", FRAME_WIN32_WINDOW (f));
  store_in_alist (alistptr, Qwindow_id,
       	   build_string (buf));
  FRAME_SAMPLE_VISIBILITY (f);
  store_in_alist (alistptr, Qvisibility,
		  (FRAME_VISIBLE_P (f) ? Qt
		   : FRAME_ICONIFIED_P (f) ? Qicon : Qnil));
}

/* Decide if color named COLOR is valid for the display associated with
   the selected frame; if so, return the rgb values in COLOR_DEF.
   If ALLOC is nonzero, allocate a new colormap cell.  */

int
defined_color (color, color_def)
     Lisp_Object color;
     Color *color_def;
{
  register Lisp_Object tem ;

  tem = x_to_win32_color (color) ;

  if (! NILP (tem)) {
    /* Modified by himi */
      *color_def = XUINT(tem);
      return 1 ;
  } else {
      return 0 ;
  }
}

/* Given a string ARG naming a color, compute a pixel value from it
   suitable for screen F.
   If F is not a color screen, return DEF (default) regardless of what
   ARG says.  */
/* 96.10.1 modified by himi */
int
win32_decode_color (arg, def)
     Lisp_Object arg;
     int def;
{
  Color cdef;
  TCHAR *colstr;

  CHECK_STRING (arg, 0);

  colstr = XSTRING(arg) -> data;

  if (strcmp (colstr, "black") == 0)
    return BLACK_PIX_DEFAULT;
  else if (strcmp (colstr, "white") == 0)
    return WHITE_PIX_DEFAULT;

  if (win32_screen_planes == 1)
    return def;

  if (defined_color (arg, &cdef, 1))
    return cdef;
  else
    Fsignal (Qundefined_color, Fcons (arg, Qnil));
}

/* Functions called only from `win32_set_frame_param'
   to set individual parameters.

   If FRAME_WIN32_WINDOW (f) is 0,
   the frame is being created and its X-window does not exist yet.
   In that case, just record the parameter's new value
   in the standard place; do not attempt to change the window.  */

void
win32_set_foreground_color (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  f->display.win32->foreground_pixel = win32_decode_color (arg, BLACK_PIX_DEFAULT);
  if (FRAME_WIN32_WINDOW (f) != 0)
    {
      recompute_basic_faces (f);
      if (FRAME_VISIBLE_P (f))
        redraw_frame (f);
    }
}

void
win32_set_background_color (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  f->display.win32->background_pixel = win32_decode_color (arg, WHITE_PIX_DEFAULT);

  if (FRAME_WIN32_WINDOW (f) != 0)
    {
      recompute_basic_faces (f);

      if (FRAME_VISIBLE_P (f))
        redraw_frame (f);
    }
}

void
win32_set_mouse_color (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
#if 0
  Cursor cursor, nontext_cursor, mode_cursor, cross_cursor;
#endif
  int mask_color;

  if (!EQ (Qnil, arg))
    f->display.win32->mouse_pixel = win32_decode_color (arg, BLACK_PIX_DEFAULT);
  mask_color = f->display.win32->background_pixel;
				/* No invisible pointers. */
  if (mask_color == f->display.win32->mouse_pixel
	&& mask_color == f->display.win32->background_pixel)
    f->display.win32->mouse_pixel = f->display.win32->foreground_pixel;

  BLOCK_INPUT;
#if 0

  /* It's not okay to crash if the user selects a screwy cursor.  */
  win32_catch_errors ();

  if (!EQ (Qnil, Vwin32_pointer_shape))
    {
      CHECK_NUMBER (Vwin32_pointer_shape, 0);
      cursor = XCreateFontCursor (x_current_display, XINT (Vwin32_pointer_shape));
    }
  else
    cursor = XCreateFontCursor (dsp, XC_xterm);
  win32_check_errors ("bad text pointer cursor: %s");

  if (!EQ (Qnil, Vwin32_nontext_pointer_shape))
    {
      CHECK_NUMBER (Vwin32_nontext_pointer_shape, 0);
      nontext_cursor = XCreateFontCursor (dsp,
					  XINT (Vwin32_nontext_pointer_shape));
    }
  else
    nontext_cursor = XCreateFontCursor (dsp, XC_left_ptr);
  win32_check_errors ("bad nontext pointer cursor: %s");

  if (!EQ (Qnil, Vwin32_mode_pointer_shape))
    {
      CHECK_NUMBER (Vwin32_mode_pointer_shape, 0);
      mode_cursor = XCreateFontCursor (dsp,
					  XINT (Vwin32_mode_pointer_shape));
    }
  else
    mode_cursor = XCreateFontCursor (dsp, XC_xterm);
  win32_check_errors ("bad modeline pointer cursor: %s");

  if (!EQ (Qnil, Vwin32_sensitive_text_pointer_shape))
    {
      CHECK_NUMBER (Vwin32_sensitive_text_pointer_shape, 0);
      cross_cursor
	= XCreateFontCursor (dsp,
			     XINT (Vwin32_sensitive_text_pointer_shape));
    }
  else
    cross_cursor = XCreateFontCursor (dsp, XC_crosshair);

  /* Check and report errors with the above calls.  */
  win32_check_errors ("can't set cursor shape: %s");
  win32_uncatch_errors ();
  {
    XColor fore_color, back_color;

    fore_color.pixel = f->display.win32->mouse_pixel;
    back_color.pixel = mask_color;
    XQueryColor (dsp,
		 DefaultColormap (dsp,
				  DefaultScreen (dsp)),
		 &fore_color);
    XQueryColor (dsp,
		 DefaultColormap (dsp,
				  DefaultScreen (dsp)),
		 &back_color);
    XRecolorCursor (dsp, cursor,
		    &fore_color, &back_color);
    XRecolorCursor (dsp, nontext_cursor,
		    &fore_color, &back_color);
    XRecolorCursor (dsp, mode_cursor,
		    &fore_color, &back_color);
    XRecolorCursor (dsp, cross_cursor,
                    &fore_color, &back_color);
  }
  if (FRAME_WIN32_WINDOW (f) != 0)
    {
      XDefineCursor (XDISPLAY FRAME_WIN32_WINDOW (f), cursor);
    }

  if (cursor != f->display.win32->text_cursor && f->display.win32->text_cursor != 0)
      XFreeCursor (XDISPLAY f->display.win32->text_cursor);
  f->display.win32->text_cursor = cursor;
#ifdef HAVE_X11
  if (nontext_cursor != f->display.win32->nontext_cursor
      && f->display.win32->nontext_cursor != 0)
      XFreeCursor (XDISPLAY f->display.win32->nontext_cursor);
  f->display.win32->nontext_cursor = nontext_cursor;

  if (mode_cursor != f->display.win32->modeline_cursor
      && f->display.win32->modeline_cursor != 0)
      XFreeCursor (XDISPLAY f->display.win32->modeline_cursor);
  f->display.win32->modeline_cursor = mode_cursor;
  if (cross_cursor != f->display.win32->cross_cursor
      && f->display.win32->cross_cursor != 0)
      XFreeCursor (XDISPLAY f->display.win32->cross_cursor);
  f->display.win32->cross_cursor = cross_cursor;
#endif
#endif

  UNBLOCK_INPUT;
}

void
win32_set_cursor_color (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  unsigned long fore_pixel;

  if (!EQ (Vwin32_cursor_fore_pixel, Qnil))
    fore_pixel = win32_decode_color (Vwin32_cursor_fore_pixel, WHITE_PIX_DEFAULT);
  else
    fore_pixel = f->display.win32->background_pixel;
  f->display.win32->cursor_pixel = win32_decode_color (arg, BLACK_PIX_DEFAULT);
  
  /* Make sure that the cursor color differs from the background color.  */
  if (f->display.win32->cursor_pixel == f->display.win32->background_pixel)
    {
      f->display.win32->cursor_pixel = f->display.win32->mouse_pixel;
      if (f->display.win32->cursor_pixel == fore_pixel)
	fore_pixel = f->display.win32->background_pixel;
    }
  f->display.win32->cursor_foreground_pixel = fore_pixel;

  if (FRAME_WIN32_WINDOW (f) != 0)
    {
      if (FRAME_VISIBLE_P (f))
	{
	  win32_display_cursor (f, 0);
	  win32_display_cursor (f, 1);
	}
    }
}

/* Set the border-color of frame F to value described by ARG.
   ARG can be a string naming a color.
   The border-color is used for the border that is drawn by the X server.
   Note that this does not fully take effect if done before
   F has an x-window; it must be redone when the window is created.

   Note: this is done in two routines because of the way X10 works.

   Note: under X11, this is normally the province of the window manager,
   and so emacs' border colors may be overridden. */

void
win32_set_border_color (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  CHECK_STRING (arg, 0);

  win32_set_border_pixel (f, win32_decode_color (arg, BLACK_PIX_DEFAULT));
}

/* Set the border-color of frame F to pixel value PIX.
   Note that this does not fully take effect if done before
   F has an x-window.  */

win32_set_border_pixel (f, pix)
     struct frame *f;
     int pix;
{
  f->display.win32->border_pixel = pix;

  if (FRAME_WIN32_WINDOW (f) != 0 && f->display.win32->border_width > 0)
    {
      if (FRAME_VISIBLE_P (f))
        redraw_frame (f);
    }
}

void
win32_set_cursor_type (f, arg, oldval)
     FRAME_PTR f;
     Lisp_Object arg, oldval;
{
  if (EQ (arg, Qbar))
    FRAME_DESIRED_CURSOR (f) = bar_cursor;
  else
      FRAME_DESIRED_CURSOR (f) = filled_box_cursor;
  /* Error messages commented out because people have trouble fixing
     .Xdefaults with Emacs, when it has something bad in it.  */

  /* Make sure the cursor gets redrawn.  This is overkill, but how
     often do people change cursor types?  */
  update_mode_lines++;
}

extern Lisp_Object win32_new_font ();

#ifdef WIN32_FONTSET
/* 96.4.1 by himi */
void
win32_set_fontset (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  CHECK_STRING (arg, 1);

  win32_change_fontset(f, XSTRING(arg) -> data);
  store_frame_param (f, Qfontset, arg);
}

void
win32_set_font_property (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  Lisp_Object result;

  CHECK_NUMBER (arg, 1);

  BLOCK_INPUT;
  result = win32_new_font (f, (f) -> display.win32 -> font -> name,
			   XFASTINT(arg));
  UNBLOCK_INPUT;
  
  if (EQ (result, Qnil))
    error ("This property in the font \"%s\" is not defined");
  else if (EQ (result, Qt))
    error ("the characters of the given font have varying widths");
  else if (STRINGP (result))
    {
      recompute_basic_faces (f);
      store_frame_param (f, Qfont_property, XFASTINT(arg));
    }
  else
    abort ();
}

#endif

void
win32_set_font (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  Lisp_Object result;

  CHECK_STRING (arg, 1);

  BLOCK_INPUT;
#ifdef WIN32_FONTSET
  result = win32_new_font (f, XSTRING (arg)->data, 0);
#else
  result = win32_new_font (f, XSTRING (arg)->data);
#endif
  UNBLOCK_INPUT;
  
  if (EQ (result, Qnil))
#ifdef MULE			/* 94.6.6 by K.Handa */
#ifdef WIN32_FONTSET            /* 96.4.1 by himi */
    error ("Font in the fontset \"%s\" is not defined", XSTRING (arg)->data);
#else
    error ("Font or fontset \"%s\" is not defined", XSTRING (arg)->data);
#endif
#else   /* not MULE */
    error ("Font \"%s\" is not defined", XSTRING (arg)->data);
#endif  /* not MULE */
  else if (EQ (result, Qt))
    error ("the characters of the given font have varying widths");
  else if (STRINGP (result))
    {
      recompute_basic_faces (f);
      store_frame_param (f, Qfont, result);
    }
  else
    abort ();
}

#ifdef MULE
void
win32_set_line_space (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  int upper, lower;
  char *s, *p;

  CHECK_STRING (arg, 0);
  s = (char *) XSTRING (arg)->data;

  upper = lower = 0;
  if (*s) {
    if (*s != '+')
      upper = atoi(s);
    if (p = (char *) index (s, '+')) {
      lower = atoi(p);
    } else {
      lower = upper - (upper / 2);
      upper /= 2;
    }
  }
  if (upper < 0) upper = 0;
  if (lower < 0) lower = 0;

  if (f->display.win32->upper_space != upper
      || f->display.win32->lower_space != lower) {
    f->display.win32->line_height +=
      upper - f->display.win32->upper_space +
	lower - f->display.win32->lower_space;
    f->display.win32->upper_space = upper;
    f->display.win32->lower_space = lower;
    if (FRAME_WIN32_WINDOW (f) != 0) {
      BLOCK_INPUT;
      win32_set_window_size (f, 0, f->width, f->height);
      UNBLOCK_INPUT;
      SET_FRAME_GARBAGED (f);
    }
  }
}
#endif /* MULE */

void
win32_set_border_width (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  CHECK_NUMBER (arg, 0);

  if (XINT (arg) == f->display.win32->border_width)
    return;

  if (FRAME_WIN32_WINDOW (f) != 0)
    error ("Cannot change the border width of a window");

  f->display.win32->border_width = XINT (arg);
}

void
win32_set_internal_border_width (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  int mask;
  int old = f->display.win32->internal_border_width;

  CHECK_NUMBER (arg, 0);
  f->display.win32->internal_border_width = XINT (arg);
  if (f->display.win32->internal_border_width < 0)
    f->display.win32->internal_border_width = 0;

  if (f->display.win32->internal_border_width == old)
    return;

  if (FRAME_WIN32_WINDOW (f) != 0)
    {
      BLOCK_INPUT;
      win32_set_window_size (f, 0, f->width, f->height);
      UNBLOCK_INPUT;
      SET_FRAME_GARBAGED (f);
    }
}

void
win32_set_visibility (f, value, oldval)
     struct frame *f;
     Lisp_Object value, oldval;
{
  Lisp_Object frame;
  XSET (frame, Lisp_Frame, f);

  if (NILP (value))
    Fmake_frame_invisible (frame, Qt);
  else if (EQ (value, Qicon))
    Ficonify_frame (frame);
  else
    Fmake_frame_visible (frame);
}

static void
win32_set_menu_bar_lines_1 (window, n)
  Lisp_Object window;
  int n;
{
  struct window *w = XWINDOW (window);

  XFASTINT (w->top) += n;
  XFASTINT (w->height) -= n;

  /* Handle just the top child in a vertical split.  */
  if (!NILP (w->vchild))
    win32_set_menu_bar_lines_1 (w->vchild, n);

  /* Adjust all children in a horizontal split.  */
  for (window = w->hchild; !NILP (window); window = w->next)
    {
      w = XWINDOW (window);
      win32_set_menu_bar_lines_1 (window, n);
    }
}

void
win32_set_menu_bar_lines (f, value, oldval)
     struct frame *f;
     Lisp_Object value, oldval;
{
  int nlines;
  int olines = FRAME_MENU_BAR_LINES (f);

  /* Right now, menu bars don't work properly in minibuf-only frames;
     most of the commands try to apply themselves to the minibuffer
     frame itslef, and get an error because you can't switch buffers
     in or split the minibuffer window.  */
  if (FRAME_MINIBUF_ONLY_P (f))
    return;

  if (XTYPE (value) == Lisp_Int)
    nlines = XINT (value);
  else
    nlines = 0;

  FRAME_MENU_BAR_LINES (f) = 0;
  if (nlines)
    FRAME_EXTERNAL_MENU_BAR (f) = 1;
  else
    {
      if (FRAME_EXTERNAL_MENU_BAR (f) == 1)
	free_frame_menubar (f);
      FRAME_EXTERNAL_MENU_BAR (f) = 0;
    }
}

/* Change the name of frame F to NAME.  If NAME is nil, set F's name to
       win32_id_name.

   If EXPLICIT is non-zero, that indicates that lisp code is setting the
       name; if NAME is a string, set F's name to NAME and set
       F->explicit_name; if NAME is Qnil, then clear F->explicit_name.

   If EXPLICIT is zero, that indicates that Emacs redisplay code is
       suggesting a new name, which lisp code should override; if
       F->explicit_name is set, ignore the new name; otherwise, set it.  */

void
win32_set_name (f, name, explicit)
     struct frame *f;
     Lisp_Object name;
     int explicit;
{
  /* Make sure that requests from lisp code override requests from 
     Emacs redisplay code.  */
  if (explicit)
    {
      /* If we're switching from explicit to implicit, we had better
	 update the mode lines and thereby update the title.  */
      if (f->explicit_name && NILP (name))
	update_mode_lines = 1;

      f->explicit_name = ! NILP (name);
    }
  else if (f->explicit_name)
    return;

  /* If NAME is nil, set the name to the win32_id_name.  */
  if (NILP (name))
    {
      /* Check for no change needed in this very common case
	 before we do any consing.  */
      if (!strcmp (win32_id_name, XSTRING (f->name)->data))
	return;
      name = build_string (win32_id_name);
    }
  else
    CHECK_STRING (name, 0);

  /* Don't change the name if it's already NAME.  */
  if (! NILP (Fstring_equal (name, f->name)))
    return;

  if (FRAME_WIN32_WINDOW (f))
    {
#ifdef MULE /* 96.10.12 modified by himi */
      extern Lisp_Object Vwin32_system_coding_system;
      TCHAR *buf;
      int size;
      coding_type mccode;

      encode_code(Fcheck_code(Vwin32_system_coding_system), &mccode);
      buf = alloca(DECODE_BUF_SIZE(XSTRING(name) -> size +1, &mccode));
      CODE_CNTL (&mccode) |= CC_END;
      size = decode(&mccode, XSTRING(name) -> data, buf,
		    XSTRING(name) -> size);
      buf[size] = (TCHAR) NULL;

#endif
      BLOCK_INPUT;
#ifdef MULE
      SetWindowText(FRAME_WIN32_WINDOW (f), buf);
#else     
      SetWindowText(FRAME_WIN32_WINDOW (f),XSTRING(name)->data) ;
#endif
      UNBLOCK_INPUT;
    }

  f->name = name;
}

/* This function should be called when the user's lisp code has
   specified a name for the frame; the name will override any set by the
   redisplay code.  */
void
win32_explicitly_set_name (f, arg, oldval)
     FRAME_PTR f;
     Lisp_Object arg, oldval;
{
  win32_set_name (f, arg, 1);
}

/* This function should be called by Emacs redisplay code to set the
   name; names set this way will never override names set by the user's
   lisp code.  */
void
win32_implicitly_set_name (f, arg, oldval)
     FRAME_PTR f;
     Lisp_Object arg, oldval;
{
  win32_set_name (f, arg, 0);
}

void
win32_set_autoraise (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  f->auto_raise = !EQ (Qnil, arg);
}

void
win32_set_autolower (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  f->auto_lower = !EQ (Qnil, arg);
}

void
win32_set_vertical_scroll_bars (f, arg, oldval)
     struct frame *f;
     Lisp_Object arg, oldval;
{
  if (NILP (arg) != ! FRAME_HAS_VERTICAL_SCROLL_BARS (f))
    {
#ifdef WIN32_SCROLLBAR
      FRAME_HAS_VERTICAL_SCROLL_BARS (f) = ! NILP (arg);

      /* We set this parameter before creating the X window for the
	 frame, so we can get the geometry right from the start.
	 However, if the window hasn't been created yet, we shouldn't
	 call win32_set_window_size.  */
      if (FRAME_WIN32_WINDOW (f))
	win32_set_window_size (f, 0, FRAME_WIDTH (f), FRAME_HEIGHT (f));
#endif
    }
}

#ifndef WIN32_FONTSET
FONT_TYPE *win32_load_font (f,name)
    FRAME_PTR f ;
    char * name;
{
    FONT_TYPE * font = NULL ;
    BOOL ok ;

    {
	LOGFONT lf ;
	
	if (!name || !x_to_win32_font(name,&lf)) 
	    return (NULL) ;
	
	font = (FONT_TYPE *) xmalloc(sizeof(FONT_TYPE)) ;
	
	if (!font) return (NULL) ;
	
	BLOCK_INPUT;
	
	font->hfont = CreateFontIndirect(&lf) ;
    }

    if (font->hfont == NULL) {
	ok = FALSE ;
    } else {
	HDC hdc ;
	HANDLE oldobj ;

	hdc = MyGetDC (FRAME_WIN32_WINDOW (f)) ;
	
	oldobj = SelectObject (hdc,font->hfont) ;
	
	ok = GetTextMetrics(hdc,&font->tm) ;
	
	SelectObject (hdc,oldobj) ;
	
	ReleaseDC (FRAME_WIN32_WINDOW (f),hdc) ;
    }

    UNBLOCK_INPUT ;

    if (ok) return (font) ;

    win32_unload_font(font) ;
    return (NULL) ;
}

void win32_unload_font (f,font)
    FRAME_PTR f ;
    FONT_TYPE * font;
{
    if (font) {
	if (font->hfont) DeleteObject(font->hfont) ;
	xfree (font) ;
    }
}
#endif


/* Subroutines of creating an X frame.  */

DEFUN ("win32-get-resource", Fwin32_get_resource, Swin32_get_resource, 2, 4, 0,
  "Return the value of ATTRIBUTE, of class CLASS, from the Win32 registory \n\
database.\n\
This uses `NAME.ATTRIBUTE' as the key and `Emacs.INSTANCE' as the\n\
class, where INSTANCE is the name under which Emacs was invoked, or\n\
the name specified by the `-name' or `-rn' command-line arguments.\n\
\n\
The optional arguments COMPONENT and SUBCLASS add to the key and the\n\
class, respectively.  You must specify both of them or neither.\n\
If you specify them, the key is `NAME.COMPONENT.ATTRIBUTE'\n\
and the class is `Emacs.INSTANCE.SUBCLASS'.")
  (attribute, class, component, subclass)
     Lisp_Object attribute, class, component, subclass;
{
  register LPBYTE value;
  char *name_key;
  char *class_key;

  CHECK_STRING (attribute, 0);
  CHECK_STRING (class, 0);

  if (!NILP (component))
    CHECK_STRING (component, 1);
  if (!NILP (subclass))
    CHECK_STRING (subclass, 2);
  if (NILP (component) != NILP (subclass))
    error ("win32-get-resource: must specify both COMPONENT and SUBCLASS or neither");

  if (NILP (component))
    {
      /* Allocate space for the components, the dots which separate them,
	 and the final '\0'.  */
      name_key = (char *) alloca (XSTRING (attribute)->size + 1);
      class_key = (char *) alloca (XSTRING (class)->size + 1) ;

      sprintf (name_key, "%s",XSTRING (attribute)->data);
      sprintf (class_key, "%s",XSTRING (class)->data);
    }
  else
    {
      name_key = (char *) alloca (XSTRING (component)->size
				  + XSTRING (attribute)->size
				  + 2);

      class_key = (char *) alloca (XSTRING (class)->size
				   + XSTRING (subclass)->size
				   + 2);

      sprintf (name_key, "%s.%s",
	       XSTRING (component)->data,
	       XSTRING (attribute)->data);
      sprintf (class_key, "%s.%s",
	       XSTRING (class)->data,
	       XSTRING (subclass)->data);
    }

  value = win32_get_string_resource (name_key, REG_ROOT);

  if (value != NULL) {
      Lisp_Object ret = build_string ((char *) value);

      xfree (value) ;

      return ret ;
  } else {
      return Qnil;
  }
}

/* Types we might convert a resource string into.  */
enum resource_types
  {
    number, eboolean, string, symbol
  };

/* Return the value of parameter PARAM.

   First search ALIST, then Vdefault_frame_alist, then the X defaults
   database, using ATTRIBUTE as the attribute name and CLASS as its class.

   Convert the resource to the type specified by desired_type.

   If no default is specified, return Qunbound.  If you call
   win32_get_arg, make sure you deal with Qunbound in a reasonable way,
   and don't let it get stored in any lisp-visible variables!  */

static Lisp_Object
win32_get_arg (alist, param, attribute, class, type)
     Lisp_Object alist, param;
     char *attribute;
     char *class;
     enum resource_types type;
{
  register Lisp_Object tem;

  tem = Fassq (param, alist);
  if (EQ (tem, Qnil))
    tem = Fassq (param, Vdefault_frame_alist);
  if (EQ (tem, Qnil))
    {

      if (attribute)
	{
	  tem = Fwin32_get_resource (build_string (attribute),
				 build_string (class),
				 Qnil, Qnil);

	  if (NILP (tem))
	    return Qunbound;

	  switch (type)
	    {
	    case number:
	      return make_number (atoi (XSTRING (tem)->data));

	    case eboolean:
	      tem = Fdowncase (tem);
	      if (!strcmp (XSTRING (tem)->data, "on")
		  || !strcmp (XSTRING (tem)->data, "true"))
		return Qt;
	      else 
		return Qnil;

	    case string:
	      return tem;

	    case symbol:
	      /* As a special case, we map the values `true' and `on'
		 to Qt, and `false' and `off' to Qnil.  */
	      {
		Lisp_Object lower;
		lower = Fdowncase (tem);
		if (!strcmp (XSTRING (lower)->data, "on")
		    || !strcmp (XSTRING (lower)->data, "true"))
		  return Qt;
		else if (!strcmp (XSTRING (lower)->data, "off")
		      || !strcmp (XSTRING (lower)->data, "false"))
		  return Qnil;
		else
		  return Fintern (tem, Qnil);
	      }

	    default:
	      abort ();
	    }
	}
      else
	return Qunbound;
    }
  return Fcdr (tem);
}

/* Record in frame F the specified or default value according to ALIST
   of the parameter named PARAM (a Lisp symbol).
   If no value is specified for PARAM, look for an X default for XPROP
   on the frame named NAME.
   If that is not found either, use the value DEFLT.  */

static Lisp_Object
win32_default_parameter (f, alist, prop, deflt, xprop, xclass, type)
     struct frame *f;
     Lisp_Object alist;
     Lisp_Object prop;
     Lisp_Object deflt;
     char *xprop;
     char *xclass;
     enum resource_types type;
{
  Lisp_Object tem;

  tem = win32_get_arg (alist, prop, xprop, xclass, type);
  if (EQ (tem, Qunbound))
    tem = deflt;
  win32_set_frame_parameters (f, Fcons (Fcons (prop, tem), Qnil));
  return tem;
}

/* Calculate the desired size and position of this window,
   and return the flags saying which aspects were specified.

   This function does not make the coordinates positive.  */

#define DEFAULT_ROWS(f) ((GetSystemMetrics(SM_CYFULLSCREEN) * 2 \
			 /(FRAME_FONT_HEIGHT(f) * 3)))
#define DEFAULT_COLS(f) ((GetSystemMetrics(SM_CXFULLSCREEN) - \
			  VERTICAL_SCROLL_BAR_PIXEL_WIDTH(f)) * 2 \
			 /(FRAME_FONT_WIDTH(f) * 3))

static int
win32_figure_window_size (f, parms)
     struct frame *f;
     Lisp_Object parms;
{
  register Lisp_Object tem0, tem1, tem2;
  int height, width, left, top;
  register int geometry;
  long window_prompting = 0;

  /* Default values if we fall through.
     Actually, if that happens we should get
     window manager prompting. */
  f->width = DEFAULT_COLS(f);
  f->height = DEFAULT_ROWS(f);
  /* Window managers expect that if program-specified
     positions are not (0,0), they're intentional, not defaults.  */
  f->display.win32->top_pos = 0;
  f->display.win32->left_pos = 0;

  tem0 = win32_get_arg (parms, Qheight, 0, 0, number);
  tem1 = win32_get_arg (parms, Qwidth, 0, 0, number);
  tem2 = win32_get_arg (parms, Quser_size, 0, 0, number);
  if (! EQ (tem0, Qunbound) || ! EQ (tem1, Qunbound))
    {
      if (!EQ (tem0, Qunbound))
	{
	  CHECK_NUMBER (tem0, 0);
	  f->height = XINT (tem0);
	}
      if (!EQ (tem1, Qunbound))
	{
	  CHECK_NUMBER (tem1, 0);
	  f->width = XINT (tem1);
	}
#if 0
      if (!NILP (tem2) && !EQ (tem2, Qunbound))
	window_prompting |= USSize;
      else
	window_prompting |= PSize;
#endif
    }

  f->display.win32->vertical_scroll_bar_extra = 0 ;

  f->display.win32->pixel_width = CHAR_TO_PIXEL_WIDTH (f, f->width);
  f->display.win32->pixel_height = CHAR_TO_PIXEL_HEIGHT (f, f->height);

  tem0 = win32_get_arg (parms, Qtop, 0, 0, number);
  tem1 = win32_get_arg (parms, Qleft, 0, 0, number);
  tem2 = win32_get_arg (parms, Quser_position, 0, 0, number);
  if (! EQ (tem0, Qunbound) || ! EQ (tem1, Qunbound))
    {
      if (EQ (tem0, Qminus))
	{
	  f->display.win32->top_pos = 0;
#if 0
	  window_prompting |= YNegative;
#endif
	}
      else if (EQ (tem0, Qunbound))
	f->display.win32->top_pos = 0;
      else
	{
	  CHECK_NUMBER (tem0, 0);
	  f->display.win32->top_pos = XINT (tem0);
#if 0
	  if (f->display.win32->top_pos < 0)
	    window_prompting |= YNegative;
#endif
	}

      if (EQ (tem1, Qminus))
	{
	  f->display.win32->left_pos = 0;
#if 0
	  window_prompting |= XNegative;
#endif
	}
      else if (EQ (tem1, Qunbound))
	f->display.win32->left_pos = 0;
      else
	{
	  CHECK_NUMBER (tem1, 0);
	  f->display.win32->left_pos = XINT (tem1);
#if 0
	  if (f->display.win32->left_pos < 0)
	    window_prompting |= XNegative;
#endif
	}

#if 0
      if (!NILP (tem2))
	window_prompting |= USPosition;
      else
	window_prompting |= PPosition;
#endif
    }

  return window_prompting;
}

#if 0
HDC MapMode(hdc)
    HDC hdc ;
{
    if (hdc) {
#if 0
	/* Make mapping mode be in 1/20 of point */
	
	SetMapMode (hdc, MM_ANISOTROPIC) ;
	SetWindowExtEx (hdc, 1440, 1440,NULL) ;
	SetViewportExtEx (hdc,
			  GetDeviceCaps (hdc, LOGPIXELSX),
			  GetDeviceCaps (hdc, LOGPIXELSY),
			  NULL) ;
#endif
    }

    return (hdc) ;
}
#endif

#ifdef MULE
HANDLE win32_ime_string_handle(HANDLE hStr)
{
  HANDLE hwin32_ir_string;
  LPTSTR lpStr;
  LPTSTR lpCode;

  if (!hStr) return 0;
  lpStr = GlobalLock (hStr);
  if (!lpStr) return 0;
  hwin32_ir_string = 
    GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, 
		strlen(lpStr) + 1);
  if (!hwin32_ir_string) abort();
  lpCode = GlobalLock(hwin32_ir_string);
  strcpy(lpCode, lpStr);
  GlobalUnlock(hwin32_ir_string);
  GlobalUnlock (hStr);
  return hwin32_ir_string;
}
#endif

/* Create and set up the X widget for frame F.  */

LRESULT CALLBACK win32_WndProc (hwnd,msg,wParam,lParam)
    HWND hwnd ;
    UINT msg ;
    WPARAM wParam ;
    LPARAM lParam ;
{
#ifdef IME_CONTROL
  extern int IME_event_off_count;
#endif
  struct frame *f;
  WPARAM SCParam;
  LRESULT ret = 1 ;

  switch (msg) {
  case WM_ERASEBKGND:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    f = win32_window_to_frame (hwnd) ;

    if (f)
      {
	RECT rect ;
	
	GetClientRect (FRAME_WIN32_WINDOW (f),&rect) ;
	win32_clear_rect (f,MapMode((HDC)wParam),&rect) ;
	ret = 0 ;
      }

    /*	ReleaseMutex(hMutex) ; */

    if (ret == 0) return (0) ;

    goto dflt ;
  case WM_PAINT:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    {
      PAINTSTRUCT paintStruct ;

      f = win32_window_to_frame (hwnd) ;
      if (f) {
	if (f->async_visible == 0)
	  {
	    f->async_visible = 1;
	    f->async_iconified = 0;
	    SET_FRAME_GARBAGED (f);
	  } else {
	    RECT rect ;
		    
	    BeginPaint (hwnd,&paintStruct) ;
	    rect = paintStruct.rcPaint ;
	    EndPaint (hwnd,&paintStruct) ;
		    
	    dumprectangle(f,
			  rect.left,
			  rect.top,
			  rect.right-rect.left+1,
			  rect.bottom-rect.top+1) ;
	    ret = 0 ;
	  }
      }
    }
		    
    /*	ReleaseMutex(hMutex) ; */

    if (ret == 0) return (0) ;

    goto dflt ;
  case WM_SETFOCUS:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    f = win32_window_to_frame (hwnd) ;

    win32_new_focus_frame (f) ;

    PostMessage(hwnd, WM_EMACS_ACTIVATE, 
		(WPARAM) 0, (LPARAM) 0);
    /* To switch selected-frame  by himi */

#if defined(MULE) && defined(IME_CONTROL)
    if (f)
      win32_set_ime_font(hwnd, 
			 get_logfont(FRAME_FONTSET(f),
				     f->display.win32->font->fs_font, 
				     f->display.win32->font->
				     property_index));
#endif

    /*	ReleaseMutex(hMutex) ; */

    return (0) ;
  case WM_KILLFOCUS:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    f = win32_window_to_frame (hwnd) ;

    if (f == win32_focus_frame)
      win32_new_focus_frame (NULL) ;

    /*	ReleaseMutex(hMutex) ; */

    return (0) ;
  case WM_MOVE:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    f = win32_window_to_frame (hwnd) ;
	
    if (f && !f->async_iconified && f->async_visible)
      {
	RECT rect;

	GetWindowRect(hwnd, &rect);

	f->display.win32->left_pos = rect.left;
	f->display.win32->top_pos = rect.top;
      }

    /*	ReleaseMutex(hMutex) ; */

    break ;
  case WM_SIZE:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    f = win32_window_to_frame (hwnd) ;
	
    if (f && !f->async_iconified && f->async_visible)
      {
	RECT rect ;
	int rows ;
	int columns ;
	int width ;
	int height ;
	    
	GetClientRect(hwnd,&rect) ;

	height = rect.bottom - rect.top + 1 ;
	width = rect.right - rect.left + 1 ;

	rows = PIXEL_TO_CHAR_HEIGHT (f,height) ;
	columns = PIXEL_TO_CHAR_WIDTH (f,width) ;

	/* Even if the number of character rows and columns has
	   not changed, the font size may have changed, so we need
	   to check the pixel dimensions as well.  */
	    
	if (columns != f->width
	    || rows != f->height
	    || width != f->display.win32->pixel_width
	    || height != f->display.win32->pixel_height)
	  {
	    change_frame_size (f, rows, columns, 0, 0);
	    //		change_frame_size (f, rows, columns, 0, 1);
	    SET_FRAME_GARBAGED (f);
	    
	    f->display.win32->pixel_width = width ;
	    f->display.win32->pixel_height = height ;
	  }
      }

    /*	ReleaseMutex(hMutex) ; */

    break;
  case WM_WINDOWPOSCHANGING:
    /*	WaitForSingleObject (hMutex,INFINITE) ; */

    f = win32_window_to_frame (hwnd) ;
	
    if (f && !f->async_iconified && f->async_visible)
      {
	LPWINDOWPOS lppos = (WINDOWPOS *) lParam ;

	if (!(lppos->flags & SWP_NOSIZE))
	  {
	    RECT rect;
	    int wdiff;
	    int hdiff;

	    memset(&rect, 0, sizeof(rect));
	    AdjustWindowRect(&rect, 
			     f->display.win32->dwStyle, 
			     FRAME_EXTERNAL_MENU_BAR (f));

	    /* All windows have an extra pixel so subtract 1 */

#ifdef WIN32_FONTSET
	    wdiff = (lppos->cx - (rect.right - rect.left) - 1 - 
		     f->display.win32->vertical_scroll_bar_extra) % 
		       FRAME_FONT_WIDTH(f);
#else
	    wdiff = (lppos->cx - (rect.right - rect.left) - 1) %
	      FONT_WIDTH (f->display.win32->font) ;
#endif
	    hdiff = (lppos->cy - (rect.bottom - rect.top) - 1) %
	      f->display.win32->line_height ;

	    if (wdiff || hdiff) {
	      lppos->cx -= wdiff ;
	      lppos->cy -= hdiff ;
			    
	      ret = 0 ;
	    }
	  }
      }

    /*	ReleaseMutex(hMutex) ; */

    if (ret == 0) return (0) ;

    goto dflt ;
  case WM_SYSCOMMAND:
    SCParam = wParam & 0xFFF0;
    switch (SCParam) {
    case SC_CLOSE:
      PostMessage(hwnd,WM_EMACS_DESTROY,wParam,lParam) ;
      return (0) ;
    case SC_MINIMIZE:
      /*	    WaitForSingleObject (hMutex,INFINITE) ; */

      f = win32_window_to_frame (hwnd) ;
	
      if (f) {
	f->async_visible = 1;
	f->async_iconified = 1;
      }

      /*	    ReleaseMutex(hMutex) ; */

      break ;
    case SC_MAXIMIZE:
    case SC_RESTORE:
      /*	    WaitForSingleObject (hMutex,INFINITE) ; */

      f = win32_window_to_frame (hwnd) ;
	
      if (f) {
	f->async_visible = 1;
	f->async_iconified = 0;
	SET_FRAME_GARBAGED (f);
      }

      /*	    ReleaseMutex(hMutex) ; */

      break ;
    }
    goto dflt ;
  case WM_CLOSE:
  case WM_DESTROY:
    PostMessage(hwnd,WM_EMACS_DESTROY,wParam,lParam) ;
    return (0) ;
  case WM_CREATE:
    {
      HDC hdc = MyGetDC(hwnd) ;

      /* Make mapping mode be in 1/20 of point */

      MapMode (hdc) ;

      ReleaseDC(hwnd,hdc) ;
    }
    return (0) ;
#ifdef WIN32_SCROLLBAR
  case WM_VSCROLL:
    {
      if ((LOWORD(wParam) == SB_THUMBTRACK) || (LOWORD(wParam) == SB_THUMBPOSITION)) {
	extern struct scroll_bar *
	  win32_window_to_scroll_bar (HWND window_id);
	extern int update_frame();

	struct scroll_bar *bar;
	Lisp_Object window, top_range, pos;
	
	bar = win32_window_to_scroll_bar ((HWND)lParam);
	window = bar->window;
	f = XFRAME(WINDOW_FRAME(XWINDOW(window)));
	XSET(top_range, Lisp_Int, 
	     VERTICAL_SCROLL_BAR_TOP_RANGE(XINT(bar->height)));
	XSET(pos, Lisp_Int, (int) HIWORD(wParam));
	
	
	call3 (intern ("win32-scrollbar-thumbtrack"),
	       window, top_range, pos);
	update_frame(f, 1, 0);
	redisplay();
	
      } else {
	PostMessage(NULL, WM_EMACS_VSCROLL, wParam, lParam);
	WIN32_SPECIAL_INPUT;
	read_input_waiting();
	WIN32_SPECIAL_INPUT_END;
      }
#if 0
      if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) &&
	  (msg.message==WM_EMACS_VSCROLL))
	{
	  printf("Scroll Bar Message Posted %d\n", msg.wParam);
	}
      else
	{
	  printf("Scroll Bar Message Not Posted %d\n");
	}
      fflush(stdout);
#endif
      return 0;
    }
#endif
#ifdef WIN32_INTELLIMOUSE
  case WM_MOUSEWHEEL:
    PostMessage(hwnd, WM_MOUSEWHEEL, wParam, lParam);
    return 0;
#endif
#ifdef MULE			/* 95.10.1 by himi */
#ifdef WIN32_VER4
  case WM_IME_NOTIFY:
    if (wParam == IMN_SETOPENSTATUS)
      {
	if (!IME_event_off_count)
	  PostMessage(hwnd, WM_MULE_IME_STATUS, 0, 0);
	else
	  IME_event_off_count--;
      }
    goto dflt;

  case WM_IME_COMPOSITION:
    {
      if (lParam & GCS_RESULTSTR) {
	extern BOOL win32_get_ime_composion_string(HWND);
	if (win32_get_ime_composition_string(hwnd))
	  return 0;
	else 
	  break;
      }
      goto dflt;
    }
#endif
#if !defined(WIN32_VER4) || defined(__BORLANDC__) && (__BORLANDC__ < 0x500)
    /*
     * ???: When this program is compiled by BCC 4.5,
     * There's no WM_IME_NOTIFY sent to this procedure.
     * So we stick to use WM_IME_REPORT even in case with WIN32_VER4.
     * (bsh: 1997-Jul-28)
     */
  case WM_IME_REPORT:
    switch (wParam)
      {
      case IR_STRING:
	{
	  struct frame *f;
	  HANDLE himestr;

	  himestr = win32_ime_string_handle((HANDLE)lParam);
	  if (!himestr) break;

	  f = win32_window_to_frame(hwnd);
	  PostMessage(NULL, WM_MULE_IME_REPORT, 
		      (WPARAM) himestr,(LPARAM) f);
	  return 1;
	}
      case IR_OPENCONVERT:
      case IR_CLOSECONVERT:
	if (!IME_event_off_count)
	  PostMessage(hwnd, WM_MULE_IME_STATUS, 0, 0);
	else
	  IME_event_off_count--;
	return 0;
      }
#endif
#endif /* MULE */
  default:
  dflt:
    return DefWindowProc(hwnd,msg,wParam,lParam) ;
  }

  return (1) ;
}

BOOL win32_init_app (hinst)
    HINSTANCE hinst;
{
    WNDCLASS wc ;

    wc.style = CS_HREDRAW | CS_VREDRAW ;
    wc.lpfnWndProc = (WNDPROC) win32_WndProc ;
    wc.cbClsExtra = 0 ;
    wc.cbWndExtra = 0 ;
    wc.hInstance = hinst ;
    wc.hIcon = LoadIcon (hinst,EMACS_CLASS) ;
    wc.hCursor = LoadCursor (NULL,IDC_ARROW) ;
    wc.hbrBackground = GetStockObject (WHITE_BRUSH) ;
    wc.lpszMenuName = NULL ;
    wc.lpszClassName = EMACS_CLASS ;

    return (RegisterClass(&wc)) ;
}

static void
win32_window (f, window_prompting, minibuffer_only)
     struct frame *f;
     long window_prompting;
     int minibuffer_only;
{
  HWND hwnd ;
  char* name;

  BLOCK_INPUT;

  if (STRINGP (f->name))
    name = (char*) XSTRING (f->name)->data;
  else
    name = "emacs";

  hwnd = CreateWindow (EMACS_CLASS,
		       name,
		       f->display.win32->dwStyle,
		       f->display.win32->left_pos,
		       f->display.win32->top_pos,
		       PIXEL_WIDTH (f),
		       PIXEL_HEIGHT (f),
		       NULL,
		       NULL,
		       hinst,
		       NULL) ;

  FRAME_WIN32_WINDOW (f) = hwnd ;

  if (!minibuffer_only && FRAME_EXTERNAL_MENU_BAR (f))
    initialize_frame_menubar (f);

  win32_calc_absolute_position (f);

  /* win32_set_name normally ignores requests to set the name if the
     requested name is the same as the current name.  This is the one
     place where that assumption isn't correct; f->name is set, but
     the X server hasn't been told.  */
  {
    Lisp_Object name;
    int explicit = f->explicit_name;

    f->explicit_name = 0;
    name = f->name;
    f->name = Qnil;
    win32_set_name (f, name, explicit);
  }

#if 0
  XDefineCursor (XDISPLAY FRAME_WIN32_WINDOW (f),
		 f->display.win32->text_cursor);
#endif

  DragAcceptFiles(FRAME_WIN32_WINDOW(f), TRUE);

  UNBLOCK_INPUT;

  if (FRAME_WIN32_WINDOW (f) == 0)
    error ("Unable to create window");

#ifdef WIN32_FONTSET
  win32_set_ime_font(hwnd, 
		     get_logfont(FRAME_FONTSET(f),
				 f->display.win32->font->fs_font, 
				 f->display.win32->font->
				 property_index));
#endif
}

DEFUN ("win32-create-frame", Fwin32_create_frame, Swin32_create_frame,
       1, 1, 0,
  "Make a new window, which is called a \"frame\" in Emacs terms.\n\
Return an Emacs frame object representing the window.\n\
ALIST is an alist of frame parameters.\n\
If the parameters specify that the frame should not have a minibuffer,\n\
and do not specify a specific minibuffer window to use,\n\
then `default-minibuffer-frame' must be a frame whose minibuffer can\n\
be shared by the new frame.")
  (parms)
     Lisp_Object parms;
{
  struct frame *f;
  Lisp_Object frame, tem;
  Lisp_Object name;
  int minibuffer_only = 0;
  long window_prompting = 0;
  int width, height;
  int count = specpdl_ptr - specpdl;
  struct gcpro gcpro1;

  name = win32_get_arg (parms, Qname, "title", "Title", string);
  if (XTYPE (name) != Lisp_String
      && ! EQ (name, Qunbound)
      && ! NILP (name))
    error ("win32-create-frame: name parameter must be a string");

  tem = win32_get_arg (parms, Qminibuffer, 0, 0, symbol);
  if (EQ (tem, Qnone) || NILP (tem))
    f = make_frame_without_minibuffer (Qnil);
  else if (EQ (tem, Qonly))
    {
      f = make_minibuffer_frame ();
      minibuffer_only = 1;
    }
  else if (XTYPE (tem) == Lisp_Window)
    f = make_frame_without_minibuffer (tem);
  else
    f = make_frame (1);

  /* Note that win32 does NOT support scroll bars.  */
  FRAME_CAN_HAVE_SCROLL_BARS (f) = 0;

  /* Set the name; the functions to which we pass f expect the name to
     be set.  */
  if (EQ (name, Qunbound) || NILP (name))
    {
      f->name = build_string (win32_id_name);
      f->explicit_name = 0;
    }
  else
    {
      f->name = name;
      f->explicit_name = 1;
      /* use the frame's title when getting resources for this frame.  */
      specbind (Qwin32_resource_name, name);
    }

  XSET (frame, Lisp_Frame, f);
  GCPRO1 (frame);

  f->output_method = output_win32;
  f->display.win32 = (struct win32_display *) xmalloc (sizeof (struct win32_display));
  bzero (f->display.win32, sizeof (struct win32_display));

  /* Note that the frame has no physical cursor right now.  */
  f->phys_cursor_x = -1;

  /* Fontset Initialization */
#ifdef WIN32_FONTSET
  {
    Lisp_Object fontset;
    fontset = win32_get_arg (parms, Qfontset, "fontset", "Fontset", string);
    if (STRINGP (fontset))
      win32_default_parameter (f, parms, Qfontset, fontset, 
			       "fontset", "Fontset", string);
    else
      win32_default_parameter (f, parms, Qfontset, build_string("default"), 
			       "fontset", "Fontset", string);
  }
#endif

  /* Extract the window parameters from the supplied values
     that are needed to determine window geometry.  */
  {
    Lisp_Object font;

    font = win32_get_arg (parms, Qfont, "font", "Font", string);
#ifdef WIN32_FONTSET
    if (! STRINGP (font))
      font = build_string ("default");
#else
    BLOCK_INPUT;
    /* First, try whatever font the caller has specified.  */
    if (STRINGP (font))
      font = win32_new_font (f, XSTRING (font)->data);

    /* If those didn't work, look for something which will at least work.  */
    if (! STRINGP (font))
      font = win32_new_font (f, "-*-system-normal-r-normal-*-*-200-*-*-c-120-*-*");
    UNBLOCK_INPUT;
#ifdef MULE_FONTSET
    if (! STRINGP (font))
      font = build_string ("default-fontset");
#endif  /* MULE */
    if (! STRINGP (font))
      font = build_string ("-*-system-");
#endif /* not WIN32_FONTSET */

    win32_default_parameter (f, parms, Qfont, font, 
			 "font", "Font", string);
  }

#ifdef MULE
  win32_default_parameter (f, parms, Qline_space, build_string ("2"),
		       "linespace", "LineSpace", string);
#endif  /* MULE */

  win32_default_parameter (f, parms, Qborder_width, make_number (0),
		       "borderwidth", "BorderWidth", number);
  /* This defaults to 0 in order to match win32.  We recognize either
     internalBorderWidth or internalBorder (which is what xterm calls
     it).  */
  if (NILP (Fassq (Qinternal_border_width, parms)))
    {
      Lisp_Object value;

      value = win32_get_arg (parms, Qinternal_border_width,
			 "internalBorder", "BorderWidth", number);
      if (! EQ (value, Qunbound))
	parms = Fcons (Fcons (Qinternal_border_width, value),
		       parms);
    }
  win32_default_parameter (f, parms, Qinternal_border_width, make_number (0),
		       "internalBorderWidth", "BorderWidth", number);
  win32_default_parameter (f, parms, Qvertical_scroll_bars, Qt,
		       "verticalScrollBars", "ScrollBars", eboolean);

  /* Also do the stuff which must be set before the window exists. */
  win32_default_parameter (f, parms, Qforeground_color, build_string ("black"),
		       "foreground", "Foreground", string);
  win32_default_parameter (f, parms, Qbackground_color, build_string ("white"),
		       "background", "Background", string);
  win32_default_parameter (f, parms, Qmouse_color, build_string ("black"),
		       "pointerColor", "Foreground", string);
  win32_default_parameter (f, parms, Qcursor_color, build_string ("black"),
		       "cursorColor", "Foreground", string);
  win32_default_parameter (f, parms, Qborder_color, build_string ("black"),
		       "borderColor", "BorderColor", string);

  win32_default_parameter (f, parms, Qmenu_bar_lines, make_number (0),
		       "menuBarLines", "MenuBarLines", number);

  f->display.win32->parent_desc = ROOT_WINDOW;

  /* 96.10.3 modified by himi (to clip scroll bar) */
  f->display.win32->dwStyle =
    WS_OVERLAPPEDWINDOW |
      WS_CLIPCHILDREN |
      WS_VISIBLE;
  
  window_prompting = win32_figure_window_size (f, parms);

#if 0
  if (window_prompting & XNegative)
    {
      if (window_prompting & YNegative)
	f->display.win32->win_gravity = SouthEastGravity;
      else
	f->display.win32->win_gravity = NorthEastGravity;
    }
  else
    {
      if (window_prompting & YNegative)
	f->display.win32->win_gravity = SouthWestGravity;
      else
	f->display.win32->win_gravity = NorthWestGravity;
    }

  f->display.win32->size_hint_flags = window_prompting;
#endif

  win32_window (f, window_prompting, minibuffer_only);
  init_frame_faces (f);

  /* We need to do this after creating the X window, so that the
     icon-creation functions can say whose icon they're describing.  */
  win32_default_parameter (f, parms, Qauto_raise, Qnil,
		       "autoRaise", "AutoRaiseLower", eboolean);
  win32_default_parameter (f, parms, Qauto_lower, Qnil,
		       "autoLower", "AutoRaiseLower", eboolean);
  win32_default_parameter (f, parms, Qcursor_type, Qbox,
		       "cursorType", "CursorType", symbol);

  /* Dimensions, especially f->height, must be done via change_frame_size.
     Change will not be effected unless different from the current
     f->height. */
  width = f->width;
  height = f->height;
  f->height = f->width = 0;
  change_frame_size (f, height, width, 1, 0);
  win32_set_window_size(f,0,width,height) ;

  tem = win32_get_arg (parms, Qunsplittable, 0, 0, eboolean);
  f->no_split = minibuffer_only || EQ (tem, Qt);

  UNGCPRO;

  /* It is now ok to make the frame official
     even if we get an error below.
     And the frame needs to be on Vframe_list
     or making it visible won't work.  */
  Vframe_list = Fcons (frame, Vframe_list);

  /* Make the window appear on the frame and enable display,
     unless the caller says not to.  */
  {
    Lisp_Object visibility;
    int exist_frames = (XGCTYPE (Vframe_list) == Lisp_Cons);

    visibility = win32_get_arg (parms, Qvisibility, 0, 0, symbol);
    if (EQ (visibility, Qunbound))
      visibility = Qt;

    if (EQ (visibility, Qicon))
      win32_iconify_frame (f);
    else {
	win32_make_frame_visible (f);
	win32_new_focus_frame (f) ;
	SetFocus (FRAME_WIN32_WINDOW(f)) ;
	Fselect_frame(frame, Qnil);
	/* To switch selected-frame  by himi */
    }
  }

  return unbind_to (count, frame);
}

Lisp_Object
win32_get_focus_frame ()
{
  Lisp_Object xfocus;
  if (! win32_focus_frame)
    return Qnil;

  XSET (xfocus, Lisp_Frame, win32_focus_frame);
  return xfocus;
}

DEFUN ("focus-frame", Ffocus_frame, Sfocus_frame, 1, 1, 0,
  "Set the focus on FRAME.")
  (frame)
     Lisp_Object frame;
{
  CHECK_LIVE_FRAME (frame, 0);

  if (FRAME_WIN32_P (XFRAME (frame)))
    {
      BLOCK_INPUT;
      win32_focus_on_frame (XFRAME (frame));
      UNBLOCK_INPUT;
      return frame;
    }

  return Qnil;
}

DEFUN ("unfocus-frame", Funfocus_frame, Sunfocus_frame, 0, 0, 0,
  "If a frame has been focused, release it.")
  ()
{
  if (win32_focus_frame)
    {
      BLOCK_INPUT;
      win32_unfocus_frame (win32_focus_frame);
      UNBLOCK_INPUT;
    }

  return Qnil;
}


#ifndef WIN32_FONTSET
typedef struct enumfont_t {
    HDC hdc ;
    FONT_TYPE * size_ref ;
    Lisp_Object * pattern ;
    Lisp_Object * head ;
    Lisp_Object * tail ;
} enumfont_t ;

int CALLBACK EnumFontCB2 (lplf,lptm,FontType,lpef)
    ENUMLOGFONT * lplf ;
    NEWTEXTMETRIC * lptm ;
    int FontType ;
    enumfont_t * lpef ;
{
    if (lplf->elfLogFont.lfStrikeOut || lplf->elfLogFont.lfUnderline || 
	(lplf->elfLogFont.lfCharSet != ANSI_CHARSET && lplf->elfLogFont.lfCharSet != OEM_CHARSET))
	return (1) ;
	
    if (!lpef->size_ref || lptm->tmMaxCharWidth == FONT_WIDTH(lpef->size_ref))
    {
	char buf[100] ;

	if (!win32_to_x_font(lplf,buf,100)) return (0) ;

	if (NILP(*(lpef->pattern)) || win32_font_match(buf,XSTRING(*(lpef->pattern))->data))
	{
	    *lpef->tail = Fcons (build_string (buf), Qnil);
	    lpef->tail = &XCONS (*lpef->tail)->cdr;
	}
    }

    return (1) ;
}

int CALLBACK EnumFontCB1 (lplf,lptm,FontType,lpef)
    ENUMLOGFONT * lplf ;
    NEWTEXTMETRIC * lptm ;
    int FontType ;
    enumfont_t * lpef ;
{
    return EnumFontFamilies (lpef->hdc,
			     lplf->elfLogFont.lfFaceName,
			     (FONTENUMPROC)EnumFontCB2,
			     (LPARAM)lpef) ;
}


DEFUN ("win32-list-fonts", Fwin32_list_fonts, Swin32_list_fonts, 1, 3, 0,
  "Return a list of the names of available fonts matching PATTERN.\n\
If optional arguments FACE and FRAME are specified, return only fonts\n\
the same size as FACE on FRAME.\n\
\n\
PATTERN is a string, perhaps with wildcard characters;\n\
  the * character matches any substring, and\n\
  the ? character matches any single character.\n\
  PATTERN is case-insensitive.\n\
FACE is a face name - a symbol.\n\
\n\
The return value is a list of strings, suitable as arguments to\n\
set-face-font.\n\
\n\
Fonts Emacs can't use (i.e. proportional fonts) may or may not be excluded\n\
even if they match PATTERN and FACE.")
  (pattern, face, frame)
    Lisp_Object pattern, face, frame;
{
  Lisp_Object list;
  enumfont_t ef ;
  BOOL ok ;
  FRAME_PTR f = NILP (frame) ? selected_frame : XFRAME (frame);

  CHECK_STRING (pattern, 0);
  if (!NILP (face))
    CHECK_SYMBOL (face, 1);
  if (!NILP (frame))
    CHECK_LIVE_FRAME (frame, 2);

  if (NILP (face))
    ef.size_ref = 0;
  else
    {
      int face_id;

      /* Don't die if we get called with a terminal frame.  */
      if (! FRAME_WIN32_P (f))
	error ("non-win32 frame used in `win32-list-fonts'");

      face_id = face_name_id_number (f, face);

      if (face_id < 0 || face_id >= FRAME_N_PARAM_FACES (f)
	  || FRAME_PARAM_FACES (f) [face_id] == 0)
	ef.size_ref = f->display.win32->font;
      else
	{
	  ef.size_ref = FRAME_PARAM_FACES (f) [face_id]->font;
	  if (ef.size_ref == (FONT_TYPE *) (~0))
	    ef.size_ref = f->display.win32->font;
	}
    }

  list = Qnil;
  ef.pattern = &pattern ;
  ef.tail = ef.head = &list;

  BLOCK_INPUT;

  {
      ef.hdc = MyGetDC (FRAME_WIN32_WINDOW(f)) ;

      ok = EnumFontFamilies (ef.hdc,NULL,(FONTENUMPROC)EnumFontCB1,(LPARAM)&ef) ;

      ReleaseDC  (FRAME_WIN32_WINDOW(f),ef.hdc) ;
  }

  UNBLOCK_INPUT;

  return list;
}

DEFUN ("win32-find-font", Fwin32_find_font, Swin32_find_font, 1, 3, 0,
  "Return a list of the names of available fonts matching PATTERN.\n\
If optional arguments FACE and FRAME are specified, return only fonts\n\
the same size as FACE on FRAME.\n\
\n\
PATTERN is a string, perhaps with wildcard characters;\n\
  the * character matches any substring, and\n\
  the ? character matches any single character.\n\
  PATTERN is case-insensitive.\n\
FACE is a face name - a symbol.\n\
\n\
The return value is a list of strings, suitable as arguments to\n\
set-face-font.\n\
\n\
Fonts Emacs can't use (i.e. proportional fonts) may or may not be excluded\n\
even if they match PATTERN and FACE.")
  (pattern, face, frame)
    Lisp_Object pattern, face, frame;
{
  LOGFONT lf ;

  CHECK_STRING (pattern, 0);
  if (!NILP (face))
    CHECK_SYMBOL (face, 1);
  if (!NILP (frame))
    CHECK_LIVE_FRAME (frame, 2);

  if (!x_to_win32_font(XSTRING(pattern)->data,&lf))
      return Qnil ;
  
  if (!NILP (face))
  {
      FONT_TYPE * size_ref ;
      int face_id;
      FRAME_PTR f = NILP (frame) ? selected_frame : XFRAME (frame);
      
      /* Don't die if we get called with a terminal frame.  */
      if (! FRAME_WIN32_P (f))
	  error ("non-win32 frame used in `win32-find-font'");
      
      face_id = face_name_id_number (f, face);
      
      if (face_id < 0 || face_id >= FRAME_N_PARAM_FACES (f)
	  || FRAME_PARAM_FACES (f) [face_id] == 0)
	  size_ref = f->display.win32->font;
      else
      {
	  size_ref = FRAME_PARAM_FACES (f) [face_id]->font;
	  if (size_ref == (FONT_TYPE *) (~0))
	    size_ref = f->display.win32->font;
      }

      if (size_ref) {
	  if (lf.lfWidth == 0) lf.lfWidth= FONT_WIDTH(size_ref) ;
	  if (lf.lfHeight == 0) lf.lfHeight = FONT_HEIGHT(size_ref) ;
      }
  }
  
  lf.lfPitchAndFamily = FIXED_PITCH ;
  if (lf.lfFaceName[0] == 0) {
      strcpy (lf.lfFaceName,"System") ;
  }
	  
  {
      char buf[100] ;
      
      if (!win32_to_x_font(&lf,buf,100)) return Qnil ;
      
      return Fcons (build_string (buf), Qnil);
  }
  
}
#endif /* not WIN32_FONTSET */


DEFUN ("win32-color-defined-p", Fwin32_color_defined_p, Swin32_color_defined_p, 1, 1, 0,
  "Return non-nil if the X display supports the color named COLOR.")
  (color)
     Lisp_Object color;
{
  Color foo;
  
  CHECK_STRING (color, 0);

  if (defined_color (color, &foo, 0))
    return Qt;
  else
    return Qnil;
}

DEFUN ("win32-color-values", Fwin32_color_values, Swin32_color_values, 1, 1, 0,
  "Return a description of the color named COLOR.\n\
The value is a list of integer RGB values--(RED GREEN BLUE).\n\
These values appear to range from 0 to 65280; white is (65280 65280 65280).")
  (color)
     Lisp_Object color;
{
  Color foo;

  CHECK_STRING (color, 0);

  if (defined_color (color, &foo, 0))
    {
      Lisp_Object rgb[3];

      rgb[0] = make_number (GetRValue(foo));
      rgb[1] = make_number (GetGValue(foo));
      rgb[2] = make_number (GetBValue(foo));
      return Flist (3, rgb);
    }
  else
    return Qnil;
}

DEFUN ("win32-display-color-p", Fwin32_display_color_p, Swin32_display_color_p, 0, 0, 0,
  "Return t if the X screen currently in use supports color.")
  ()
{
    HDC hdc ;
    int cap ;

    hdc = MyGetDC (ROOT_WINDOW) ;

    cap = GetDeviceCaps(hdc,BITSPIXEL) * GetDeviceCaps(hdc,PLANES) ;

    ReleaseDC(ROOT_WINDOW,hdc) ;
    
    return (cap <= 2)? Qnil : Qt ;
}

DEFUN ("win32-display-pixel-width", Fwin32_display_pixel_width, Swin32_display_pixel_width,
       0, 1, 0,
       "Returns the width in pixels of the display FRAME is on.")
    (frame)
    Lisp_Object frame;
{
    HDC hdc ;
    int cap ;
    HWND hwnd ;

    if (!NILP (frame))
	CHECK_LIVE_FRAME (frame, 0);
    
    if (!NILP(frame) && FRAME_WIN32_P (XFRAME (frame)))
	hwnd = FRAME_WIN32_WINDOW (XFRAME (frame)) ;
    else
	hwnd = ROOT_WINDOW ;

    hdc = MyGetDC (hwnd) ;

    cap = GetDeviceCaps(hdc,HORZRES) ;

    ReleaseDC(hwnd,hdc) ;
    
    return make_number (cap);
}

DEFUN ("win32-display-pixel-height", Fwin32_display_pixel_height,
  Swin32_display_pixel_height, 0, 1, 0,
  "Returns the height in pixels of the display FRAME is on.")
  (frame)
     Lisp_Object frame;
{
    HDC hdc ;
    int cap ;
    HWND hwnd ;

    if (!NILP (frame))
	CHECK_LIVE_FRAME (frame, 0);
    
    if (!NILP(frame) && FRAME_WIN32_P (XFRAME (frame)))
	hwnd = FRAME_WIN32_WINDOW (XFRAME (frame)) ;
    else
	hwnd = ROOT_WINDOW ;

    hdc = MyGetDC (hwnd) ;

    cap = GetDeviceCaps(hdc,VERTRES) ;

    ReleaseDC(hwnd,hdc) ;
    
    return make_number (cap);
}

DEFUN ("win32-display-planes", Fwin32_display_planes, Swin32_display_planes,
       0, 1, 0,
       "Returns the number of bitplanes of the display FRAME is on.")
    (frame)
    Lisp_Object frame;
{
    HDC hdc ;
    int cap ;
    HWND hwnd ;

    if (!NILP (frame))
	CHECK_LIVE_FRAME (frame, 0);
    
    if (!NILP(frame) && FRAME_WIN32_P (XFRAME (frame)))
	hwnd = FRAME_WIN32_WINDOW (XFRAME (frame)) ;
    else
	hwnd = ROOT_WINDOW ;

    hdc = MyGetDC (hwnd) ;

    cap = GetDeviceCaps(hdc,PLANES) * GetDeviceCaps(hdc, BITSPIXEL);
    /* 96.4.27 by himi */

    ReleaseDC(hwnd,hdc) ;
    
    return make_number (cap);
}

DEFUN ("win32-display-color-cells", Fwin32_display_color_cells, Swin32_display_color_cells,
       0, 1, 0,
       "Returns the number of color cells of the display FRAME is on.")
    (frame)
    Lisp_Object frame;
{
    HDC hdc ;
    int cap ;
    HWND hwnd ;

    if (!NILP (frame))
	CHECK_LIVE_FRAME (frame, 0);
    
    if (!NILP(frame) && FRAME_WIN32_P (XFRAME (frame)))
	hwnd = FRAME_WIN32_WINDOW (XFRAME (frame)) ;
    else
	hwnd = ROOT_WINDOW ;

    hdc = MyGetDC (hwnd) ;

    cap = GetDeviceCaps(hdc,NUMCOLORS) ;

    ReleaseDC(hwnd,hdc) ;
    
    return make_number (cap);

}

DEFUN ("win32-display-mm-height", Fwin32_display_mm_height, Swin32_display_mm_height, 0, 1, 0,
       "Returns the height in millimeters of the X screen FRAME is on.")
    (frame)
    Lisp_Object frame;
{
    HDC hdc ;
    int cap ;
    HWND hwnd ;

    if (!NILP (frame))
	CHECK_LIVE_FRAME (frame, 0);
    
    if (!NILP(frame) && FRAME_WIN32_P (XFRAME (frame)))
	hwnd = FRAME_WIN32_WINDOW (XFRAME (frame)) ;
    else
	hwnd = ROOT_WINDOW ;

    hdc = MyGetDC (hwnd) ;

    cap = GetDeviceCaps(hdc,VERTSIZE) ;

    ReleaseDC(hwnd,hdc) ;
    
    return make_number (cap);
}

DEFUN ("win32-display-mm-width", Fwin32_display_mm_width, Swin32_display_mm_width, 0, 1, 0,
       "Returns the width in millimeters of the X screen FRAME is on.")
    (frame)
    Lisp_Object frame;
{
    HDC hdc ;
    int cap ;
    HWND hwnd ;

    if (!NILP (frame))
	CHECK_LIVE_FRAME (frame, 0);
    
    if (!NILP(frame) && FRAME_WIN32_P (XFRAME (frame)))
	hwnd = FRAME_WIN32_WINDOW (XFRAME (frame)) ;
    else
	hwnd = ROOT_WINDOW ;

    hdc = MyGetDC (hwnd) ;

    cap = GetDeviceCaps(hdc,HORZSIZE) ;

    ReleaseDC(hwnd,hdc) ;
    
    return make_number (cap);
}


win32_pixel_width (f)
     register struct frame *f;
{
  return PIXEL_WIDTH (f);
}

win32_pixel_height (f)
     register struct frame *f;
{
  return PIXEL_HEIGHT (f);
}

win32_char_width (f)
     register struct frame *f;
{
#ifdef WIN32_FONTSET
  return FRAME_FONT_WIDTH(f);
#else
  return FONT_WIDTH (f->display.win32->font);
#endif
}

win32_char_height (f)
     register struct frame *f;
{
  return f->display.win32->line_height;
}

/* Offset in buffer of character under the pointer, or 0. */
int mouse_buffer_offset;

DEFUN ("win32-open-connection", Fwin32_open_connection, Swin32_open_connection,
       0, 0, 0, "Open a connection to an X server.\n\
DISPLAY is the name of the display to connect to.\n\
Optional second arg XRM_STRING is a string of resources in xrdb format.")
  ()
{
  HDC hdc ;

  /* 95.10.20 Modified by himi */
  win32_init_app(hinst);
    
  /* This is what opens the connection and sets dsp.
     This also initializes many symbols, such as those used for input. */

  win32_term_init ();

  Vwin32_color_map = Fwin32_default_color_map() ;

/*  if (!hMutex) {
      hMutex = CreateMutex (NULL,FALSE,NULL) ;
  } */

#if 0
  hUndoAccel = CreateAcceleratorTable(&undo_accel, 1);
#endif

  hdc = MyGetDC (ROOT_WINDOW) ;
  
  win32_screen_planes = GetDeviceCaps(hdc,PLANES) * GetDeviceCaps(hdc,BITSPIXEL) ;
  win32_screen_height = GetDeviceCaps(hdc,VERTRES) ;
  win32_screen_width  = GetDeviceCaps(hdc,HORZRES) ;
  win32_screen_height_in = GetDeviceCaps(hdc,LOGPIXELSX) ;
  win32_screen_width_in  = GetDeviceCaps(hdc,LOGPIXELSY) ;

  ReleaseDC(ROOT_WINDOW,hdc) ;

  return Qnil;
}

DEFUN ("win32-close-current-connection", Fwin32_close_current_connection,
       Swin32_close_current_connection,
       0, 0, 0, "Close the current connection.")
  ()
{
#if 0
  if (hMutex)
  {
      CloseHandle (hMutex) ;
      hMutex = NULL ;
  }
#endif
#if 0
  DestroyAcceleratorTable(hUndoAccel);
#endif
  return Qnil;
}

syms_of_win32fns ()
{
  /* The section below is built by the lisp expression at the top of the file,
     just above where these variables are declared.  */
  /*&&& init symbols here &&&*/
  Qauto_raise = intern ("auto-raise");
  staticpro (&Qauto_raise);
  Qauto_lower = intern ("auto-lower");
  staticpro (&Qauto_lower);
  Qbackground_color = intern ("background-color");
  staticpro (&Qbackground_color);
  Qbar = intern ("bar");
  staticpro (&Qbar);
  Qborder_color = intern ("border-color");
  staticpro (&Qborder_color);
  Qborder_width = intern ("border-width");
  staticpro (&Qborder_width);
  Qbox = intern ("box");
  staticpro (&Qbox);
  Qcursor_color = intern ("cursor-color");
  staticpro (&Qcursor_color);
  Qcursor_type = intern ("cursor-type");
  staticpro (&Qcursor_type);
  Qfont = intern ("font");
  staticpro (&Qfont);
#ifdef WIN32_FONTSET
  Qfontset = intern ("fontset");
  staticpro (&Qfontset);
  Qfont_property = intern ("font-property");
  staticpro (&Qfont_property);
#endif
  Qforeground_color = intern ("foreground-color");
  staticpro (&Qforeground_color);
  Qgeometry = intern ("geometry");
  staticpro (&Qgeometry);
  Qicon_left = intern ("icon-left");
  staticpro (&Qicon_left);
  Qicon_top = intern ("icon-top");
  staticpro (&Qicon_top);
  Qicon_type = intern ("icon-type");
  staticpro (&Qicon_type);
  Qinternal_border_width = intern ("internal-border-width");
  staticpro (&Qinternal_border_width);
  Qleft = intern ("left");
  staticpro (&Qleft);
  Qmouse_color = intern ("mouse-color");
  staticpro (&Qmouse_color);
  Qnone = intern ("none");
  staticpro (&Qnone);
  Qparent_id = intern ("parent-id");
  staticpro (&Qparent_id);
  Qsuppress_icon = intern ("suppress-icon");
  staticpro (&Qsuppress_icon);
  Qtop = intern ("top");
  staticpro (&Qtop);
  Qundefined_color = intern ("undefined-color");
  staticpro (&Qundefined_color);
  Qvertical_scroll_bars = intern ("vertical-scroll-bars");
  staticpro (&Qvertical_scroll_bars);
  Qvisibility = intern ("visibility");
  staticpro (&Qvisibility);
  Qwindow_id = intern ("window-id");
  staticpro (&Qwindow_id);
  Qwin32_frame_parameter = intern ("win32-frame-parameter");
  staticpro (&Qwin32_frame_parameter);
  Qwin32_resource_name = intern ("win32-resource-name");
  staticpro (&Qwin32_resource_name);
  Quser_position = intern ("user-position");
  staticpro (&Quser_position);
  Quser_size = intern ("user-size");
  staticpro (&Quser_size);
#ifdef MULE			/* 94.6.17 by K.Handa */
  Qline_space = intern ("line-space");
  staticpro (&Qline_space);
#endif  /* MULE */
  /* This is the end of symbol initialization.  */

  Fput (Qundefined_color, Qerror_conditions,
	Fcons (Qundefined_color, Fcons (Qerror, Qnil)));
  Fput (Qundefined_color, Qerror_message,
	build_string ("Undefined color"));

  init_win32_parm_symbols ();

  DEFVAR_INT ("mouse-buffer-offset", &mouse_buffer_offset,
    "The buffer offset of the character under the pointer.");
  mouse_buffer_offset = 0;

  DEFVAR_LISP ("win32-pointer-shape", &Vwin32_pointer_shape,
    "The shape of the pointer when over text.\n\
Changing the value does not affect existing frames\n\
unless you set the mouse color.");
  Vwin32_pointer_shape = Qnil;

  Vwin32_nontext_pointer_shape = Qnil;

  Vwin32_mode_pointer_shape = Qnil;

  DEFVAR_INT ("win32-sensitive-text-pointer-shape",
	      &Vwin32_sensitive_text_pointer_shape,
	      "The shape of the pointer when over mouse-sensitive text.\n\
This variable takes effect when you create a new frame\n\
or when you set the mouse color.");
  Vwin32_sensitive_text_pointer_shape = Qnil;

  DEFVAR_LISP ("win32-cursor-fore-pixel", &Vwin32_cursor_fore_pixel,
	       "A string indicating the foreground color of the cursor box.");
  Vwin32_cursor_fore_pixel = Qnil;

  DEFVAR_LISP ("mouse-grabbed", &Vmouse_depressed,
	       "Non-nil if a mouse button is currently depressed.");
  Vmouse_depressed = Qnil;

  DEFVAR_LISP ("win32-no-window-manager", &Vwin32_no_window_manager,
	       "Non-nil if no X window manager is in use.");

  DEFVAR_LISP ("win32-color-map", &Vwin32_color_map,
	       "A array of color name mappings for windows.");
  Vwin32_color_map = Qnil;

  defsubr (&Swin32_get_resource);
  defsubr (&Swin32_display_color_p);

#ifndef WIN32_FONTSET
  defsubr (&Swin32_list_fonts);
  defsubr (&Swin32_find_font);
#endif

  defsubr (&Swin32_color_defined_p);
  defsubr (&Swin32_color_values);
  defsubr (&Swin32_display_pixel_width);
  defsubr (&Swin32_display_pixel_height);
  defsubr (&Swin32_display_mm_width);
  defsubr (&Swin32_display_mm_height);
  defsubr (&Swin32_display_planes);
  defsubr (&Swin32_display_color_cells);
  defsubr (&Swin32_create_frame);
  defsubr (&Sfocus_frame);
  defsubr (&Sunfocus_frame);
  defsubr (&Swin32_open_connection);
  defsubr (&Swin32_close_current_connection);
#if 0
  defsubr (&Swin32_rgb);
#else
  defsubr (&Swin32_default_color_map);
#endif
}

#undef abort

void win32_abort()
{
    MessageBox (NULL,
		"A fatal error has occurred - aborting!",
		"Emacs Abort Dialog",
		MB_OK|MB_ICONEXCLAMATION) ;
    RaiseException(100, EXCEPTION_NONCONTINUABLE, 0, NULL);
    abort() ;
}

