/***************************************
  Win32 Window System Fontset Manager
        Written by himi
***************************************/
#include <config.h>
#include "lisp.h"
#include "win32term.h"
#include "win32fontset.h"
#include "dispextern.h"
#include "coding.h"

int fontset_num;
Lisp_Object Vwin32_system_coding_system;
static win32_fontset **win32_fontset_table = 0;

LPLOGFONT get_logfont(win32_fontset *fs, fs_fonts *sf, int property)
{
  LOGFONT *lf;

  lf = FS_GET_LOGFONT(sf, property);
  if (!lf) {
    switch (property) {
    case 0:
      lf = &default_normal_font;
      break;
    case 1:
      lf = &default_bold_font;
      break;
    case 2:
      lf = &default_italic_font;
      break;
    case 3:
      lf = &default_italic_bold_font;
      break;
    default :
      return (LPLOGFONT) NULL;
    }
    lf->lfWidth = fs -> hmetric;
    lf->lfHeight = fs -> vmetric;
  }
  return lf;
}

HFONT get_hfont(win32_fontset *fs, fs_fonts *sf, int property)
{
  LOGFONT *lf;
  HFONT hf;

  lf = get_logfont(fs, sf, property);
  hf = CreateFontIndirect(lf);
  FS_GET_FONT(sf, property) = hf;
  return hf;
}

FONT_TYPE *get_font_type(unsigned char *name,
			 fs_fonts* sf, int property)
{
  FONT_TYPE *result;
  int nbytes;

  result = (FONT_TYPE *)xmalloc(sizeof(FONT_TYPE));
  nbytes = strlen(name) + 1;
  result -> name = (unsigned char *) xmalloc(nbytes);
  memcpy(result -> name, name, nbytes);
  result -> phfont = &FS_GET_FONT(sf, property);
  result -> property_index = property;
  result -> fs_font = sf;

  return result;
}

FONT_TYPE *win32_load_font(struct frame *f, unsigned char *name, int property)
{
  fs_fonts **sf;
  LOGFONT *lf;
  HFONT hf;
  int i;

  sf = FRAME_FONTSET(f) -> fonts;
  for (i = 0;i < (FRAME_FONTSET(f) -> font_num);i++){
    if (strcmp((*sf) -> name, name) == 0) {
      hf = FS_GET_FONT(*sf, property);
      if (hf) return get_font_type(name, *sf, property);
      hf = get_hfont(FRAME_FONTSET(f), *sf, property);
      if (!hf) return (FONT_TYPE *) NULL;
      return get_font_type(name, *sf, property);
    }
    sf++;
  }
  return (FONT_TYPE *) NULL;
}

void win32_unload_font(struct frame *f, FONT_TYPE *font)
{
  HFONT *phf;

  phf =  &FS_GET_FONT(font->fs_font, font -> property_index);

  if (!(*phf)) return;
/*  DeleteObject(*phf);
  *phf = NULL; */
  xfree(font->name);
  xfree(font);

  return;
}

win32_fontset *win32_get_fontset(unsigned char *name)
{
  int i;

  for (i = 0;i < fontset_num; i++) {
    if (strcmp(win32_fontset_table[i] -> fs_name, name) == 0) {
      return win32_fontset_table[i];
    }
  }
  return 0;
}

void expand_fontset_table(int num)
{
  int i;
  win32_fontset *fs;
  win32_fontset_table = (win32_fontset **) 
    xrealloc(win32_fontset_table, 
	     (fontset_num + num) * sizeof(win32_fontset *));
  fs = (win32_fontset *) xmalloc(num * sizeof(win32_fontset));
  for (i = 0;i < num;fontset_num++, fs++, i++) {
    win32_fontset_table[fontset_num] = fs;
  }
}

void internal_add_fontset(int location,
			  const unsigned char *name,
			  int width, int height, int base, int overhang)
{
  win32_fontset *tp;
  tp = win32_fontset_table[location];
  tp -> fs_name = (unsigned char *) xmalloc(strlen(name) + 1);
  strcpy(tp -> fs_name, name);
  tp -> vmetric = height;
  tp -> hmetric = width;
  tp -> font_base = base;
  tp -> overhang = overhang;
  tp -> font_num = 0;
  tp -> fonts = 0;
}

void expand_font_table(win32_fontset *fontset, int num)
{
  int i;
  fs_fonts *fs;
  fontset->fonts = (fs_fonts **)
    xrealloc(fontset->fonts,
	     (fontset->font_num + num) * sizeof(fs_fonts*));
  fs = (fs_fonts *) xmalloc(num * sizeof(fs_fonts));
  for (i = 0;i < num;fontset->font_num++, fs++, i++) {
    fontset->fonts[fontset->font_num] = fs;
  }
}

void internal_add_font(win32_fontset *fontset, int location,
		       const unsigned char *name, 
		       Lisp_Object coding_system)
{
  fs_fonts *fp;
  int i;
  Fcheck_code(coding_system);
  fp = (fontset -> fonts)[location];
  fp -> name = (unsigned char *)xmalloc(strlen(name) + 1);
  strcpy(fp -> name, name);
  fp -> coding_system = coding_system;
  for (i = 0;i < FONT_PROPERTY_NUM;i++){
    fp -> font_properties[i] = (HFONT) 0;
    fp -> logfont_properties[i] = (LPLOGFONT) 0;
  }
}

Lisp_Object logfont_to_lisp_object(LOGFONT *lf)
{
  Lisp_Object logfont[9];
  /* [0] is name.
     [1] is width. 
     [2] is height. 
     [3] is weight.
     [4] is orientation.
     [5] is italic.
     [6] is strikeout.
     [7] is charset.
     [8] is quality. 
     [9] is OutPrecision 
     [10] is PitchAndFamily */

  Lisp_Object found_code;
  unsigned char *buf;
  int nbytes, size;
  coding_type mccode;

  nbytes = strlen (lf->lfFaceName);
  encode_code (Fcheck_code (Vwin32_system_coding_system), &mccode);
  CODE_CNTL (&mccode) |= CC_END;
  buf = (unsigned char *)xmalloc (ENCODE_BUF_SIZE (nbytes, &mccode));
  size = encode (&mccode, lf->lfFaceName, buf, nbytes, &found_code);
  logfont[0] = make_string(buf, size);
  xfree(buf);

  logfont[1] = make_number(lf->lfWidth);
  logfont[2] = make_number(lf->lfHeight);
  XFASTINT(logfont[3]) = lf->lfWeight;
  XFASTINT(logfont[4]) = lf->lfOrientation;
  logfont[5] = (lf->lfItalic) ? Qt:Qnil;
  logfont[6] = (lf->lfStrikeOut) ? Qt:Qnil;
  XFASTINT(logfont[7]) = lf->lfCharSet;
  XFASTINT(logfont[8]) = lf->lfQuality;
  XFASTINT(logfont[9]) = lf->lfOutPrecision;
  XFASTINT(logfont[10]) = lf->lfPitchAndFamily;

  return Flist(XFASTINT(11), logfont);
}

LOGFONT lisp_object_to_logfont(Lisp_Object font)
{
  coding_type mccode;
  LOGFONT lf;
  Lisp_Object tmpcdr, tmpcar;
  unsigned char *buf;
  int size;

  tmpcar = Fcar(font);
  encode_code(Fcheck_code(Vwin32_system_coding_system), &mccode);

  buf = alloca(DECODE_BUF_SIZE(XSTRING(tmpcar) -> size +1, &mccode));

  /* LOGFONT is NEVER accessed by lisp interpreter. */
  CODE_CNTL (&mccode) |= CC_END;
  size = decode(&mccode, XSTRING(tmpcar) -> data, buf,
		XSTRING(tmpcar) -> size);
  if (size >= LOGFONT_FACENAME_MAX) {
    error ("Font name is too long. ");
   }
  memcpy(lf.lfFaceName, buf, size);
  lf.lfFaceName[size] = (TCHAR) NULL;

  tmpcdr = Fcdr(font);
  tmpcar = Fcar(tmpcdr);
  lf.lfWidth = XINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfHeight = XINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfWeight = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfOrientation = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfItalic = !NILP(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfStrikeOut = !NILP(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfCharSet = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfQuality = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfOutPrecision = XFASTINT(tmpcar);
  tmpcdr = Fcdr(tmpcdr);
  tmpcar = Fcar(tmpcdr);
  lf.lfPitchAndFamily = XFASTINT(tmpcar);

/* Set the default value... */
  lf.lfEscapement = lf.lfOrientation;
  lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
  lf.lfUnderline = 0;

  return lf;
}

DEFUN("win32-check-font", Fwin32_check_font, Swin32_check_font, 1, 1, 0,
      "Check validity of WIN32-FONT. \n\
The validity of WIN32-FONT is it is a list that have 11 elements.\n\
(name width height weight orientation italic strikeout charset quality\n\
OutPrecision PitchAndFamily)")
(font)
     Lisp_Object font;
/* The width must be a string that describes the name of its font.\n\
The height must be a number that describes the width of its font.\n\
The weight must be a number that describes the weight.(1-1000)\n\
The orientation must be a number that describes the orientation.(angle)\n\
If the italic is nil, it is not italic font.\n\
If the strikeout is nil, it doesn't have a strikeout line.\n\
The charset must be a number that describes the charset that it uses.\n\
(128 is SJIS, 0 is ANSI)\n\
The quality must be a number that describes its quality.(0-2) */
{
  Lisp_Object tmpcar, tmpcdr;
  CHECK_LIST(font, 0);
  tmpcar = Fcar(font);
  tmpcdr = Fcdr(font);
  CHECK_STRING(tmpcar, 1); /* name */
  if ((XSTRING(tmpcar)->size) > LOGFONT_FACENAME_MAX) {
    error("The name is too long. Max is %d", LOGFONT_FACENAME_MAX);
  }
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar,2); /* width */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar,3); /* height */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 4); /* weight */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 5); /* orientation */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  if (!((EQ(tmpcar, Qnil)) || (EQ(tmpcar, Qt)))) { /* italic */
    error("italic cell must be nil or t");
  }
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  if (!((EQ(tmpcar, Qnil)) || (EQ(tmpcar, Qt)))) { /* strikeout */
    error("strikeout cell must be nil or t");
  }
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 6); /* charset */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 7); /* quality */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 8); /* OutPrecision */
  tmpcar = Fcar(tmpcdr);
  tmpcdr = Fcdr(tmpcdr);
  CHECK_NUMBER(tmpcar, 9); /* PitchAndFamily */
  if (!NILP(tmpcdr)) {
    error ("invalid font. its length must be 11.");
  }
  return Qt;
}

void win32_load_font_tmp(win32_fontset *fs, FONT_TYPE *font, int forcedef)
{
  fs_fonts **sf;
  LOGFONT *lf;
  HFONT hf;
  int i, flag;

  sf = fs -> fonts;
  for (i = 0;i < (fs -> font_num);i++){
    if (forcedef) flag = strcmp((*sf) -> name, default_font);
    else flag = strcmp((*sf) -> name, font -> name);

    if (flag == 0) {
      font -> fs_font = *sf;
      hf = FS_GET_FONT(*sf, font -> property_index);
      if (hf) {
	font -> phfont = &FS_GET_FONT(*sf, font -> property_index);
	return;
      }
      hf = get_hfont(fs, *sf, font -> property_index);
      if (!hf) {
	error("Unable to change fontset. Fonts in the fontset you specified are not found.");
      }
      font -> phfont = &FS_GET_FONT(*sf, font -> property_index);
      return;
    }
    sf++;
  }

 error:
#if 0
  error("Fail to change fontset!. There is no font that is in the previous fontset in the specified fontset.");
#else
  if (!forcedef)
    win32_load_font_tmp(fs, font, 1);
  else
  error("Fail to change fontset!. There is no font that is in the previous fontset in the specified fontset. Furturemore, there is no default font!.");
#endif
}

void win32_change_fontset(struct frame *f, unsigned char *name)
{

  win32_fontset *fs;
  int i;
  HWND hwnd;

  fs = win32_get_fontset(name);
  if (!fs) {
    error ("Such fontset does not exist!");
  }

  /* display.win32 member is initialized by bzero */
  if (f->display.win32->font)
    win32_load_font_tmp(fs, f->display.win32->font, 0);
  for (i = 0; i < FRAME_N_PARAM_FACES(f); i++)
    if (FRAME_PARAM_FACES(f)[i] != 0
	&& FRAME_PARAM_FACES(f)[i]->font != (FONT_TYPE *) FACE_DEFAULT)
      {
/*	win32_unload_font_tmp(FRAME_FONTSET(f),
			      f->display.win32->param_faces[i]->font); */
	win32_load_font_tmp(fs, FRAME_PARAM_FACES(f)[i]->font, 0);
      }
  FRAME_FONTSET(f) = fs;
  frame_update_line_height (f);
  win32_set_window_size (f, 0, f->width, f->height);
  hwnd = FRAME_WIN32_WINDOW(f);
  if (hwnd){
    win32_set_ime_font(hwnd, 
		       get_logfont(fs,
				   f->display.win32->font->fs_font, 
				   f->display.win32->font->
				   property_index));
  }
}

/* fontset manager */
DEFUN("win32-add-fontset", Fwin32_add_fontset, Swin32_add_fontset, 5, 5, 0,
      "Add fontset. First ARG is a name of fontset.\n\
Second ARG, Third ARG, Forth ARG, and Fifth ARG are \n\
metric infomation, width, height, base, and overhang.")
(name, width, height, base, overhang)
     Lisp_Object name, width, height, base, overhang;
{
  CHECK_STRING(name, 0);
  CHECK_NUMBER(width, 1);
  CHECK_NUMBER(height, 1);
  CHECK_NUMBER(base, 1);
  CHECK_NUMBER(overhang, 1);
  expand_fontset_table(1);
  internal_add_fontset(fontset_num - 1, XSTRING(name) -> data, 
		       width, height, base, overhang);
  return Qnil;
}

DEFUN("win32-fontset-list", Fwin32_fontset_list, Swin32_fontset_list, 0, 0, 0,
      "Return a list of all win32 fontset")
()
{
  Lisp_Object val = Qnil;
  int i;

  for (i = 0;i < fontset_num;i++) {
    val = Fcons(build_string(win32_fontset_table[i] -> fs_name), val);
  }
  return val;
}

DEFUN("win32-refresh-fontset", Fwin32_refresh_fontset,
      Swin32_refresh_fontset, 1, 1, 0,
      "Refresh the fontset of the frame.\n\
FRAME shows the frame you want to refresh fontset.")
(frame)
     Lisp_Object frame;
{
  CHECK_FRAME(frame, 0);

  win32_change_fontset(XFRAME(frame),
		       FRAME_FONTSET(XFRAME(frame)) -> fs_name);
  return Qnil;
}

DEFUN("win32-change-fontset-attribute", Fwin32_change_fontset_attribute,
      Swin32_change_fontset_attribute, 4, 5, 0,
      "Change the fontset attribute.\n\
The first arg WIDTH is the width of the fontset. The second arg HEIGHT is\n\
the height of it.  The third arg BASE is the base length. The forth arg \n\
OVERHANG is the overhang. The Fifth optional arg NAME is the name of the \n\
fontset you want to change. If it is nil, the fontset of the current frame\n\
is changed.")
(width, height, base, overhang, name)
     Lisp_Object width, height, base, overhang, name;
{
  win32_fontset *fs;

  CHECK_NUMBER(width, 0);
  CHECK_NUMBER(height, 1);
  CHECK_NUMBER(overhang, 2);
  CHECK_NUMBER(base, 3);

  if (NILP(name)) {
    fs = FRAME_FONTSET(selected_frame);
  }else{
    CHECK_STRING(name, 4);
    fs = win32_get_fontset(XSTRING(name)->data);
    if (!fs) {
      error("Such fontset does NOT exist!.");
    }
  }
  fs->hmetric = XFASTINT(width);
  fs->vmetric = XFASTINT(height);
  fs->overhang = XFASTINT(overhang);
  fs->font_base = XFASTINT(base);

  return Qnil;
}

/* font manager */
DEFUN("win32-add-font", Fwin32_add_font, Swin32_add_font, 2, 3, 0,
      "Add font. The first ARG is a name of font.\n\
The second ARG is the coding_system that is used with its font.\n\
The third optional arg is a name of the fontset that this font is added to.")
(name, coding_system, fontset)
{
  win32_fontset *fs;
  int font_location;

  CHECK_STRING(name, 0);

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 1);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }
  font_location = fs -> font_num;
  expand_font_table(fs, 1);
  internal_add_font(fs, font_location,
		    XSTRING(name) -> data, coding_system);

  return Qnil;
}

DEFUN("win32-font-list", Fwin32_font_list, Swin32_font_list, 0, 1, 0,
      "Return a list of all win32 font. \n\
The optional arg is the fontset where you want to get fonts.")
(fontset)
{
  win32_fontset *fs;
  int i;
  Lisp_Object val = Qnil;

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 0);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }
  for(i = 0;i < fs->font_num;i++) {
    val = Fcons(build_string((fs -> fonts)[i] -> name), val);
  }
  return val;
}
  
/* font property manager */
DEFUN("win32-change-font-property", Fwin32_change_font_property,
      Swin32_change_font_property, 3, 4, 0,
      "Change font property. \n\
The first arg is a name of the font where the property is changed.\n\
The second ARG is a number of the property.\n\
The third ARG is the logfont infomation.\n\
The forth optional arg is a name of the fontset \n\
where this property is changed.")
(font, number, logfont, fontset)
{
  win32_fontset *fs;
  fs_fonts **sf;
  LPLOGFONT plf;
  HFONT *phf;
  
  int i, prop_num;
  CHECK_STRING(font, 0);
  CHECK_NUMBER(number, 1);
  prop_num = XFASTINT(number);
  if (prop_num >= FONT_PROPERTY_NUM) {
    error("Too large font property number.");
  }

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 2);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }
  Fwin32_check_font(logfont);

  sf = fs -> fonts;
  for (i = 0;i < fs -> font_num; sf++, i++){
    if (strcmp((*sf)->name, XSTRING(font)->data) == 0) {
      plf = FS_GET_LOGFONT(*sf, XFASTINT(number));
      if (!plf) {
	plf = (LPLOGFONT)xmalloc(sizeof(LOGFONT));
	FS_GET_LOGFONT(*sf, XFASTINT(number)) = plf;
      }
      *plf = lisp_object_to_logfont(logfont);
      
      /* to delete old font handle */
      phf =  &FS_GET_FONT(*sf, XFASTINT(number));
      if (!(*phf)) return Qnil;
      DeleteObject(*phf);
      *phf = NULL; 
      return Qnil;
    }
  }
  error ("Such font does not exist!");
}

DEFUN ("win32-get-font-property", Fwin32_get_font_property, 
       Swin32_get_font_property, 2, 3, 0,
       "Get the property of the font you specified.\n\
The first arg is a name of the font.\n\
The second arg is a number of the property.\n\
The third optional arg is a fontset of the font.")
(font, number, fontset)
{
  win32_fontset *fs;
  fs_fonts **sf;
  int i, prop_num;
  LOGFONT *lf;

  CHECK_STRING(font, 0);
  CHECK_NUMBER(number, 1);
  prop_num = XFASTINT(number);
  if (prop_num >= FONT_PROPERTY_NUM) {
    error("Too large font property number.");
  }

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 2);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }

  sf = fs -> fonts;
  for (i = 0;i < fs -> font_num; sf++, i++){
    if (strcmp((*sf) -> name, XSTRING(font)->data) == 0) {
      if (lf = FS_GET_LOGFONT(*sf, XFASTINT(number))) {
	return logfont_to_lisp_object(lf);
      }
      return Qnil;
    }
  }
  error ("Such font does not exist!");
}

DEFUN ("win32-get-font-metric", Fwin32_get_font_metric, 
       Swin32_get_font_metric, 2, 3, 0,
       "Get the metric of the font you specified.\n\
The first arg is a name of the font.\n\
The second arg is a number of the property.\n\
The third optional arg is a fontset of the font.\n\
The returned metric infomation is as follows.\n\
(width, height, base, overhang)")
(font, number, fontset)
{
  win32_fontset *fs;
  fs_fonts **sf;
  int i, prop_num;
  HFONT hf;

  CHECK_STRING(font, 0);
  CHECK_NUMBER(number, 1);
  prop_num = XFASTINT(number);
  if (prop_num >= FONT_PROPERTY_NUM) {
    error("Too large font property number.");
  }

  if (NILP(fontset)){
    fs = CURRENT_FONTSET;
  }else{
    CHECK_STRING(fontset, 2);
    fs = win32_get_fontset(XSTRING(fontset) -> data);
    if (!fs) {
      error ("Such fontset does not exist!");
    }
  }

  sf = fs -> fonts;
  for (i = 0;i < fs -> font_num; sf++, i++){
    if (strcmp((*sf) -> name, XSTRING(font)->data) == 0) {
      if (hf = get_hfont(fs, *sf, XFASTINT(number))) {
	HDC hdc;
	HANDLE oldobj;
	TEXTMETRIC tm;

	hdc = MyGetDC(FRAME_WIN32_WINDOW(selected_frame));
	oldobj = SelectObject(hdc, hf);
	if (!GetTextMetrics(hdc, &tm)) {
	  error("Fail to get metrics!!");
	}
	SelectObject(hdc, oldobj);
	ReleaseDC(FRAME_WIN32_WINDOW(selected_frame), hdc);
	return Fcons(make_number((int)tm.tmAveCharWidth), 
		     Fcons(make_number((int)tm.tmHeight),
			   Fcons(make_number((int)tm.tmAscent),
				 Fcons(make_number((int)tm.tmOverhang), 
				       Qnil))));
      }
      return Qnil;
    }
  }
  error ("Such font does not exist!");
}

DEFUN ("win32-query-get-logfont", Fwin32_query_get_logfont,
       Swin32_query_get_logfont, 0, 0, 0,
       "Return the font that you selected with Font Select Dialog Box.")
     ()
{
  CHOOSEFONT cf;
  LOGFONT lf;
  unsigned char buf[100];

  cf.lStructSize = sizeof(CHOOSEFONT);
  cf.hwndOwner = FRAME_WIN32_WINDOW(selected_frame);
  cf.hDC = (HDC)NULL;
  cf.lpLogFont = &lf;
  cf.Flags = CF_SCREENFONTS | CF_FIXEDPITCHONLY | 
    CF_FORCEFONTEXIST | CF_EFFECTS;
  cf.rgbColors = RGB(0,0,0);
  cf.lpfnHook = (LPCFHOOKPROC)NULL;
  cf.lpTemplateName = (LPSTR)NULL;
  cf.hInstance = (HINSTANCE) NULL;
  cf.lpszStyle = (LPSTR)NULL;
  cf.nFontType = SCREEN_FONTTYPE;

  ChooseFont(&cf);
  return logfont_to_lisp_object(&lf);
}

/* syms */

syms_of_win32fontset()
{
  DEFVAR_LISP("win32-system-coding-system",
	      &Vwin32_system_coding_system,
	      "coding system in win32 system");

  defsubr(&Swin32_check_font);
  defsubr(&Swin32_change_font_property);
  defsubr(&Swin32_add_font);
  defsubr(&Swin32_add_fontset);
  defsubr(&Swin32_refresh_fontset);
  defsubr(&Swin32_change_fontset_attribute);
  defsubr(&Swin32_check_font);
  defsubr(&Swin32_fontset_list);
  defsubr(&Swin32_font_list);
  defsubr(&Swin32_get_font_property);
  defsubr(&Swin32_get_font_metric);
  defsubr(&Swin32_query_get_logfont);
/*
   defsubr(&Swin32_delete_fontset);
   defsubr(&Swin32_delete_font);
   defsubr(&Swin32_get_fontset_info);
   defsubr(&Swin32_get_font_info);
*/
}
