/* Definitions and headers for communication with WIN32
   Copyright (C) 1989, 1993, 1994 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef __WIN32TERM__
#define __WIN32TERM__
#include <windows.h>
#ifdef WIN32_FONTSET
#include "win32fontset.h"
#endif
#include "frame.h"

/* Define a queue for X-events.  One such queue is used for mouse clicks.
   Another is used for expose events.  */

#define EVENT_BUFFER_SIZE 64

/* Max and Min sizes in character columns. */
#define MINWIDTH 10
#define MINHEIGHT 10
#define MAXWIDTH 300
#define MAXHEIGHT 80

typedef struct win32_font_t
{
#ifdef WIN32_FONTSET
  unsigned char *name;
  fs_fonts *fs_font;
  int property_index;
  HFONT *phfont;
#else
  TEXTMETRIC tm ;
  HFONT hfont ;
#endif
} win32_font_t ;

#define Color COLORREF

#define PIX_TYPE Color
#define BLACK_PIX_DEFAULT RGB(0,0,0)
#define WHITE_PIX_DEFAULT RGB(255,255,255)
//#define DISPLAY_CELLS DisplayCells (x_current_display, XDefaultScreen (x_current_display))
#define ROOT_WINDOW GetDesktopWindow()
#define FONT_TYPE win32_font_t

#define Bitmap HBITMAP

#ifdef MULE
#ifdef WIN32_FONTSET
#define FONTSET_WIDTH(fs) ((fs) -> hmetric)
#define FONTSET_HEIGHT(fs) ((fs) -> vmetric)
#define FONTSET_OVERHANG(fs) ((fs) -> overhang)
#define FONTSET_BASE(fs) ((fs) -> font_base)
#define FRAME_FONTSET(f) ((f) -> win32_fs)
#define FONT_HFONT(font, fs) ((*((font)->phfont)!=0)?*((font)->phfont):\
get_hfont((fs), (font)->fs_font, (font)-> property_index))
#define FRAME_FONT_WIDTH(f) (FONTSET_WIDTH(FRAME_FONTSET(f)))
#define FRAME_FONT_HEIGHT(f) (FONTSET_HEIGHT(FRAME_FONTSET(f)))
#define FRAME_FONT_OVERHANG(f) (FONTSET_OVERHANG(FRAME_FONTSET(f)))
#define FRAME_FONT_BASE(f) (FONTSET_BASE(FRAME_FONTSET(f)))
#endif
#define FONT_WIDTH(f)	((f)->tm.tmAveCharWidth) 
#define FONT_HEIGHT(f)	((f)->tm.tmHeight)
#define FONT_OVERHANG(f) ((f)->tm.tmOverhang) /* 96.3.22 by himi */
#else
#define FONT_WIDTH(f)	((f)->tm.tmMaxCharWidth)
#define FONT_HEIGHT(f)	((f)->tm.tmHeight)
#endif
#define FONT_BASE(f)    ((f)->tm.tmAscent)

typedef DWORD Time ;

struct event_queue
  {
    int rindex;			/* Index at which to fetch next. */
    int windex;			/* Index at which to store next.  */
    MSG win32rep[EVENT_BUFFER_SIZE];
  };

/* Queue for mouse clicks.  */
extern struct event_queue win32_mouse_queue;

/* This checks to make sure we have a display.  */
extern void check_win32 ();

extern struct frame *win32_window_to_frame ();

/* The frame (if any) which has the X window that has keyboard focus.
   Zero if none.  This is examined by Ffocus_frame in xfns.c */

extern struct frame *win32_focus_frame;

/* Height of this X screen in pixels.  */
extern int win32_screen_height;

/* Height of this X screen in millimeters.  */
extern int win32_screen_height_mm;

/* Width of this X screen in pixels.  */
extern int win32_screen_width;

/* Width of this X screen in millimeters.  */
extern int win32_screen_width_mm;

/* Number of planes for this screen.  */
extern int win32_screen_planes;

/* Mask of which mouse buttons are currently held down.  */
extern unsigned int win32_mouse_grabbed;


enum text_cursor_kinds {
  filled_box_cursor, hollow_box_cursor, bar_cursor
};

/* Each Win32 frame object points to its own struct win32_display object
   in the display.win32 field.  The win32_display structure contains all
   the information that is specific to win32.  */

struct win32_display
{
#ifdef TERMINAL_FACE
  struct f_display f;
#endif

  /* Position of the X window (x and y offsets in root window).  */
  int left_pos;
  int top_pos;

  /* Border width of the X window as known by the X window system.  */
  int border_width;

  /* Size of the X window in pixels.  */
  int pixel_height, pixel_width;

  /* Height of a line, in pixels.  */
  int line_height;

#ifdef MULE
  /* Upper space of each line, in pixels. */
  int upper_space;

  /* Lower space of each line, in pixels. */
  int lower_space;
#endif  /* MULE */

  /* Width of the internal border.  This is a line of background color
     just inside the window's border.  When the frame is selected,
     a highlighting is displayed inside the internal border.  */
  int internal_border_width;

  /* The X window used for this frame.
     May be zero while the frame object is being created
     and the X window has not yet been created.  */
  HWND window_desc;

  /* The X window used for the bitmap icon;
     or 0 if we don't have a bitmap icon.  */
  HWND icon_desc;

  /* The X window that is the parent of this X window.
     Usually but not always RootWindow.  */
  HWND parent_desc;

  FONT_TYPE *font;

  /* Pixel values used for various purposes.
     border_pixel may be -1 meaning use a gray tile.  */
  PIX_TYPE background_pixel;
  PIX_TYPE foreground_pixel;
  PIX_TYPE cursor_pixel;
  PIX_TYPE border_pixel;
  PIX_TYPE mouse_pixel;
  PIX_TYPE cursor_foreground_pixel;

  /* Descriptor for the cursor in use for this window.  */
  HCURSOR text_cursor;
  HCURSOR nontext_cursor;
  HCURSOR modeline_cursor;
  HCURSOR cross_cursor;

  /* The name that was associated with the icon, the last time
     it was refreshed.  Usually the same as the name of the
     buffer in the currently selected window in the frame */
  char *icon_label;

  /* Flag to set when the X window needs to be completely repainted.  */
  int needs_exposure;

  /* What kind of text cursor is drawn in this window right now?
     (If there is no cursor (phys_cursor_x < 0), then this means nothing.)  */
  enum text_cursor_kinds current_cursor;

  /* What kind of text cursor should we draw in the future?
     This should always be filled_box_cursor or bar_cursor.  */
  enum text_cursor_kinds desired_cursor;

  /* The window style flags */

  DWORD dwStyle ;

  /* The size of the extra width currently allotted for vertical
     scroll bars, in pixels.  */
  int vertical_scroll_bar_extra;

  /* Table of parameter faces for this frame.  Any X resources (pixel
     values, fonts) referred to here have been allocated explicitly
     for this face, and should be freed if we change the face.  */
  struct face **param_faces;
  int n_param_faces;

#ifndef TERMINAL_FACE
  /* Table of computed faces for this frame.  These are the faces
     whose indexes go into the upper bits of a glyph, computed by
     combining the parameter faces specified by overlays, text
     properties, and what have you.  The X resources mentioned here
     are all shared with parameter faces.  */
  struct face **computed_faces;
  int n_computed_faces;		/* How many are valid */
  int size_computed_faces;	/* How many are allocated */
#endif
};

#ifndef TERMINAL_FACE
/* Get at the computed faces of an Win32 frame.  */
#define FRAME_PARAM_FACES(f) ((f)->display.win32->param_faces)
#define FRAME_N_PARAM_FACES(f) ((f)->display.win32->n_param_faces)
#define FRAME_DEFAULT_PARAM_FACE(f) (FRAME_PARAM_FACES (f)[0])
#define FRAME_MODE_LINE_PARAM_FACE(f) (FRAME_PARAM_FACES (f)[1])

#define FRAME_COMPUTED_FACES(f) ((f)->display.win32->computed_faces)
#define FRAME_N_COMPUTED_FACES(f) ((f)->display.win32->n_computed_faces)
#define FRAME_SIZE_COMPUTED_FACES(f) ((f)->display.win32->size_computed_faces)
#define FRAME_DEFAULT_FACE(f) ((f)->display.win32->computed_faces[0])
#define FRAME_MODE_LINE_FACE(f) ((f)->display.win32->computed_faces[1])
#endif

/* Return the window associated with the frame F.  */
#define FRAME_WIN32_WINDOW(f) ((f)->display.win32->window_desc)

/* These two really ought to be called FRAME_PIXEL_{WIDTH,HEIGHT}.  */
#define PIXEL_WIDTH(f) ((f)->display.win32->pixel_width)
#define PIXEL_HEIGHT(f) ((f)->display.win32->pixel_height)

#define FRAME_DESIRED_CURSOR(f) ((f)->display.win32->desired_cursor)


/* When X windows are used, a glyph may be a 16 bit unsigned datum.
   The high order byte is the face number and is used as an index
   in the face table.  A face is a font plus:
    1) the unhighlighted foreground color,
    2) the unhighlighted background color.
   For highlighting, the two colors are exchanged.
   Face number 0 is unused.  The low order byte of a glyph gives
   the character within the font.  All fonts are assumed to be
   fixed width, and to have the same height and width.  */

#define MAX_FACES_AND_GLYPHS 256
extern struct face *win32_face_table[];


/* X-specific scroll bar stuff.  */

/* We represent scroll bars as lisp vectors.  This allows us to place
   references to them in windows without worrying about whether we'll
   end up with windows referring to dead scroll bars; the garbage
   collector will free it when its time comes.

   We use struct scroll_bar as a template for accessing fields of the
   vector.  */

struct scroll_bar {

  /* These fields are shared by all vectors.  */
  int size_from_Lisp_Vector_struct;
  struct Lisp_Vector *next_from_Lisp_Vector_struct;

  /* The window we're a scroll bar for.  */
  Lisp_Object window;

  /* The next and previous in the chain of scroll bars in this frame.  */
  Lisp_Object next, prev;

  /* The X window representing this scroll bar.  Since this is a full
     32-bit quantity, we store it split into two 32-bit values.  */
  Lisp_Object win32_window_low, win32_window_high;

  /* The position and size of the scroll bar in pixels, relative to the
     frame.  */
  Lisp_Object top, left, width, height;

  /* The starting and ending positions of the handle, relative to the
     handle area (i.e. zero is the top position, not
     SCROLL_BAR_TOP_BORDER).  If they're equal, that means the handle
     hasn't been drawn yet.

     These are not actually the locations where the beginning and end
     are drawn; in order to keep handles from becoming invisible when
     editing large files, we establish a minimum height by always
     drawing handle bottoms VERTICAL_SCROLL_BAR_MIN_HANDLE pixels below
     where they would be normally; the bottom and top are in a
     different co-ordinate system.  */
  Lisp_Object start, end;

  /* If the scroll bar handle is currently being dragged by the user,
     this is the number of pixels from the top of the handle to the
     place where the user grabbed it.  If the handle isn't currently
     being dragged, this is Qnil.  */
  Lisp_Object dragging;
};

/* The number of elements a vector holding a struct scroll_bar needs.  */
#define SCROLL_BAR_VEC_SIZE \
  ((sizeof (struct scroll_bar) - sizeof (int) - sizeof (struct Lisp_Vector *)) \
   / sizeof (Lisp_Object))

/* Turning a lisp vector value into a pointer to a struct scroll_bar.  */
#define XSCROLL_BAR(vec) ((struct scroll_bar *) XPNTR (vec))


/* Building a 32-bit C integer from two 16-bit lisp integers.  */
#define SCROLL_BAR_PACK(low, high) (XINT (high) << 16 | XINT (low))

/* Setting two lisp integers to the low and high words of a 32-bit C int.  */
#define SCROLL_BAR_UNPACK(low, high, int32) \
  (XSET ((low),  Lisp_Int,  (int32)        & 0xffff), \
   XSET ((high), Lisp_Int, ((int32) >> 16) & 0xffff))


/* Extract the X window id of the scroll bar from a struct scroll_bar.  */
#define SCROLL_BAR_WIN32_WINDOW(ptr) \
  ((HWND) SCROLL_BAR_PACK ((ptr)->win32_window_low, (ptr)->win32_window_high))

/* Store a window id in a struct scroll_bar.  */
#define SET_SCROLL_BAR_WIN32_WINDOW(ptr, id) \
  (SCROLL_BAR_UNPACK ((ptr)->win32_window_low, (ptr)->win32_window_high, (int) id))


/* Return the outside pixel width for a vertical scroll bar on frame F.  */
#if 1
#define VERTICAL_SCROLL_BAR_PIXEL_WIDTH(f) \
(GetSystemMetrics(SM_CXVSCROLL))
#else
#define VERTICAL_SCROLL_BAR_PIXEL_WIDTH(f) \
(2*FONT_WIDTH ((f)->display.win32->font))
#endif
#ifdef VERTICAL_SCROLL_BAR_WIDTH
#undef VERTICAL_SCROLL_BAR_WIDTH
#define VERTICAL_SCROLL_BAR_WIDTH \
((VERTICAL_SCROLL_BAR_PIXEL_WIDTH(f) + FRAME_FONT_WIDTH(f) - 1) \
 / FRAME_FONT_WIDTH(f))
#endif

/* Return the outside pixel height for a vertical scroll bar HEIGHT
   rows high on frame F.  */
#define VERTICAL_SCROLL_BAR_PIXEL_HEIGHT(f, height) \
  ((height) * (f)->display.win32->line_height)

/* Return the inside width of a vertical scroll bar, given the outside
   width.  */
#define VERTICAL_SCROLL_BAR_INSIDE_WIDTH(width) \
  ((width) - VERTICAL_SCROLL_BAR_LEFT_BORDER - VERTICAL_SCROLL_BAR_RIGHT_BORDER)

/* Return the length of the rectangle within which the top of the
   handle must stay.  This isn't equivalent to the inside height,
   because the scroll bar handle has a minimum height.  

   This is the real range of motion for the scroll bar, so when we're
   scaling buffer positions to scroll bar positions, we use this, not
   VERTICAL_SCROLL_BAR_INSIDE_HEIGHT.  */
#if 0 /* modified by himi 96.9.18 */
#define VERTICAL_SCROLL_BAR_TOP_RANGE(height) \
  (VERTICAL_SCROLL_BAR_INSIDE_HEIGHT (height) - VERTICAL_SCROLL_BAR_MIN_HANDLE)
#else
#define VERTICAL_SCROLL_BAR_TOP_RANGE(height) 0xffff
#endif

/* Return the inside height of vertical scroll bar, given the outside
   height.  See VERTICAL_SCROLL_BAR_TOP_RANGE too.  */
#define VERTICAL_SCROLL_BAR_INSIDE_HEIGHT(height) \
  ((height) - VERTICAL_SCROLL_BAR_TOP_BORDER - VERTICAL_SCROLL_BAR_BOTTOM_BORDER)


/* Border widths for scroll bars.

   Scroll bar windows don't have any X borders; their border width is
   set to zero, and we redraw borders ourselves.  This makes the code
   a bit cleaner, since we don't have to convert between outside width
   (used when relating to the rest of the screen) and inside width
   (used when sizing and drawing the scroll bar window itself).

   The handle moves up and down/back and forth in a rectangle inset
   from the edges of the scroll bar.  These are widths by which we
   inset the handle boundaries from the scroll bar edges.  */
#define VERTICAL_SCROLL_BAR_LEFT_BORDER (2)
#define VERTICAL_SCROLL_BAR_RIGHT_BORDER (2)
#define VERTICAL_SCROLL_BAR_TOP_BORDER (2)
#define VERTICAL_SCROLL_BAR_BOTTOM_BORDER (2)

/* Minimum lengths for scroll bar handles, in pixels.  */
#define VERTICAL_SCROLL_BAR_MIN_HANDLE (5)


/* Manipulating pixel sizes and character sizes.
   Knowledge of which factors affect the overall size of the window should
   be hidden in these macros, if that's possible.

/* Return the upper/left pixel position of the character cell on frame F
   at ROW/COL.  */
#define CHAR_TO_PIXEL_ROW(f, row) \
  ((f)->display.win32->internal_border_width \
   + (row) * (f)->display.win32->line_height)
#ifdef WIN32_FONTSET
#define CHAR_TO_PIXEL_COL(f, col) \
  ((f)->display.win32->internal_border_width \
   + (col) * FRAME_FONT_WIDTH(f))
#else
#define CHAR_TO_PIXEL_COL(f, col) \
  ((f)->display.win32->internal_border_width \
   + (col) * FONT_WIDTH ((f)->display.win32->font))
#endif

/* Return the pixel width/height of frame F if it has
   WIDTH columns/HEIGHT rows.  */
#define CHAR_TO_PIXEL_WIDTH(f, width) \
  (CHAR_TO_PIXEL_COL (f, width) \
   + (f)->display.win32->vertical_scroll_bar_extra \
   + (f)->display.win32->internal_border_width)
#define CHAR_TO_PIXEL_HEIGHT(f, height) \
  (CHAR_TO_PIXEL_ROW (f, height) \
   + (f)->display.win32->internal_border_width)


/* Return the row/column (zero-based) of the character cell containing 
   the pixel on FRAME at ROW/COL.  */
#define PIXEL_TO_CHAR_ROW(f, row) \
  (((row) - (f)->display.win32->internal_border_width) \
   / (f)->display.win32->line_height)

#ifdef WIN32_FONTSET
#define PIXEL_TO_CHAR_COL(f, col) \
  (((col) - (f)->display.win32->internal_border_width) \
   / FRAME_FONT_WIDTH (f))
#else
#define PIXEL_TO_CHAR_COL(f, col) \
  (((col) - (f)->display.win32->internal_border_width) \
   / FONT_WIDTH ((f)->display.win32->font))
#endif

/* How many columns/rows of text can we fit in WIDTH/HEIGHT pixels on
   frame F?  */
#define PIXEL_TO_CHAR_WIDTH(f, width) \
  (PIXEL_TO_CHAR_COL (f, ((width) \
			  - (f)->display.win32->internal_border_width \
			  - (f)->display.win32->vertical_scroll_bar_extra)))
#define PIXEL_TO_CHAR_HEIGHT(f, height) \
  (PIXEL_TO_CHAR_ROW (f, ((height) \
			  - (f)->display.win32->internal_border_width)))

/* If a struct input_event has a kind which is selection_request_event
   or selection_clear_event, then its contents are really described
   by this structure.  */

/* For an event of kind selection_request_event,
   this structure really describes the contents.  */
struct selection_input_event
{
  int kind;
  HWND requestor;
  ATOM selection, target, property;
  Time time;
};

#define SELECTION_EVENT_REQUESTOR(eventp)	\
  (((struct selection_input_event *) (eventp))->requestor)
#define SELECTION_EVENT_SELECTION(eventp)	\
  (((struct selection_input_event *) (eventp))->selection)
#define SELECTION_EVENT_TARGET(eventp)	\
  (((struct selection_input_event *) (eventp))->target)
#define SELECTION_EVENT_PROPERTY(eventp)	\
  (((struct selection_input_event *) (eventp))->property)
#define SELECTION_EVENT_TIME(eventp)	\
  (((struct selection_input_event *) (eventp))->time)


/* Interface to the face code functions.  */

/* Create the first two computed faces for a frame -- the ones that
   have GC's.  */
extern void init_frame_faces (/* FRAME_PTR */);

/* Free the resources for the faces associated with a frame.  */
extern void free_frame_faces (/* FRAME_PTR */);

/* Given a computed face, find or make an equivalent display face
   in face_vector, and return a pointer to it.  */
extern struct face *intern_face (/* FRAME_PTR, struct face * */);

/* Given a frame and a face name, return the face's ID number, or
   zero if it isn't a recognized face name.  */
extern int face_name_id_number (/* FRAME_PTR, Lisp_Object */);

/* Return non-zero if FONT1 and FONT2 have the same size bounding box.
   We assume that they're both character-cell fonts.  */
extern int same_size_fonts (/* XFontStruct *, XFontStruct * */);

/* Recompute the GC's for the default and modeline faces.
   We call this after changing frame parameters on which those GC's
   depend.  */
extern void recompute_basic_faces (/* FRAME_PTR */);

/* Return the face ID associated with a buffer position POS.  Store
   into *ENDPTR the next position at which a different face is
   needed.  This does not take account of glyphs that specify their
   own face codes.  F is the frame in use for display, and W is a
   window displaying the current buffer.

   REGION_BEG, REGION_END delimit the region, so it can be highlighted.  */
extern int compute_char_face (/* FRAME_PTR frame,
				 struct window *w,
				 int pos,
				 int region_beg, int region_end,
				 int *endptr */);
/* Return the face ID to use to display a special glyph which selects
   FACE_CODE as the face ID, assuming that ordinarily the face would
   be BASIC_FACE.  F is the frame.  */
extern int compute_glyph_face (/* FRAME_PTR, int */);

#define WM_EMACS_SIZE     (WM_USER+2000)
#define WM_EMACS_MOVE     (WM_USER+2001)
#define WM_EMACS_DESTROY  (WM_USER+2002)
#define WM_EMACS_ACTIVATE (WM_USER+2003) 
/* to report "switch-buffer" event by himi */
#define WM_IME_REPORT     0x0280
#define IR_STRING         0x140
#define IR_OPENCONVERT    0x120
#define IR_CLOSECONVERT   0x122

#define IDM_UNDO 100
#ifdef WIN32_SCROLLBAR
#define WM_EMACS_VSCROLL (WM_USER+2004)
#define WM_EMACS_CREATE_SCROLLBAR (WM_USER+2005)
#define WM_EMACS_DESTROY_SCROLLBAR (WM_USER+2006)
#endif
#ifdef WIN32_INTELLIMOUSE
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 0x20A
#endif
#ifndef WHEEL_DELTA
#define WHEEL_DELTA 120
#endif
#ifndef SPI_GETWHEELSCROLLLINES
#define SPI_GETWHEELSCROLLLINES 104
#endif
#ifndef WHEEL_PAGESCROLL
#define WHEEL_PAGESCROLL (0xffffffff)
#endif
#endif
#ifdef MULE
#ifndef VK_KANJI
#define VK_KANJI 0x19
#endif
#ifndef VK_KANA
#define VK_KANA  0x15
#endif
#define VK_COMPEND 0x1A
#define WM_MULE_IME_REPORT (WM_USER+2100)
#define WM_MULE_IME_STATUS (WM_USER+2101)
#endif

extern void win32_fill_rect () ;

#define win32_fill_area(f,hdc,pix,x,y,nx,ny) \
{ \
    RECT rect ; \
    rect.left = x ; \
    rect.top = y ; \
    rect.right = x + nx ; \
    rect.bottom = y + ny ; \
    win32_fill_rect (f,hdc,pix,&rect) ; \
}

#define win32_clear_rect(f,hdc,lprect) \
win32_fill_rect (f,hdc,f->display.win32->background_pixel,lprect)

#define win32_clear_area(f,hdc,x,y,nx,ny) \
win32_fill_area (f,hdc,f->display.win32->background_pixel,x,y,nx,ny)

extern FONT_TYPE *win32_load_font () ;
extern void win32_unload_font () ;

#define MapMode(hdc) (hdc)
/* extern HDC MapMode() ; */
/* Modified by himi */

#define MyGetDC(hwnd) (MapMode(GetDC(hwnd)))

extern int GetKeyModifierState();
#define WIN32GETMODIFIER GetKeyModifierState()

#define WIN32GETMOUSEMODIFIER(mouse) \
 (WIN32GETMODIFIER | (mouse ? up_modifier : down_modifier))

#ifdef MULE
#define MULE_WIN32_DECODE_BUF buf_mule_decode_01211

#define MULE_WIN32_DECODE_ALLOC_PREDEFINE \
TCHAR *MULE_WIN32_DECODE_BUF; \
coding_type mccode_mule_decode_01211; \
extern Lisp_Object Vwin32_system_coding_system;


#define MULE_WIN32_DECODE_ALLOC(len) \
(encode_code(Fcheck_code(Vwin32_system_coding_system), \
	     &mccode_mule_decode_01211)); \
(MULE_WIN32_DECODE_BUF = alloca(DECODE_BUF_SIZE((len), \
						&mccode_mule_decode_01211))); \
(CODE_CNTL(&mccode_mule_decode_01211) |= CC_END)

#define MULE_WIN32_DECODE(string, len) \
(decode(&mccode_mule_decode_01211, \
	(string), MULE_WIN32_DECODE_BUF, (len)))
#endif

#endif /* __WIN32TERM__ */
